#include <QFile>
#include <QMessageBox>
#include <math.h>

#include "CScorrInterp.h"
#include "XYQt/PeriodicTable.h"
#include "../corteo/techniques.h"

// Create a CScorrInterp class for cross-section correction by interpolation in a table
//
// The CS stamp is in the format AB.MMMZZmmmzz where A is the simulation type, B is the correction type (1 for Ratio to Rutherfor, 2 for value from table),
// MMM is the beam mass, ZZ is the beam atomic number, mmmm is the recoil mass, zz is the recoil atomic number
// (The stamp has been gererated in this format so it looks like a number and there are no problems reading Corteo output with Python, Matlab or other software)
//
// The constructor tries to generate a filename in the format, e.g.,
// O-16_He-4_He-4_scat_RR.ccs for alpha scattering off oxygen in "relative to Rutherford" units
// or Si-nat_H-1_H-1_scat_CS.ccs for protons scattring off Si with natural isotope abundances, in mb (millibarn) units
// It then tries to load that file.
//
// Upon success, the constructor sets dataLoaded = true.
// If anything fails (invalid CSstamp, could not open file, etc.) dataLoaded = false,
// and an error message is available in errorMess
CScorrInterp::CScorrInterp(QString _CSstamp) {

    CSstamp = _CSstamp;
    errorMess = "";

    int dot = CSstamp.indexOf('.');
    if(dot==-1) {
        errorMess = "No '.' in CS stamp.";
        dataLoaded = false;
        return;
    }

    bool ok;
    simulationType = CSstamp.left(dot-1).toInt(&ok);
    if(!ok) {
        errorMess = QString("In CS stamp, could not convert '%1' into a simulation type integer value").arg(CSstamp.left(dot-1));
        dataLoaded = false;
        return;
    }

    int val = CSstamp.mid(dot-1,1).toInt(&ok);
    if(ok && val == 1) ratio2Rutherford = true;
    else if(ok && val == 2) ratio2Rutherford = false;
    else {
        errorMess = "In CS stamp, type was not set to 1 (ratio to Rutherford) or 2 (assign value from table)";
        dataLoaded = false;
        return;
    }

    if(ok) M1 = CSstamp.mid(dot+1,3).toDouble(&ok);
    if(ok) Z1 = CSstamp.mid(dot+4,2).toInt(&ok);
    if(ok) M2 = CSstamp.mid(dot+6,3).toDouble(&ok);
    if(ok) Z2 = CSstamp.mid(dot+9,2).toInt(&ok);

    if(!ok || isnan(M1) || isnan(M2) || isinf(M1) || isinf(M2) || Z1==0 || M1 <=0 || Z2==0 || M2 <0) {
        errorMess = QString("Could not convert this part of a CS stamp: '%1'."
                            "\nFormat expected is 'MMMZZmmmzz' where M1=MMM, Z1=ZZ, M2=mmm, Z2=zz.")
                .arg(CSstamp.mid(dot+1,10));
        dataLoaded = false;
        return;
    }

    bool useNatural = M2 == 0; // if M2 set to 0, use the correction file for natural isotope abundances of the element

    // generate file name in the format, e.g., O-16_He-4_He-4_scat_RR.ccs for alpha scattering off oxygen in "relative to Rutherford" units
    // or Si-nat_H-1_H-1_scat_CS.ccs for protons scattring off Si with natural isotope abundances, in mb (millibarn) units
    PeriodicTable pt;
    reactName = "data/" + pt.returnAtom(Z2).symbol + "-";

    QString Ms;
    if(useNatural)
        Ms += "nat";
    else {
        Ms.sprintf("%.0f", M2);
    }
    reactName += Ms;

    Ms.sprintf("%.0f", M1);
    QString reactant = "_" + pt.returnAtom(Z1).symbol + "-" + Ms;
    reactName += reactant + reactant;

    switch (simulationType) {
    case TECH_ERD:
    case TECH_ERD_TOF:
        reactName += "_recoil";
        break;
    case TECH_COINC:
    case TECH_RBS:
    case TECH_RBS_TOF:
        reactName += "_scat";
        break;
    default:
        errorMess = QString("In CS stamp, simulation type specified '%1' does not exist or cannot have a cross-section correction.'%1' into a simulation type integer value").arg(simulationType);
        dataLoaded = false;
        return;
    }


    if(ratio2Rutherford) {
        reactName += "_RR.ccs"; // ratio to Rutherford
    } else {
        reactName += "_CS.ccs"; // plain cross-section in Angs^2/str
    }

    dataLoaded = loadCSTable();

}

// load the cross-section table related to the Z1, Z2 and simulationType
// Energies are in eV, angles in degrees, cross-section in Angs.^2,
bool CScorrInterp::loadCSTable() {

    int i, j;
    QFile file(reactName);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        errorMess = QString("Could not open file '%1'").arg(reactName);
        return false;
    }

    // Expected file structure is as follows (where CSij are cross-section values):
    // ... arbitrary number of lines...
    // ... contaning for example the header of R33 files...
    // Units: rr {or} mb   <- we scan until we find this and check that the cross section units are as specified
    // Data:    <- then we scan the file until we find this!
    // M N      <- Where M is the number of energy values (number of rows is M+1) and N is the number of angle values (number of colums is N+1)
    // 0 Angle1 Angle2 ... AngleN
    // Energy1 CS11 CS12 ... CS1N
    // Energy2 CS21 CS22 ... CS2N
    // ...
    // EnergyM CSM1 CSM2 ... CSMN

    // Energies are in keV and angles are in degrees
    // Angle and Energy values must be in strictly increasing order (no identical values)

    // so, scan file until we find "Data:"
    QString line;
    while (!file.atEnd()) {
          line = QString(file.readLine());
          if(line.startsWith("Units:"))
              break;
    }
    if(file.atEnd()) {
        errorMess = QString("Did not find line 'Units:' in cross-section correction file %1").arg(reactName);
        return false;
    }

    if(ratio2Rutherford && !line.toLower().contains("rr")) {
        errorMess = QString("In cross-section correction file %1, 'Units:' are supposed to be 'rr' (i.e. Rutherford ratio)").arg(reactName);
        return false;
    }
    if(!ratio2Rutherford && !line.toLower().contains("mb")) {
        errorMess = QString("In cross-section correction file %1, 'Units:' are supposed to be 'mb' (i.e. millibarns)").arg(reactName);
        return false;
    }

    while (!file.atEnd()) {
          line = QString(file.readLine());
          if(line.startsWith("Data:"))
              break;
    }
    if(file.atEnd()) {
        errorMess = QString("Did not find line 'Data:' in cross-section correction file %1").arg(reactName);
        return false;
    }

    // first line after "Data:" contains the number of angles (columns) and energies (rows)
    line = QString(file.readLine());
    QStringList val = line.simplified().split(" ");
    int numEner = val[0].toInt();
    int numAngle = val[1].toInt();
    if(numEner<=0 || numAngle <=0 || val.count()!=2) {
        errorMess =  QString("Number of energies and/or angles not set properly in cross-section correction file %1").arg(reactName);
        return false;
    }

    CSarray.resize(numAngle);
    angle.resize(numAngle);
    E.resize(numEner);

    // following line contains the angles (but first value is a dummy = 0, so angles are at the top of the column, since the first
    line = QString(file.readLine());
    QStringList vals = line.simplified().split(" ");
    if(vals.size()!=numAngle+1){
        errorMess = QString("Wrong number of angle values in cross-section correction file %1").arg(reactName);
        return false;
    }
    for( j=0; j<numAngle; j++) {
        bool ok;
        angle[j]=vals[j+1].toDouble(&ok);
        if(!ok) {
            errorMess =  QString("Error converting '%1' into an angle in cross-section correction file %2").arg(vals[j+1]).arg(reactName);
            return false;
        }
        if(j>0 && angle[j-1]>=angle[j]) {
            errorMess =  QString("Angle values not in strictly increasing order in cross-section correction file %1").arg(reactName);
            return false;
        }
    }

    // read energy rows, each containing an energy value followed by numAngle cross-section values
    for( i=0; i<numEner; i++) {
        line = QString(file.readLine());
        QStringList vals = line.simplified().split(" ");
        if(vals.size()!=numAngle+1){
            errorMess =  QString("Wrong number of values at row %1 in cross-section correction file %2").arg(i+1).arg(reactName);
            return false;
        }
        CSarray[i].resize(numAngle);
        for( j=0; j<numAngle+1; j++) {
            bool ok;
            double val =vals[j].toDouble(&ok);
            if(!ok) {
                errorMess =  QString("Error converting '%1' into a number at row %2 in cross-section correction file %3").arg(vals[j]).arg(j).arg(reactName);
                return false;
            }
            if(j==0) {
                E[i]=val*1000.; // keV converted to eV
                if(i>0 && E[i-1]>=E[i]) {
                    errorMess =  QString("Energy values not in strictly increasing order in cross-section correction file %1").arg(reactName);
                    return false;
                }

            } else {
                if(!ratio2Rutherford) val *= 1E16*1E-27; // if we use directly the tabulated values, convert millibarn to Angs^2
                CSarray[i][j-1] = val;
            }
        }
    }

    dataLoaded = true;
    return true;
}


// interpolate linearly into cross-section table
// return  0 if interpolation is successfull and a correction or replacement to *CS value was applied
// return -1 if below minimum energy
// return -2 if below minimum angle (here, we cannot tell if at high energy, the *CS value is still good)
// return -3 if above maximum energy
// return -4 if above maximum angle (usually, tables are provided up to 180 deg., so that should not happen)
// return -5 if division by 0 in interpolation (should not happen since we checked that energies and
// angles were stricktly increasing values)

// hence, if the returned value is >= -1, the value can be accepted, since the value of *CS
// is either corrected according to the table, or returned without correction at low energy
// Accepting a returned value  >= -2 means that we accept *CS unchanged also for
// small angle scattering, whatever is the energy.
int CScorrInterp::interp(double Elab, double AngleLab, double *CS ) {
    int i, j;

    int nAngle = angle.count();
    int nEner  = E.count();

    if(Elab<=E[0])
        return -1; // below (or equal to) minimum energy of the table

    if(AngleLab<=angle[0])
        return -2; // below (or equal to) the minimum angle

    // find energy index for interpolation
    for(i=1; i<nEner; i++)
        if(Elab <= E[i])
            break;

    if(i==nEner)
        return *CS; // the prevous loop didn't break, hence the value is above upper energy limit: return uncorrected CS

    // find angle index for interpolation
    for(j=1; j<nAngle; j++)
        if(AngleLab <= angle[j])
            break;

    if(j==nAngle)
        return -4; // the prevous loop didn't break, hence the value is above upper limit

    double div = (angle[j]-angle[j-1])*(E[i]-E[i-1]);
    if(div==0) return -5; // should not happen since we checked that energies and angles were continuously increasing values

    // at this point, energy should be between i-1 and i, and angle between j-1 and j
    // lets carry out a bilinear interpolation
    double val = (CSarray[i-1][j-1]*(E[i  ]-Elab)*(angle[j  ]-AngleLab)
                 +CSarray[i-1][j  ]*(E[i  ]-Elab)*(AngleLab-angle[j-1])
                 +CSarray[i  ][j-1]*(Elab-E[i-1])*(angle[j  ]-AngleLab)
                 +CSarray[i  ][j  ]*(Elab-E[i-1])*(AngleLab-angle[j-1]) ) /div;

    if(ratio2Rutherford) *CS*=val; // correction by a factor
    else *CS = val; // assing table value

    return 0;

}
