/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <QHBoxLayout>
#include "CalibrationForm.h"


CalibrationForm::CalibrationForm(QWidget * parent, int ZZ, int numlayer)  : QWidget(parent) {

  Z=ZZ;

  setupUi(this);

  QDoubleValidator * dValidator = new QDoubleValidator(parent);
  QDoubleValidator * dPosValidator = new QDoubleValidator(parent); // validator positive parameters
  dPosValidator->setBottom(0);

  lineEditQuad->setValidator(dValidator);
  lineEditSlope->setValidator(dPosValidator);
  lineEditZero->setValidator(dValidator);
  lineEditResolution->setValidator(dPosValidator);
  
  // create numlayer compound correction edit fields
  tabWidgetCompCorr->removeTab(0); // remove default compound corr tab
  for(int ilayer=0;ilayer<numlayer;ilayer++) {

    leCompCorrList.append(new QLineEdit("1")  );  // compound correction edit fields
    leCompCorrList.last()->setValidator(dPosValidator);
    leCompCorrList.last()->setMinimumHeight(25);
    QLabel * label = new QLabel("Compound Corr.");
    label->setMinimumHeight(25);

    QHBoxLayout *layout = new QHBoxLayout;
    layout->addWidget(label);
    layout->addWidget(leCompCorrList.last());

    QWidget *w = new QWidget;
    w->setLayout(layout);
    tabWidgetCompCorr->addTab(w, "");
  }

}


QDomElement CalibrationForm::XMLNode(QDomDocument &doc) {

  QDomElement e = doc.createElement( "calibration" );
  e.setAttribute( "Z",     QString::number(Z) );
  e.setAttribute( "zero",  lineEditZero->text());
  e.setAttribute( "slope", lineEditSlope->text());
  e.setAttribute( "quad",  lineEditQuad->text());
  e.setAttribute( "resolution", lineEditResolution->text());

  // compound corrections
  for(int ilayer=0; ilayer<tabWidgetCompCorr->count(); ilayer++) {
    QDomElement ee;
    
    ee = doc.createElement( "layer_param" );
    ee.setAttribute( "layer",  ilayer );
    ee.setAttribute( "compound_corr", leCompCorrList[ilayer]->text() );
    e.appendChild( ee );
  }

  return e;
}


// set parameters from an XML node
void CalibrationForm::setFromNode(QDomNode &n) {

  QDomElement e = n.toElement();
  lineEditZero->setText(  e.attribute("zero", "0"));
  lineEditSlope->setText( e.attribute("slope", "1"));
  lineEditQuad->setText(  e.attribute("quad", "0"));
  lineEditResolution->setText( e.attribute("resolution", "0"));

  // read compound correction factors
  QDomNode nn = n.firstChild();
  while( !nn.isNull() ) {
    QDomElement ee = nn.toElement();
    if( !ee.isNull() ) {
      if( ee.tagName() == "layer_param" ) {
        int ilayer = ee.attribute("layer", "-1").toInt();
        if(ilayer>=0 && ilayer<tabWidgetCompCorr->count()) {
          leCompCorrList[ilayer]->setText(ee.attribute("compound_corr", "1"));
        }
      } 
    } 
    nn = nn.nextSibling();
  } 
} 


// if all string are assigned, return a string containting zero slope quad resolution 
// otherwise return a null string
QString CalibrationForm::corteo(bool *error) {

  *error = false;
  QString serror;

  if(lineEditZero->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="calibration zero";
  }
  if(lineEditSlope->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="slope";
  }
  if(lineEditQuad->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="quadratic calibration term";
  }
  if(lineEditResolution->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="calibration zero";
  }

  if(*error) return serror;

  return lineEditZero->text()  + " " 
       + lineEditSlope->text() + " " 
       + lineEditQuad->text()  + " " 
       + lineEditResolution->text();
}


// return compound correction for this element in layer ilayer
// if edit field is empty, return "1"
// if ilayer is out of bound, return a null string
QString CalibrationForm::compoundCorr(int ilayer) {
  if(ilayer<0 || ilayer>=tabWidgetCompCorr->count()) return QString();

  if(leCompCorrList[ilayer]->text().isEmpty()) return QString("1");

  return leCompCorrList[ilayer]->text();
}


void CalibrationForm::setCalibrationFieldsEnabled(bool enable) {
  labelZero->setEnabled(  enable );
  labelSlope->setEnabled(  enable );
  labelQuad->setEnabled(  enable && Z>0  ); // Z==0 is TOF, quad term not implemented
  labelResolution->setEnabled(  enable );

  lineEditZero->setEnabled(  enable );
  lineEditSlope->setEnabled(  enable );
  lineEditQuad->setEnabled(  enable && Z>0 ); // Z==0 is TOF, quad term not implemented
  lineEditResolution->setEnabled(  enable );

}

void CalibrationForm::setCompoundCorrEnabled(int ilayer, bool enable) {
  if(ilayer<leCompCorrList.count())
      leCompCorrList[ilayer]->setEnabled(  enable );
}
