/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <QMessageBox>
#include <QFileDialog>
#include <QSplashScreen>
#include <QTextCursor>
#include <QPainter>
#include <QWheelEvent>
#include <QInputDialog>
#include <QDateTime>

#include <math.h>
#include <limits.h>
#include "CorteoParamWidget.h"
#include "XYQt/FSDataSet.h"
#include "constants.h"
#include "colors.h"


#define BLACK 0xFF000000 // ARGB for black with full opacity A=255
#define SUBSTRATE_COLOR 0x123456 // in bitmap mode, special code assign to the underlying substrate layer
#define MAX_GEN_PIXMAP_WIDTH 2048 // largest size for generated pixmap

// class based on QLabel that implements wheelEvent so it scrolls the specified scrollbar
class pixelArea: public QLabel
{
  public:
    pixelArea( QWidget * pparent, QScrollBar *ssb):QLabel(pparent)  {
        sb = NULL;
        if(ssb != NULL)
            sb = ssb;
    }

    void wheelEvent(QWheelEvent* event)
    {
        if(sb)
            sb->setValue( sb->value() + (event->pixelDelta().y()/10) );
    }

private:
  QScrollBar * sb;
};


CorteoParamWidget::CorteoParamWidget(QWidget * parent, int maxlayer, int maxelement, QSplashScreen * splash)
                  : QWidget(parent)
{
  // setup base class
  setupUi(this);

  labelPixels = new pixelArea(framePixels, horizontalScrollBar);
  labelPixels->setGeometry(QRect(0,0,framePixels->frameSize().width(), framePixels->frameSize().height()));

  // icons indicating if a layer or an element is used (green) or not (gray)
  green = QIcon(":/images/green.png");
  gray  = QIcon(":/images/gray.png");

  // validators for numerical edit fields
  FSDoubleValidator * fsdValidator =    new FSDoubleValidator(this);
  FSDoubleValidator * fsdPosValidator = new FSDoubleValidator(this, 0.0); // validator positive parameters
  QIntValidator     * intValidator =    new QIntValidator(this);
  intValidator->setBottom(0);

  // set technique items
  techniqueName  << "RBS"    << "RBS-TOF"    << "ERD"    << "ERD-TOF"  << "Coincidence" << "Full RBS"
                 << "Full RBS background" << "(Implantation)" << "(Full cascade)"
                 << "Energy-depth map" << "Energy-depth list"  ;
  techniqueCode << TECH_RBS << TECH_RBS_TOF << TECH_ERD << TECH_ERD_TOF << TECH_COINC  << TECH_FULLRBS
                << TECH_FULLBACK << TECH_IMPLANT  << TECH_IMPREC << TECH_ENERDEPTHMAP << TECH_ENERDEPTHLIST;
  comboBoxTechnique->addItems(techniqueName);
  comboBoxTechnique->setCurrentIndex(0);

  comboBoxScreening->addItem("None");
  comboBoxScreening->addItem("Andersen");
  comboBoxScreening->addItem("Universal");
  comboBoxScreening->addItem("Lenz-Jensen");

  comboBoxStraggling->addItem("none");
  comboBoxStraggling->addItem("Bohr");
  comboBoxStraggling->addItem("Chu");
  comboBoxStraggling->addItem("Yang");

  comboBoxNthread->addItem("1");
#ifdef MAX_THREAD
  for(int ithread=2; ithread<=MAX_THREAD; ithread +=2)
      comboBoxNthread->addItem(QString().setNum(ithread));
#endif


  if(splash) splash->showMessage("setting up layer forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // set layer tabs
  tabWidgetLayers->removeTab(0); // remove default layer tab
  for(int ilayer=0;ilayer<maxlayer;ilayer++) {

    // create layer parameter forms
    LayerForm * lf;
    tabWidgetLayers->addTab(lf = new LayerForm(this, maxelement), "");
    layer.append(lf);

    // connect layer move buttons and useLayer checkbox
    connect( layer[ilayer]->pushButtonMoveToPrevious,   SIGNAL(clicked()), SLOT(moveLayerToPrevious()) );
    connect( layer[ilayer]->pushButtonMoveToNext,       SIGNAL(clicked()), SLOT(moveLayerToNext()) );
    connect( layer[ilayer]->checkBoxUseLayer,           SIGNAL(clicked()), SLOT(useLayerChanged()) );

    // connect element button and checkboxes for update
    for(int ielement=0;ielement<maxelement;ielement++) {
      connect(layer[ilayer]->element[ielement]->checkBoxUseElement, SIGNAL(clicked()), SLOT(update()) );
      connect(layer[ilayer]->element[ielement]->checkBoxSimulate,   SIGNAL(clicked()), SLOT(update()) );
      connect(layer[ilayer]->element[ielement]->checkBoxIsotopes,   SIGNAL(clicked()), SLOT(update()) );
    }
  }

  // create dialog to edit detector foil
  foilDialog = new QDialog(this);
  foil = new LayerForm(foilDialog, maxelement);
  foil->labelLayerNumber->setText("Absorber");

  // for foils, hide "move" buttons and other fields
  foil->pushButtonMoveToPrevious->hide();
  foil->pushButtonMoveToNext->hide();
  foil->stackedWidgetLayerNumber->hide();

  // also hide stuff related to simulation of foil elements
  int nFoilElem = foil->element.count();
  for(int i=0; i<nFoilElem; i++)
      foil->element[i]->stackedWidget->hide();

  connect( pushButtonEditFoil, SIGNAL(clicked()), SLOT(editFoil()) );

  if(splash) splash->showMessage("setting up detector forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // set detector tabs
  tabWidgetDetectors->removeTab(0); // remove default detector tab
  DetectorForm * df;

  // create detectors with specific names and include them in detector tabs
  tabWidgetDetectors->addTab(df = new DetectorForm("RBS", this), "RBS");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("ERD", this), "ERD");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("RBS-TOF", this), "RBS-TOF");
  df->labelDistance->setText("Distance from 1st [mm]");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("ERD-TOF", this), "ERD-TOF");
  df->labelDistance->setText("Distance from 1st [mm]");
  detector.append(df);

  // Calibration forms
  CalibrationForm * cf;
  if(splash) splash->showMessage("setting up calibration forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // at index 0, set a TOF calibration form
  tabWidgetCalibrations->addTab(cf = new CalibrationForm(this, 0, 0), "");
  cf->lineEditQuad->setText("0"); // quadtratic calibration term not implemented in TOF
  cf->lineEditQuad->setEnabled(false);
  cf->labelQuad->setText("Quad (disabled):");
  cf->labelQuad->setEnabled(false);

  cf->labelSlope->setText("Slope [ns/ch]:");
  cf->labelZero->setText("Zero [ns]:");
  cf->labelResolution->setText("Resolution [ns]:");

  calibration.append(cf);

  // then, append a calibration form for each element in the periodic table
  PeriodicTable periodicTable;
  for(int Z=1; Z<=periodicTable.countAtoms(); Z++) {
    tabWidgetCalibrations->addTab(cf = new CalibrationForm(this, Z, maxlayer), "");
    calibration.append(cf);
  }
  // now remove them all, update()  will put back the useful ones
  while(tabWidgetCalibrations->count()) tabWidgetCalibrations->removeTab(0);

  // verify if move button of tab 0 should be enabled
  moveLayerEnable(0);

  // set validators for toolbox numerical edit fields
  lineEditNumberProjectiles->setValidator(intValidator);
  lineEditRandomSeed->setValidator(intValidator);
  lineEditPixelPerAngs->setValidator(fsdPosValidator);
  lineEditNumberChannels->setValidator(intValidator);
  lineEditEnergyDiffDisc->setValidator(fsdPosValidator);

  lineEditMeanFreePath->setValidator(fsdValidator);
  lineEditConeAngle->setValidator(fsdPosValidator);
  lineEditMinimumEnergy->setValidator(fsdPosValidator);
  lineEditCutoff->setValidator(fsdPosValidator);
  lineEditEnergy->setValidator(fsdPosValidator);
  lineEditCharge->setValidator(fsdValidator);
  lineEditAngleToNormal->setValidator(fsdPosValidator);
  lineEditAzimAngle->setValidator(fsdValidator);
  lineEditBeamDiameter->setValidator(fsdPosValidator);
  lineEditWfactor->setValidator(fsdPosValidator);
  lineEditHfactor->setValidator(fsdPosValidator);

  // connect a few widgets to their slots
  connect( pushButtonBeam,     SIGNAL(clicked()),                SLOT(setBeam())            );
  connect( tabWidgetLayers,    SIGNAL(currentChanged(int)),      SLOT(moveLayerEnable(int)) );
  connect( pushButtonAbout,    SIGNAL(clicked()),                SLOT(aboutCorteoUi())      );

  connect( comboBoxTechnique,  SIGNAL(currentIndexChanged(int)), SLOT(update())     );

  connect( pushButtonLoad,     SIGNAL(clicked()),                SLOT(loadFile())   );
  connect( pushButtonSaveAs,   SIGNAL(clicked()),                SLOT(saveFileAs()) );

  connect( pushButtonShowCorteoLog,   SIGNAL(clicked()),         SLOT(showCorteoLog()) );
  connect( pushButtonLoadPixels,      SIGNAL(clicked()),         SLOT(loadPixels())    );
  connect( pushButtonGenPixels,      SIGNAL(clicked()),          SLOT(genPixels())    );

  connect( checkBoxUsePixels,  SIGNAL(stateChanged(int)),        SLOT(updateUsePixels()) );

  connect( horizontalScrollBar,SIGNAL(valueChanged(int)),        SLOT(drawPixels(int)));
  connect( toolBox,            SIGNAL(currentChanged(int)),      SLOT(drawPixels())   );

  // try loading last saved parameters
  if(splash) splash->showMessage("loading parameters...",Qt::AlignHCenter|Qt::AlignBottom);
  if( !loadParam("CorteoUi.cml") ) {

    // load last saved params did not work so initialize elements to some default values:
    beamElement = Atom(ElementForm::setElement(2, pushButtonBeam, comboBoxBeamMass)); // beam = He

    // set silicon as default elements in layers
    for(int ilayer=0; ilayer<maxlayer; ilayer++)
      for(int ielement=0; ielement<maxelement; ielement++) {
        layer[ilayer]->element[ielement]->setElement(14);
      }

    // set carbon as default elements in foil
    for(int ielement=0; ielement<maxelement; ielement++)
      foil->element[ielement]->setElement(6);
  }
  checkBoxUsePixels->toggle();
  checkBoxUsePixels->toggle();

  beamColor = Qt::red;
  deteColor = Qt::blue;

  horizontalScrollBar->show();
  update();
  drawPixels();
  setLoadedMaterialColors();
}


// determine straggling type from radiobutton state
int CorteoParamWidget::stragglingType() {

    return comboBoxStraggling->currentIndex();
}


// set straggling selection, or to None if wrong type
void CorteoParamWidget::setStragglingType(int type) {
    if(type <0 || type >3) {
        QMessageBox::warning(this, "Corteo user interface", "Unknown straggling type, setting it to None.");
        type = 0;
    }
    comboBoxStraggling->setCurrentIndex(type);

}


// call setElement for beam button and mass selection box
void CorteoParamWidget::setBeam() {
  beamElement = ElementForm::selectElement(this, pushButtonBeam, comboBoxBeamMass);
  update();
}


void CorteoParamWidget::useLayerChanged() {
  setUseLayerIcon(tabWidgetLayers->currentIndex());
  update(); // update the list of detected elements
}


// if using pixels, show material # and set combobox to select the color corresponding to a material
// if not using pixels, show layer # and layer-order arrows to show up
void CorteoParamWidget::updateLayer(int ilayer) {

    if(checkBoxUsePixels->isChecked()) {
        layer[ilayer]->stackedWidgetLayerNumber->setCurrentIndex(0);
        layer[ilayer]->pushButtonMoveToPrevious->hide();
        layer[ilayer]->pushButtonMoveToNext->hide();
        layer[ilayer]->comboBoxMaterial->clear();
        layer[ilayer]->labelLayerNumber->setText(QString("Mater. #%1").arg(ilayer));

        // colors is a vector of colors used
        QVector<QRgb> colors = pixelsColorUsed.toList().toVector();
        layer[ilayer]->comboBoxMaterial->clear();
        // qDebug() << "numcolor=" << colorlist.count();

        // create an icone filled with the color, and add it to the combobox...
        for(int i=0; i<colors.count(); i++) {
            if(colors.at(i) != BLACK) { // ... except black, which is reserved for vacuum
                QImage image(32, 32, QImage::Format_Indexed8);
                image.setColorTable(colors);
                QColor color = colors.at(i);
                image.fill(color);
                //qDebug() << "color=" << color.name();

                // when adding the icone, also add the color as a QVariant data (last argument)
                layer[ilayer]->comboBoxMaterial->addItem(QIcon(QPixmap::fromImage(image)),
                                                         QString("Color %1").arg(i), colors.at(i));
            }
        }

        // a material can be selected as the substate on which the bitmap sits (with QVariant data SUBSTRATE_COLOR)
        layer[ilayer]->comboBoxMaterial->addItem("Substrate", SUBSTRATE_COLOR);

        // by default, assign first color 0 to first layer, etc.
        layer[ilayer]->comboBoxMaterial->setCurrentIndex(ilayer % (colors.count()+1));
    } else {
        layer[ilayer]->stackedWidgetLayerNumber->setCurrentIndex(1);
        layer[ilayer]->pushButtonMoveToPrevious->show();
        layer[ilayer]->pushButtonMoveToNext->show();
        layer[ilayer]->labelLayerNumber->setText(QString("Layer #%1").arg(ilayer));
        layer[ilayer]->labelThickness->show();
        layer[ilayer]->lineEditThickness->show();
    }
}


// set things affected by the checking/unchecking checkBoxUsePixels
void CorteoParamWidget::updateUsePixels() {

    for(int ilayer=0; ilayer<layer.count(); ilayer++)
        updateLayer(ilayer);

    bool isPixmap = checkBoxUsePixels->isChecked();

    //labelPixels->setEnabled(isPixmap);
    horizontalScrollBar->setEnabled(isPixmap);
    labelPixPerAngs->setEnabled(isPixmap);
    lineEditPixelPerAngs->setEnabled(isPixmap);
    pushButtonLoadPixels->setEnabled(isPixmap);
    pushButtonGenPixels->setEnabled(isPixmap);

    drawPixels();
}



void CorteoParamWidget::setUseLayerIcon(int ilayer) {
  if(ilayer>=tabWidgetLayers->count() || ilayer<0) return;

  // icon green or gray whether layer is used or not
  bool bl = layer[ilayer]->isUsed();
  tabWidgetLayers->setTabIcon(ilayer, bl?green:gray);
  updateLayer(ilayer);


  // change also the icon of all calibration forms compound correction tabs, in case they are used
  for(int i=0; i<calibration.count(); i++)
    calibration[i]->tabWidgetCompCorr->setTabIcon(ilayer, bl?green:gray);
}


// swap forms and container widget of two layers
bool CorteoParamWidget::swapLayers(int layer1, int layer2) {
  if( layer1<0 || layer2<0 || layer1>=layer.count() || layer2>=layer.count() || layer1==layer2 )
    return false;

  // swap layer forms in layer list
  layer.swap(layer1,layer2);

  // insert layer1 at position of layer2 (this does not create a new tab!)
  tabWidgetLayers->insertTab(layer2, tabWidgetLayers->widget(layer1), "");

  // swap compound corrections in calibration forms
  for (int i = 0; i < calibration.count(); i++) {
    // insert widget corresponding to layer1 at position of that of layer2
    // (this does not create a new tab!)
    QWidget * w = calibration[i]->tabWidgetCompCorr->widget(layer1);
    calibration[i]->tabWidgetCompCorr->insertTab(layer2, w, "");
  }

  setUseLayerIcon(layer1);
  setUseLayerIcon(layer2);

  return true;
}


// check if current layer tab can move up or down and enable corresponding buttons
void CorteoParamWidget::moveLayerEnable(int ilayer) {
  if( ilayer<0 || ilayer>=layer.count())
    return; // layer tabs not initialized

  layer[ilayer]->pushButtonMoveToPrevious->setEnabled(ilayer>0);
  layer[ilayer]->pushButtonMoveToNext->setEnabled(ilayer<layer.count()-1);
}


// swap current and previous layers so current layer tab moves down by one
bool CorteoParamWidget::moveLayerToPrevious() {
  int ilayer = tabWidgetLayers->currentIndex();
  if(ilayer<=0) return false;

  swapLayers(ilayer, ilayer-1);
  tabWidgetLayers->setCurrentIndex(ilayer-1);

  return true;
}


// swap current and next layers so current layer tab moves up by one
bool CorteoParamWidget::moveLayerToNext() {
  int ilayer = tabWidgetLayers->currentIndex();
  if(ilayer>=layer.count()-1) return false;

  swapLayers(ilayer, ilayer+1);
  tabWidgetLayers->setCurrentIndex(ilayer+1);

  return true;
}


// set labels and enable widgets according to the selected technique
void CorteoParamWidget::update() {

  int tech = technique();

  // cone angle only for IBA
  labelConeAngle->setEnabled(tech>0);
  lineEditConeAngle->setEnabled(tech>0);

  // energy discrimination enabled only for coincidence
  labelEnergyDiffDisc->setEnabled(tech == TECH_COINC);
  lineEditEnergyDiffDisc->setEnabled(tech == TECH_COINC);
  checkBoxLoadSeparateSpectra->setEnabled(tech > 0 || tech == TECH_IMPREC);

  // angle-to-detector-axis file only possible for IBA
  checkBoxGraphAngle->setEnabled(tech>0 || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  // detection toolbox tab
  toolBox->setItemEnabled(4,tech>0 || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  // disable all detector tabs then re-enable the useful ones
  for(int idet = 0; idet<detector.count(); idet++)
    tabWidgetDetectors->setTabEnabled(idet,false);

  switch(tech) {
  case TECH_IMPLANT :
  case TECH_ENERDEPTHMAP :
  case TECH_ENERDEPTHLIST :
    // no detection
    break;
  case TECH_RBS_TOF :
    tabWidgetDetectors->setTabEnabled(2,true);
    // no break
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
    tabWidgetDetectors->setTabEnabled(0,true);
    tabWidgetDetectors->setCurrentIndex (0);
    break;
  case TECH_ERD_TOF :
    tabWidgetDetectors->setTabEnabled(3,true);
    // no break
  case TECH_ERD :
    tabWidgetDetectors->setTabEnabled(1,true);
    tabWidgetDetectors->setCurrentIndex (1);
    break;
  case TECH_COINC :
    tabWidgetDetectors->setTabEnabled(0,true);
    tabWidgetDetectors->setTabEnabled(1,true);
    tabWidgetDetectors->setCurrentIndex (0);
    break;
  default:
    QMessageBox::warning(this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
    break;
  }

  // cutoff angle possible only for RBS cross section
  labelCutoff->setEnabled(isBeamDetectionTechnique() && tech != TECH_FULLRBS);
  lineEditCutoff->setEnabled(isBeamDetectionTechnique() && tech != TECH_FULLRBS);

  // update layers according to technique
  for(int ilayer=0; ilayer<layer.count(); ilayer++) {
    bool b = layer[ilayer]->update(tech>0, tech==TECH_IMPREC);
    tabWidgetLayers->setTabIcon(ilayer, b?green:gray); // icon green or gray whether layer is used or not
    updateLayer(ilayer);
  }

  // set charge label according to technique
  labelCharge->setText(tech>0?QString("Incident charge [C]:"):QString("Fluence [at/cm<sup>2</sup>]:"));


  // calibration tabs

  // update used/unused layers in calibration forms compound correction tabs
  for(int i=0; i<calibration.count(); i++) {
    for(int ilayer=0; ilayer<layer.count(); ilayer++) {
      bool bl = layer[ilayer]->checkBoxUseLayer->isChecked();
      // icon green or gray whether layer is used or not
      calibration[i]->tabWidgetCompCorr->setTabIcon(ilayer, bl?green:gray);
      calibration[i]->setCompoundCorrEnabled(ilayer, bl);
    }
  }

  // remove current calibration tabs
  while(tabWidgetCalibrations->count())
    tabWidgetCalibrations->removeTab(0); // this doesn't destroy the forms themselves

  // beam calibration always there, at least the compound corrections...
  tabWidgetCalibrations->addTab(calibration[beamElement.Z], beamElement.symbol);
  // ... but enable its calibration entry fields only if we are doing spectrometry not involving TOF
  calibration[beamElement.Z]->setCalibrationFieldsEnabled(tech == TECH_RBS || tech == TECH_ERD
                             || tech == TECH_COINC || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  calibration[0]->setCalibrationFieldsEnabled(true); // but enable fields for TOF

  // add tabs appropriate to technique considering currently used elements
  QList<int> alreadyThere;
  switch(tech) {
  case TECH_IMPLANT :
  case TECH_IMPREC :
  case TECH_ENERDEPTHMAP :
  case TECH_ENERDEPTHLIST :
    // no calibration
    break;
  case TECH_RBS_TOF :
  case TECH_ERD_TOF :
    tabWidgetCalibrations->addTab(calibration[0], "TOF");
    break;
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
  case TECH_COINC :
    tabWidgetCalibrations->addTab(calibration[beamElement.Z], beamElement.symbol);
    break;
  case TECH_ERD :
    // recoils
    for(int ilayer=0; ilayer<layer.count(); ilayer++) {
      if(layer[ilayer]->isUsed()) {
        for(int ielement=0; ielement<layer[ilayer]->element.count(); ielement++) {
          Atom at = layer[ilayer]->element[ielement]->element();
          // if and element is used and has to be simulated and it is not already in the calibration tabs, add it
          if(layer[ilayer]->element[ielement]->numberIsotopesToSimulate()>0 &&
             !alreadyThere.contains(at.Z)) {
            tabWidgetCalibrations->addTab(calibration[at.Z], at.symbol);
            alreadyThere.append(at.Z);
          }
        }
      }
    }
    break;
  default:
    QMessageBox::warning( this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
    break;
  }

  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

}


// ask filename to load parameters from, and load them
void CorteoParamWidget::loadFile() {
  QString filename = QFileDialog::getOpenFileName(this,"Choose a Corteo User Interface parameter file",
    paramfile,"CorteoML file (*.cml)");

  if ( !filename.isEmpty() ) {
    paramfile = filename;
    loadParam( paramfile );
  }

}

// save parameters in current file unless it is not defined
void CorteoParamWidget::saveFile() {
  if(paramfile.isEmpty()) {
    saveFileAs();
    return;
  }
  saveParam( paramfile );
}

// ask filename before saving parameters
void CorteoParamWidget::saveFileAs() {

  QString filename = QFileDialog::getSaveFileName(this,"Choose a filename to save under",
    paramfile,"Corteo user interface parameter file (*.cml)");

  if ( !filename.isEmpty() ) {
    paramfile = filename;
    saveParam( paramfile );
  }
}


// save parameters in default filename
bool CorteoParamWidget::saveParam() {

  return saveParam("CorteoUi.cml");

}


// lets follow the wave and use XML to save our parameters
// (it has the adventages of being structured, human editable and understandable,
//  and we can easily add or remove parameters)
bool CorteoParamWidget::saveParam(QString filename) {

  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

  // create a CorteoML document from the parameters
  QDomDocument doc( "CorteoML" );
  doc.appendChild( XMLNode(doc) );

  // open file 'filename'
  QFile file( filename );
  if( !file.open( QIODevice::WriteOnly ) ) {
    QMessageBox::warning( 0, "Corteo user interface", "Can't open file " + filename );
    return false;
  }

  // wtite the CorteoML document to file
  QTextStream ts( &file );
  ts << doc.toString();

  file.close();
  return true;
}


// open CorteoML XML file to read parameters from
bool CorteoParamWidget::loadParam(QString filename) {

  QDomDocument doc( "CorteoML" );

  QFile file( filename );
  if( !file.open( QIODevice::ReadOnly ) )  {
    QMessageBox::warning( 0, "Corteo user interface", "Can't open parameter file " + filename );
    return false;
  }

  if( !doc.setContent( &file ) ) {
    file.close();
    QMessageBox::warning( 0, "Corteo user interface", "Can't find Corteo XML content in " + filename );
    return false;
  }

  QDomElement root = doc.documentElement();
  if( root.tagName() != "CorteoParam" )  {
    file.close();
    QMessageBox::warning( 0, "Corteo user interface", "Can't find <CorteoParam> tag in " + filename );
    return false;
  }

  setFromNode( root.firstChild() );

  file.close();
  update();

  setLoadedMaterialColors();

  return true;
}


// in case of pixmaps, try to set material color selection box to what was saved in .cml file
// (we do this after updating everything because pixmap file may have changed since .cml
// file was saved and combobox color selection is based on the actual pixmap colors)
void CorteoParamWidget::setLoadedMaterialColors() {
    if(checkBoxUsePixels->isChecked()) {
        for(int ilayer=0; ilayer<layer.count(); ilayer++) {
            if(layer[ilayer]->isUsed()) {
                for(int item=0; item<layer[ilayer]->comboBoxMaterial->count(); item++) {
                    if(layer[ilayer]->loadedMaterialColor == layer[ilayer]->comboBoxMaterial->itemData(item)  ) {
                        layer[ilayer]->comboBoxMaterial->setCurrentIndex(item);
                        break;
                    }
                }
            }
        }
    }
}


// Create an XML node with all the parameters, recursively requesting the same from layers, elements in layers, detectors, etc.
// This can then be saved in a .cml file
QDomElement CorteoParamWidget::XMLNode(QDomDocument &doc) {

  QDomElement root = doc.createElement( "CorteoParam" );

  QDomElement deSimulation = doc.createElement( "simulation" );
  deSimulation.setAttribute( "technique",      comboBoxTechnique->currentText() );
  deSimulation.setAttribute( "number_proj",    lineEditNumberProjectiles->text() );
  deSimulation.setAttribute( "straggling",     stragglingType() );
  deSimulation.setAttribute( "mean_free_path", lineEditMeanFreePath->text() );
  deSimulation.setAttribute( "cone_angle",     lineEditConeAngle->text() );
  deSimulation.setAttribute( "minimum_energy", lineEditMinimumEnergy->text() );
  deSimulation.setAttribute( "cutoff",         lineEditCutoff->text() );
  deSimulation.setAttribute( "screening",      comboBoxScreening->currentIndex() );
  deSimulation.setAttribute( "nthread",        comboBoxNthread->currentIndex() );
  deSimulation.setAttribute( "below_min_angle",checkBoxBelowMinAngle->isChecked()?1:0 );
  deSimulation.setAttribute( "random_seed",    lineEditRandomSeed->text() );
  deSimulation.setAttribute( "write_details",  checkBoxWriteDetails->isChecked()?1:0 );
  deSimulation.setAttribute( "pixels_Angs",    lineEditPixelPerAngs->text() );
  deSimulation.setAttribute( "use_pixels",    checkBoxUsePixels->isChecked()?1:0 );
  deSimulation.setAttribute( "pixel_filename", pixelFilename );

  root.appendChild( deSimulation );

  QDomElement deBeam = doc.createElement( "beam" );
  deBeam.setAttribute( "Z", beamElement.Z );
  deBeam.setAttribute( "selected_mass",   comboBoxBeamMass->currentIndex() );
  deBeam.setAttribute( "charge",          lineEditCharge->text() );
  deBeam.setAttribute( "energy",          lineEditEnergy->text() );
  deBeam.setAttribute( "angle_to_normal", lineEditAngleToNormal->text() );
  deBeam.setAttribute( "azimutal_angle",  lineEditAzimAngle->text() );
  deBeam.setAttribute( "diameter",        lineEditBeamDiameter->text() );
  root.appendChild( deBeam );

  QDomElement deLayers = doc.createElement( "layers" );
  deLayers.setAttribute( "current_layer", tabWidgetLayers->currentIndex() );
  for(int ilayer=0; ilayer<layer.count(); ilayer++) {
    deLayers.appendChild( layer[ilayer]->XMLNode(doc) );
  }
  root.appendChild( deLayers );

  QDomElement deDetectors = doc.createElement( "detectors" );
  deDetectors.setAttribute( "current_detector", tabWidgetDetectors->currentIndex() );
  deDetectors.setAttribute( "virtualWfactor",   lineEditWfactor->text()   );
  deDetectors.setAttribute( "virtualHfactor",   lineEditHfactor->text()   );
  for(int idet=0; idet<detector.count(); idet++)
    deDetectors.appendChild( detector[idet]->XMLNode(doc) );
  deDetectors.appendChild( foil->XMLNode(doc) );
  root.appendChild( deDetectors );

  QDomElement deSpectrum = doc.createElement( "spectrum" );
  deSpectrum.setAttribute( "number_channels", lineEditNumberChannels->text() );
  deSpectrum.setAttribute( "energy_disc",     lineEditEnergyDiffDisc->text() );

  for(int i=0; i<calibration.count(); i++)
    deSpectrum.appendChild( calibration[i]->XMLNode(doc) );

  root.appendChild( deSpectrum );

  return root;

}


// set parameters from a series of XML nodes starting with node n
void CorteoParamWidget::setFromNode( QDomNode n) {

  while( !n.isNull() ) {
    QDomElement e = n.toElement();
    if( !e.isNull() ) {
      if( e.tagName() == "simulation" ) {
        int index = comboBoxTechnique->findText(e.attribute("technique", techniqueName[0]));
        comboBoxTechnique->setCurrentIndex(index>=0?index:0);

        lineEditNumberProjectiles->setText( e.attribute("number_proj",     ""));
        setStragglingType(                  e.attribute("straggling",      "").toInt());
        lineEditMeanFreePath->setText(      e.attribute("mean_free_path",  ""));
        lineEditConeAngle->setText(         e.attribute("cone_angle",      ""));
        lineEditMinimumEnergy->setText(     e.attribute("minimum_energy",  ""));
        lineEditCutoff->setText(            e.attribute("cutoff",          ""));
        comboBoxScreening->setCurrentIndex( e.attribute("screening",       "").toInt() );
        comboBoxNthread->setCurrentIndex(   e.attribute("nthread",         "").toInt() );
        checkBoxBelowMinAngle->setChecked(   e.attribute("below_min_angle", "").toInt()!=0 );
        lineEditRandomSeed->setText(        e.attribute("random_seed",  "8550983"));
        checkBoxWriteDetails->setChecked(   e.attribute("write_details", "").toInt()!=0 );
        lineEditPixelPerAngs->setText(      e.attribute("pixels_Angs",     ""));
        checkBoxUsePixels->setChecked(      e.attribute("use_pixels",      "").toInt()!=0);
        pixelFilename = e.attribute("pixel_filename",   "");

        tabWidgetDetectors->setCurrentIndex(e.attribute("current_detector","").toInt() );

        if(checkBoxUsePixels->isChecked() && (pixelFilename.isEmpty() || loadPixels(pixelFilename) == false) )
                checkBoxUsePixels->setChecked(false);

      } else if( e.tagName() == "beam" ) {

        beamElement = ElementForm::setElement(e.attribute("Z","").toInt(), pushButtonBeam, comboBoxBeamMass);
        comboBoxBeamMass->setCurrentIndex( e.attribute("selected_mass",  "").toInt());
        lineEditCharge->setText(           e.attribute("charge",         ""));
        lineEditEnergy->setText(           e.attribute("energy",         ""));
        lineEditAngleToNormal->setText(    e.attribute("angle_to_normal",""));
        lineEditAzimAngle->setText(        e.attribute("azimutal_angle", ""));
        lineEditBeamDiameter->setText(     e.attribute("diameter",       ""));

      } else if( e.tagName() == "layers" ) {

        QDomNode nn = n.firstChild();
        int ilayer = 0;
        while( !nn.isNull() ) {
          if(ilayer>=layer.count()) {
            QMessageBox::warning( 0, "Corteo user interface",
              "Too many layers in parameter file, only loading first "+QString::number(layer.count()) );
            break;
          } else {
            layer[ilayer]->setFromNode(nn);
          }
          nn = nn.nextSibling();
          ilayer++;
        }
        tabWidgetLayers->setCurrentIndex(e.attribute("current_layer", "").toInt());

      } else if( e.tagName() == "detectors" ) {
        tabWidgetDetectors->setCurrentIndex(e.attribute("current_detector","").toInt() );
        lineEditWfactor->setText(           e.attribute("virtualWfactor", ""));
        lineEditHfactor->setText(           e.attribute("virtualHfactor", ""));

        QDomNode nn = n.firstChild();
        int idet = 0;
        while( !nn.isNull() ) {
          if(nn.toElement().tagName()=="layer") {
            foil->setFromNode(nn);
          } else if(idet<detector.count()) {
            detector[idet]->setFromNode(nn);
            idet++;
          } else {
            QMessageBox::warning( 0, "Corteo user interface", "Too many detectors in parameter file, only loading first "
                                      +QString::number(detector.count()) );
            break;
          }
          nn = nn.nextSibling();
        }
      } else if( e.tagName() == "spectrum" ) {
        lineEditNumberChannels->setText( e.attribute("number_channels", "1000"));
        lineEditEnergyDiffDisc->setText( e.attribute("energy_disc", "1000"));

        // read calibration for each element we have at this point in our 'projectiles' list
        QDomNode nn = n.firstChild();
        while( !nn.isNull() ) {
          QDomElement ee = nn.toElement();
          if( !ee.isNull() && ee.tagName() == "calibration" ) {

            // check if ee corresponds to an element in our list
            int Z = ee.attribute("Z", "").toInt();
            if(Z>=0 && Z < calibration.count()) {
              calibration[Z]->setFromNode(nn);
            }
          }
          nn = nn.nextSibling();
        }
      }
    }
    n = n.nextSibling();
  }
}


// generate 'corteo.bitmap' in the format specific for corteo: byte values representing the material
// set xDim, yDim, zDim values, and indicate which layer is the substrate, or -1 if no material is assigned as substrate
// corresponding to the layer number.
// The data are accessed this way: bitmap[(iz*yDim+iy)*xDim+ix], so x, which represents depth, runs fastest, then y, then z
bool CorteoParamWidget::writeCorteoBitmap(int *xDim, int *yDim, int *zDim, int *substrateLayer) {

    // open 'corteo.bitmap'
    QFile file( "corteo.bitmap" );
    if( !file.open( QIODevice::WriteOnly ) ) {
        QMessageBox::warning( 0, "CorteoUi", "Can't open 'corteo.bitmap'.");
        return false;
    }
/*
    QString ss = "Colors in pixelsColorUsed: ";
    QList<QRgb> ll = pixelsColorUsed.toList();
    for (int rr = 0; rr <ll.count(); rr++)
        ss += QString("%1 %2, ").arg(rr).arg(QString::number(ll.at(rr), 16).toUpper());
    QMessageBox::warning( 0, "CorteoUi", ss);
*/

    QHash<QRgb, int> hash; // this will contain the material number corresponding to each color
    hash.insert( BLACK, 255); // insert black (vacuum) by default, so it is not assigned to a material

    int ilayer = 0; // layer counter, incremented only when layer is used
    bool substrateUsed = false;
    bool convertOK;
    *substrateLayer = -1; // default value: no material assigned as substrate

    for(int i=0; i<layer.count(); i++) {
      if(layer[i]->isUsed()) {
          QRgb rgb = layer[i]->comboBoxMaterial->itemData( layer[i]->comboBoxMaterial->currentIndex() ).toUInt(&convertOK);

          if(!convertOK) {
              QMessageBox::warning( 0, "CorteoUi", QString("Internal error getting color associated with material %1.").arg(i));
              return false;
          }

          //check if color already in hashtable, in which case it has already been assigned to a layer
          if(hash.contains(rgb)) {
              if( rgb == BLACK )
                  QMessageBox::warning( 0, "CorteoUi", QString("Color black, assigned to material %1, is reserved for vacuum.")
                                        .arg(i));
              else
                  QMessageBox::warning( 0, "CorteoUi", QString("%1 assigned to more than one material.")
                                        .arg(layer[i]->comboBoxMaterial->currentText()));
              return false;
          }

          //  layer that is the substrate is treated separately
          if(rgb == SUBSTRATE_COLOR) {
              if(substrateUsed) {
                  QMessageBox::warning( 0, "CorteoUi", QString("More than one material designated as substrate."));
                  return false;
              }
              substrateUsed = true;
              *substrateLayer = ilayer;
          } else {
              hash.insert(rgb, ilayer ); // assign color to layer in hashtable
          }

          ilayer++;
      }
    }

    QTextStream ts( &file );

    *xDim = pixels.at(0).size().height(); // note here that we use 'height' since vertical axis going down is depth
    *yDim = pixels.at(0).size().width(); // while horizontal axis, goin right, is the y axis
    *zDim = pixels.count(); // and z axis (if any) goes from first imagge in stack (scroller left) to last (scroller right)
    // periodic boudary conditions are applied to y and z, even if z dimension is only 1 as for 2D images

    int iz = 0;
    foreach(QImage im, pixels) {
        for(int iy = 0; iy<*yDim; iy++) {
            for(int ix = 0; ix<*xDim; ix++) {
                if( hash.value(im.pixel(iy,ix), -1) == -1) {
                    // color not in hashtable, return the name it should have (Color N) in combobox
                    QMessageBox::warning( 0, "CorteoUi", QString("In image #%1, pixel at (%2, %3) contains color %4, which as not been assigned to any material.")
                                          .arg(iz).arg(iy).arg(ix).arg( pixelsColorUsed.toList().toVector().indexOf(im.pixel(iy,ix))) );
                    return false;
                }
                if( hash.value(im.pixel(iy,ix)) == 255)
                    ts << " ";  // black will be coded as a space (ASCII 32)
                else
                    ts << (char)(hash.value(im.pixel(iy,ix)) + 48);
                // + 48 means that we will use ASCII codes for numbers (easier to read, 0 will show as character "0", but
                // if there are more than 10 layers, ASCII code continues with ": ; < = > ? @ A B C D E ..."
                // (this effectively limits the number of materials to 255-48 = 208, and problems start at ASCII
                // code 127 (DEL) and continue with the various definitions of extended ASCII, after 79 different materials.
            }
            ts << endl; // note the use of an end-of-line caracter, that is not a pixel
        }
        iz++;
    }
    file.close();

    return true;
}


// return beam mass if conversion to double went OK, -1 otherwise
double CorteoParamWidget::beamMass() {
    bool ok;
    double val = comboBoxBeamMass->currentText().toDouble(&ok);

    if(ok)
        return val;
    else
        return -1;
}


// return element mass if layer is used, element is used and conversion to double went OK
// return -1 otherwise
double CorteoParamWidget::elementMass(int ilayer, int ielement) {

    if(isUsedElement(ilayer, ielement)) {
        bool ok;
        double val = layer[ilayer]->element[ielement]->comboBoxElementMass->currentText().toDouble(&ok);
        if(ok) return val;
    }
    return -1;
}


// Generate the content of a corteo input file as a string
// On error, sets 'error' true and returns an error message describing the problem
QString CorteoParamWidget::corteo(bool * error, int tech) {

  *error = false;

  QString scorteo;
  QTextStream ts( &scorteo );

  QString serror = "Can't run a Corteo simulation because the following parameters are not properly set:\n";

  // general parameters: technique, mean free path, cone angle, minimum energy, energy stragling, writeDetails, screening
  if(lineEditMeanFreePath->text().isEmpty() ) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="mean free path";
  }
  if(lineEditConeAngle->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="cone angle";
  }
  if(lineEditMinimumEnergy->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="minimum energy";
  }
  if(lineEditRandomSeed->text().isEmpty() || lineEditRandomSeed->text().toInt()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="random seed";
  }  

  ts << "% technique, mean free path, cone angle [deg], minimum energy [eV], energy stragling, writeDetails, "
        "screening, cutoff angle, random seed, nthreads" << endl;
  ts << tech                                   << " "
     << lineEditMeanFreePath->text()           << " "
     << lineEditConeAngle->text()              << " "
     << lineEditMinimumEnergy->text()          << " "
     << stragglingType()                       << " ";

  // write details
  ts << (checkBoxWriteDetails->isChecked()?1:0);

  // screening
  ts << " " << screening();

  // collision cutoff angle (only for RBS)
  if( isBeamDetectionTechnique() && !lineEditCutoff->text().isEmpty() )
    ts << " " << lineEditCutoff->text();
  else
    ts << " 0";

  //number of parallel threads
  ts << " " <<  lineEditRandomSeed->text().toInt();

  //number of parallel threads
  ts << " " << nthread()  << endl;

  // beam parameters: Z, atomic mass, number of ions, beam energy, beam radius, directional cosines
  if(lineEditNumberProjectiles->text().isEmpty() || lineEditNumberProjectiles->text().toInt()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="number of projectiles";
  }

  if(lineEditEnergy->text().isEmpty() || lineEditEnergy->text().toDouble()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam energy";
  }
  if(lineEditEnergy->text().toDouble()<=lineEditMinimumEnergy->text().toDouble()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam energy smaller or equal to minimum energy";
  }
  if(lineEditBeamDiameter->text().isEmpty() || lineEditBeamDiameter->text().toDouble()<0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam diameter";
  }
  if(lineEditAngleToNormal->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam angle to normal";
  }
  if(lineEditAzimAngle->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam azimutal angle";
  }

  int numproj = lineEditNumberProjectiles->text().toInt();

  ts << "% beam Z, atomic mass, number of ions, beam energy [eV], beam radius [A]" << endl
     << beamElement.Z                      << " "
     << comboBoxBeamMass->currentText()    << " ";
  if(tech>0 && !checkBoxUsePixels->isChecked() ) ts << "0 "; // to avoid confusion, put 0 projectiles here, and detail them for each layer
  else  ts << numproj                      << " "; // for implantation and "Full RBS", set it here
  ts << lineEditEnergy->text().toDouble()  << " "
     << lineEditBeamDiameter->text().toDouble()*0.5e7 << endl;  // divide by to 2 to get radius and mult. by 1e7 to get A

  // directionnal cosines
  double theta = lineEditAngleToNormal->text().toDouble()*PI/180;
  double phy   = lineEditAzimAngle->text().toDouble()*PI/180;
  ts << "% incident beam directional cosines"  << endl
     << cos(theta) << " " << sin(theta)*cos(phy) << " " << sin(theta)*sin(phy) << endl;


  double targetThickness = 0.;
  for(int i=0; i<layer.count(); i++)
    if(layer[i]->isUsed() )
      targetThickness += layer[i]->lineEditThickness->text().toDouble();
  if(targetThickness<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="target has no thickness";
  }

  bool error2;
  QString ss;

  // count number of used layers and print it
  int numberLayers=0;
  for(int i=0; i<layer.count(); i++)
    if(layer[i]->isUsed())
      numberLayers++;

  if(numberLayers==0) {
    // no used layer
    if(*error) serror+= ", ";
    else *error = true;
    serror+="target has no layers";
  }


  if(checkBoxUsePixels->isChecked()){
      if( lineEditPixelPerAngs->text().isEmpty() || lineEditPixelPerAngs->text().toFloat()<=0 ) {
          if(*error) serror+= ", ";
          else *error = true;
          serror+="pixels/Angs";
      }

      int xDim, yDim, zDim, substrateLayer;
      if(writeCorteoBitmap(&xDim, &yDim, &zDim, &substrateLayer)==false) {
          if(*error) serror+= ", ";
          else *error = true;
          serror+="bitmap/colors";
      }
      ts << "% number of layers, number of traveling atoms, x dim, y dim, z dim, "
            "layer designated as substrate, pixels per angstroem" << endl;
      // note the "-" in front of numberLayers; it indicates that the target is a bitmap
      ts << -numberLayers << " " << tabWidgetCalibrations->count()  << " "
         << xDim << " " << yDim << " " << zDim << " "
         << substrateLayer << " " << lineEditPixelPerAngs->text() << endl;
  } else {
      ts << "% number of layers, number of traveling atoms" << endl;
      ts << numberLayers << " " << tabWidgetCalibrations->count()  << endl;
  }

  
  // layers: thickness, density, numElements
  for(int i=0; i<layer.count(); i++) {
    if(layer[i]->isUsed()) {
      ts << "% layer thickness, roughness, atomic density, number of elements, number of projectiles" << endl;
      ss = layer[i]->corteo(QString("layer %1").arg(i), numproj, &error2 );
      if(error2) {
        serror+=ss;
        *error = true;
      } else {
        ts << ss << endl;
      }
    }
  }

  // foil/absorber layer
  if(foil->isUsed()) {
    ss = foil->corteo(QString("foil"), 0,  &error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  } else {
    ts << "0 0 0 1 1 1\n 1 1 1 1 1" << endl; // no absorber, we just need a dummy layer of thickness 0
  }

  // detectors
  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

  // find which detector we are using
  int firstDetector = -1, secondDetector = -1;
  switch(tech) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
  case TECH_ENERDEPTHMAP :
  case TECH_ENERDEPTHLIST :
    break;
  case TECH_FULLRBS :
  case TECH_RBS :
    firstDetector = 0;
    break;
  case TECH_RBS_TOF :
    firstDetector = 0;
    secondDetector = 2;
    break;
  case TECH_ERD :
    firstDetector = 1;
    break;
  case TECH_ERD_TOF :
    firstDetector = 1;
    secondDetector = 3;
    break;
  case TECH_COINC :
    firstDetector = 0;
    secondDetector = 1;
    break;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(tech));
    break;
  }

  // write detector information or, if detector not used, dummy information so corteo is happy
  ts << "% detectors: isEllipse p1x p1y p1z p2x p2y p2z p3x p3y p3z r_annulus" << endl;
  QString dummyDetector = "0  -1 0 0  -1 1 0  -1 1 -1  0";
  if(firstDetector == -1) {
    ts << dummyDetector << endl;
  } else {
    ss = detector[firstDetector]->corteo(&error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  }

  if(secondDetector == -1) {
    ts << dummyDetector << endl;
  } else {
    ss = detector[secondDetector]->corteo(&error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  }

  ts << "% virtual detector enlargement factors: width, height" << endl;
  QString Ws = lineEditWfactor->text().isEmpty()?QString("1"):lineEditWfactor->text(); // default is 1
  QString Hs = lineEditHfactor->text().isEmpty()?QString("1"):lineEditHfactor->text(); // default is 1
  if(Ws.toDouble()<1.0 || Hs.toDouble()<1.0) {
    if(*error) serror += ", ";
    else *error = true;
    serror += "virtual factors can't be smaller than 1";
  }
  ts << Ws << " " << Hs << endl;

  // compound correction of each detected/traveling ion in each layer
  // (first data is Z, then compound corr. in each used layer)
  ts << "% ion Z, and its compound correction in each layer" << endl;
  for(int ielem=0; ielem<tabWidgetCalibrations->count(); ielem++) {
      CalibrationForm *calib = (CalibrationForm *) tabWidgetCalibrations->widget(ielem);
      if(ielem>0) ts << " ";
      ts << calib->getZ();
      for(int i=0; i<layer.count(); i++)
          if(layer[i]->isUsed())
              ts << " " << calib->compoundCorr(i);
  }

  if(*error) return serror;
  return scorteo;

}


// fire up a foil edit dialog
void CorteoParamWidget::editFoil() {
  foilDialog->exec();
}

// return true if layer is used (and within allowed number of layers)
bool CorteoParamWidget::isUsedLayer(int ilayer) {
  if(ilayer<0 || ilayer>=layer.count()) return false;
  return layer[ilayer]->isUsed();
}

// return true if element and layer are used (and within allowed numbers)
bool CorteoParamWidget::isUsedElement(int ilayer, int ielement) {
  if(!isUsedLayer(ilayer)) return false;
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return false;
  return layer[ilayer]->element[ielement]->isUsed();
}

// return true if isUsedElement is true and element is selected for simulation
bool CorteoParamWidget::isSimulatedElement(int ilayer, int ielement) {
  if(isUsedElement(ilayer, ielement) && layer[ilayer]->element[ielement]->checkBoxSimulate->isChecked()) return true;
  return false;
}


// return the number of isotopes to simulate for an element in a layer
// return 0 if isSimulatedElement returns false
// return 1 if checkBoxIsotopes is not checked
// otherwise return the number of isotopes with abundance > 1%
int  CorteoParamWidget::numberIsotopesToSimulate(int ilayer, int ielement) {
  if(!isSimulatedElement(ilayer, ielement)) return 0;
  // the previous condition also checked the validity of ilayer, ielement arguments through a call to isUsedElement
  // so we can safely access this element form without further checking
  return layer[ilayer]->element[ielement]->numberIsotopesToSimulate();

}


// return atom structure corresponfing to element ielement in layer ilayer
// (return null atom if ielement or ilayer are out of bounds)
Atom CorteoParamWidget::element(int ilayer, int ielement) {
  if(ilayer<0   || ilayer>=layer.count()) return Atom();
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return Atom();
  return layer[ilayer]->element[ielement]->element();
}

// fraction of ielement in layer
double CorteoParamWidget::fraction(int ilayer, int ielement) {
  if(ilayer<0   || ilayer>=layer.count()) return 0;
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return 0;

  double sum = 0;
  for(int i=0; i<layer[ilayer]->element.count(); i++)
    sum += layer[ilayer]->element[i]->concentration();

  if(sum>0) return layer[ilayer]->element[ielement]->concentration()/sum;
  return 0;
}

// number of element in layer ilayer
int CorteoParamWidget::numberElement(int ilayer) {
  if(ilayer<0 || ilayer>=layer.count()) return -1;
  return layer[ilayer]->element.count();
}

// number of layers
int CorteoParamWidget::numberLayer() {
  return layer.count();
}

// currently selected technique
int CorteoParamWidget::technique() {
  return techniqueCode[comboBoxTechnique->currentIndex()];
}

// currently selected screening
int CorteoParamWidget::screening() {
  return comboBoxScreening->currentIndex();
}

// currently selected screening
int CorteoParamWidget::nthread() {
  return comboBoxNthread->currentText().toInt();
}


// is the detected particle corresponding to a simulated element is actually the beam?
bool CorteoParamWidget::isBeamDetectionTechnique () {
  switch(technique()) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
  case TECH_ENERDEPTHMAP :
  case TECH_ENERDEPTHLIST :
  case TECH_COINC :
  case TECH_ERD :
  case TECH_ERD_TOF :
    return false;
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
  case TECH_RBS_TOF :
    return true;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(technique()));
    break;
  }
  return false;
}

// is the detected particle corresponding to a simulated element is actually the beam?
bool CorteoParamWidget::isTOF () {
  switch(technique()) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
  case TECH_ENERDEPTHMAP :
  case TECH_ENERDEPTHLIST :
  case TECH_COINC :
  case TECH_ERD :
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
    return false;
  case TECH_ERD_TOF :
  case TECH_RBS_TOF :
    return true;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(technique()));
    break;
  }
  return false;
}

// show image at position z in 'pixels' list
// (the default value of z is -1, in which case z gets the value of horizontalScrollBar)
// if no image, draw a rectangle
// also draw beam and detectors on top of image
void CorteoParamWidget::drawPixels(int z) {

    // convert image to pixmap
    QSize size = labelPixels->size();
    int mx = size.width()/2., my = size.height()/2.; // middle point of viewport
    double radius = sqrt(mx*mx+my*my); // gives an order of magnetude of the size
    bool isPixmap = checkBoxUsePixels->isChecked();

    QPixmap pm(size); // produce a new pixmap
    pm.fill(QWidget::palette().color(labelPixels->backgroundRole())); // fill pixmap with background color of labelPixels

    if(z<0) z = horizontalScrollBar->value();

    QPainter mypainter(&pm);
    // if there is no image loaded yet, z == horizontalScrollBar value == 0 == pixels.count()
    // and the condition below is not satisfied
    if( isPixmap && z < pixels.count() ) {
        // get a pixmap that is a scalled version of pixels[z] so it fits in labelPixels ...
        QPixmap im = QPixmap::fromImage(pixels[z].scaled(size, Qt::KeepAspectRatio, Qt::SmoothTransformation));
        mypainter.drawPixmap( mx-im.width()/2, my-im.height()/2, im); // ... and paste it in the middle of pm
    } else if( !isPixmap ) {
        // normal target: fill the area with black and draw a green rectangle representing a flat target
        pm.fill(Qt::black );
        mypainter.fillRect(0, my*9/10, mx*2, my*5/10, Qt::darkGreen );
        horizontalScrollBar->hide();
    }

    // if beam angles are set, draw beam
    if(!lineEditAngleToNormal->text().isEmpty() &&  !lineEditAzimAngle->text().isEmpty() ) {
        double theta = lineEditAngleToNormal->text().toDouble()*PI/180;
        double phy   = lineEditAzimAngle->text().toDouble()*PI/180;
        double xb = cos(theta) ;
        double yb = sin(theta)*cos(phy);
        // double zb = sin(theta)*sin(phy);

        // x-y projection
        QLineF line(mx, my, mx-radius*yb, my-radius*xb );
        mypainter.setPen(QPen(QBrush(beamColor), 3, Qt::DashLine));
        mypainter.drawLine(line);
/*
        // x-z projection
        QLineF line2(mx*2/3, my, mx*2/3-radius*zb, my-radius*xb );
        mypainter.setPen(QPen(QBrush(beamColor), 3, Qt::DashLine));
        mypainter.drawLine(line2);
*/
    }

    // for the first 2 detectors (RBS & ERD)...
    for(int ii = 0; ii<2; ii++) {

        // check if detector is enabled
        if( tabWidgetDetectors->isTabEnabled(ii)) {
            detector[ii]->updateDetectorPoints();

            // check there are entries in all coordinate fields
            if( !detector[ii]->emptyFields() ) {

                // get detector coordinates
                // (The plane shown by our viewport (labelPixels) is the x,y plane
                // so we need the projection of the detectors coordinates on that plane.
                // Fortunately enough, this happens to be simply the x,y components of the
                // coordinates of each points defining the coordinates of the detector in 3D
                double p1x = detector[ii]->lineEditP1x->text().toDouble();
                double p1y = detector[ii]->lineEditP1y->text().toDouble();
                double p1z = detector[ii]->lineEditP1z->text().toDouble();

                double p2x = detector[ii]->lineEditP2x->text().toDouble();
                double p2y = detector[ii]->lineEditP2y->text().toDouble();
                //double p2z = detector[ii]->lineEditP2z->text().toDouble();

                double p3x = detector[ii]->lineEditP3x->text().toDouble();
                double p3y = detector[ii]->lineEditP3y->text().toDouble();
                double p3z = detector[ii]->lineEditP3z->text().toDouble();

                // empirical scale factor so it fits in viewport
                double scale = radius/sqrt(pow((p1x+p3x)/2.,2)+pow((p1y+p3y)/2.,2)+pow((p1z+p3z)/2.,2))/2;

                // x-y projection
                // draw detector plane itself
                QLineF line1(mx+p1y*scale, my+p1x*scale, mx+p2y*scale, my+p2x*scale );
                QLineF line2(mx+p3y*scale, my+p3x*scale, mx+p2y*scale, my+p2x*scale );
                mypainter.setPen(QPen(QBrush(deteColor), 4));
                mypainter.drawLine(line1);
                mypainter.drawLine(line2);

                // draw the axis line from image center to detector center
                QLineF line3(mx, my, mx+(p1y+p3y)/2.*scale, my+(p1x+p3x)/2.*scale );
                mypainter.setPen(QPen(QBrush(deteColor), 2));
                mypainter.drawLine(line3);
/*
                // x-z projection
                // draw detector plane itself
                QLineF line1b(mx*2/3+p1z*scale, my+p1x*scale, mx*2/3+p2z*scale, my+p2x*scale );
                QLineF line2b(mx*2/3+p3z*scale, my+p3x*scale, mx*2/3+p2z*scale, my+p2x*scale );
                mypainter.setPen(QPen(QBrush(deteColor), 4));
                mypainter.drawLine(line1b);
                mypainter.drawLine(line2b);

                // draw the axis line from image center to detector center
                QLineF line3b(mx*2/3, my, mx*2/3+(p1z+p3z)/2.*scale, my+(p1x+p3x)/2.*scale );
                mypainter.setPen(QPen(QBrush(deteColor), 2));
                mypainter.drawLine(line3b);
*/
            }
        }

    }
    labelPixels->setPixmap(pm);

}

void CorteoParamWidget::loadPixels() {
    QString filename = QFileDialog::getOpenFileName(this,"Select a 2D pixel image or a 3D XRAW file",
          paramfile,"Windows Bitmap (*.bmp);;Graphic Interchange Format (*.gif);;"
                    "Portable Network Graphics (*.png);;MagicaVoxel 3D XRAW format (*.xraw)");

    if ( filename.isEmpty() )
        return;

    loadPixels(filename);

}


// load a 2D image in a format readable by Qt, or a XRAW 3D image as generated by MagicaVoxel
bool CorteoParamWidget::loadPixels(QString filename) {

    MyColors my;
    if(filename.toUpper().endsWith(".XRAW")) {

        // open XRAW file
        QFile file(filename);
        if (!file.open(QIODevice::ReadOnly)) {
            QMessageBox::warning( 0, "Corteo user interface", QString("Unable to open %1").arg(filename));
            return false;
        }

        // read 24-bytes header
        char header[24];
        file.read(header, 24);

        // IMPORTANT NOTE:
        // Axes are not defined the same way in an XRAW file than in corteo:
        //  - when drawing an XRAW file, the user should assume that the beam is comming from the top
        //    i.e. in a direction with a negative z component
        //  - in corteo, the x axis is the depht axis, the incident beam having always a positive x component,
        //    then the y-axis runs from left to right on the display port in CorteoUi, and z-axis comes out of the screen
        // So xDim, yDim, zDim here have noting to do with the same variables in corteo.
        // The drawing on the viewport should remove any ambiguity
        int xDim = header[8] | header[9] << 8 | header[10] << 16 | header[11] << 24;   // turn 4-bytes into an int!
        int yDim = header[12] | header[13] << 8 | header[14] << 16 | header[15] << 24;
        int zDim = header[16] | header[17] << 8 | header[18] << 16 | header[19] << 24;
        int sizePalette = header[20] | header[21] << 8 | header[22] << 16 | header[23] << 24;

        // from   https://voxel.codeplex.com/wikipage?title=XRAW%20Format
        // bytes 0-3 are supposed to contain letters "XRAW" and byte 4 is supposed to be 0, so we use th latter as a
        // string terminator. Byte 5 is supposed to 4 (number of channels: RGBA), byte 6 and 7 are supposed to
        // be 8 (bits per channel, and bits per index). Palette size is supposed to be 256.
        if( QString::compare(QString(header), "XRAW")!=0 || header[5]!=4 ||  header[6]!=8 || header[7]!=8 || sizePalette!=256) {
            QMessageBox::warning( 0, "Corteo user interface", QString("File %1 has an invalid XRAW header.").arg(filename));
            return false;
        }

        //QMessageBox::warning( 0, "Corteo user interface", QString(" %1 %2 %3").arg(xDim).arg(yDim).arg(zDim));

        // pixels are storred x running fasters, then y, then z, so we should have in fact zDim images which size is xDim*yDim
        int i, j, k, imageSize = xDim*yDim*zDim;
        uchar *voxeldata = new uchar[imageSize];
        if( voxeldata==NULL ) {
            QMessageBox::warning( 0, "Corteo user interface", QString("Memory allocation error before reading %1*%2*%3=%4 voxels in %5")
                                  .arg(xDim).arg(yDim).arg(zDim).arg(imageSize).arg(filename));
            return false;
        }
        if( file.read((char *)voxeldata, imageSize) != imageSize ) {
            QMessageBox::warning( 0, "Corteo user interface", QString("Error reading voxels in %1").arg(filename));
            return false;
        }

        uchar palettedata[1024];
        if( file.read((char *)palettedata, 1024) != 1024 ) {
            QMessageBox::warning( 0, "Corteo user interface", QString("Error reading palette in %1").arg(filename));
            return false;
        }

        // generate a palette
        QVector<QRgb> palette;
        for(i=0; i<1024; i+=4)
            palette.append(qRgb(palettedata[i], palettedata[i+1], palettedata[i+2])); // opacity 255 (full opacity) by default

        // now that we have been able to load the image and allocate thus far,
        // lets delete the previous image list before creating the new one
        pixels.clear();
        pixelsColorUsed.clear();
        int indx;

        // store a new list of images in 'pixels'
        for(i=0; i<xDim; i++) {
            uchar *imagedata = new uchar[yDim*zDim];
            for(j=0; j<yDim; j++)
                for(k=0; k<zDim; k++) {
                    indx = (zDim-1-k)*yDim+j;
                    imagedata[indx] = voxeldata[i+(j*xDim)+k*xDim*yDim];
                    pixelsColorUsed << palette.at(imagedata[indx]);
                }

            QImage im( imagedata, yDim, zDim, yDim, QImage::Format_Indexed8 );
            im.setColorTable( palette );

            pixels.append(im);
        }

        delete voxeldata; // not needed anymore since data are now stored on image list 'pixels'

        // set scroll bar length, and set its value to half-way
        horizontalScrollBar->setMaximum(xDim-1);
        horizontalScrollBar->setValue(xDim/2);

    } else {
        // load image and convert it to 8 bits with a color table
        // (each color is a material, exept black which is vacuum)
        QImage image = QImage(filename).convertToFormat(QImage::Format_Indexed8);

        if(image.isNull()) {
            QMessageBox::warning( 0, "Corteo user interface", QString("Unable to load %1").arg(filename));
            return false;
        }

        // remove previous image(s) and replace by the loaded one
        pixels.clear();
        pixels.append(image);

        // get a set of actually used colors in the image
        // (this is needed because the image colorTable usually contains many more colors
        //  than actually used, and in CorteoUi, each used color is a material)
        pixelsColorUsed.clear();
        int w = image.width(), h = image.height();
        for(int i=0; i<w; i++)
            for(int j=0; j<h; j++) {
                if( QColor(image.pixel(i,j)).alpha() != 255 ) {
                    QMessageBox::warning( 0, "Corteo user interface",
                                          QString("%1 contains transparent pixels.").arg(filename));
                    return false;
                }
                pixelsColorUsed << image.pixel(i,j);
            }

        horizontalScrollBar->setMaximum(0);
    }

    // unsing the set of color actually used by all images, check if the number of colors is larger
    // than the number of layers (i.e. the number of different materials in the target)
    if(pixelsColorUsed.count() > numberLayer()) {
        QMessageBox::warning( 0, "Corteo user interface",
                              QString("Number of colors (%1) larger than number of layers (%2). "
                                      "In bitmap mode, a color corresponds to a material. "
                                      "(Black is reserved for vacuum.)\n\n"
                                      "Decrease the number of colors or increase the number of layers.")
                                       .arg(pixelsColorUsed.count()).arg(numberLayer()));

        pixels.clear();
        horizontalScrollBar->setMaximum(0);
        horizontalScrollBar->hide();
        return false;
    }

    if(pixels.count()>1)
        horizontalScrollBar->show(); // show scroll bar where there are many images
    else
        horizontalScrollBar->hide();

    beamColor = my.colors[99].color();
    deteColor = my.colors[100].color();

    //qDebug() << "beamColor=" << beamColor.name() << "deteColor=" << deteColor.name();

    pixelFilename = filename; // filename was successfully loaded, so keep it for saving in cml files
    updateUsePixels();

    return true;

}

// generate a 2D image from an x,y function
// Depth scale has to start at 0, be regularly spaced, and in increasing value order.
// Concentration scale has to be between 0 and 1.
void CorteoParamWidget::genPixels() {

    QString filename = QFileDialog::getOpenFileName(this,"Select depth profile data");
    if ( filename.isEmpty() )
        return;

    FSDataSet * d = new FSDataSet();
    if(d->loadfile(filename)==false) {
        QMessageBox::warning( this, "CorteoUi", QString("Cannot open depth profile '%1'. Stopping here.").arg(filename));
        return;
    }

    // check if data are regularly spaced, ordered in depth and if max concentration <= 1
    // also find the maximum concentration
    double prevDepth, deltaDepth = d->xp[1]-d->xp[0];
    double maxConc = 0;
    int i, j, n = d->getndata();
    QString depthScaleErrorMessage = "Depth scale must start at 0, be evenly spaced, and in increasing value order.";
    if(d->xp[0]!=0 || deltaDepth<=0) {
        QMessageBox::warning( this, "CorteoUi", depthScaleErrorMessage);
        return;
    }

    for(i=0; i<n; i++) {
        if(i>0 && (d->xp[i]<= prevDepth || fabs((d->xp[i]-prevDepth)/deltaDepth)-1.)>0.001 ) {
              QMessageBox::warning( this, "CorteoUi", depthScaleErrorMessage);
              return;
        }
        if(d->yp[i]>1.|| d->yp[i]<0.) {
            QMessageBox::warning( this, "CorteoUi", "Concentration has to be between 0 and 1.");
            return;
        }
        prevDepth = d->xp[i];
        if(d->yp[i] > maxConc) maxConc = d->yp[i];
    }

    if(maxConc==0) {
        QMessageBox::warning( this, "CorteoUi", "Concentration is 0 everywhere!");
        return;
    }

    // ask for image width
    // (the suggested value is such that if maximum concentration is low, at least 10 pixels will be colored)
    bool ok;
    int suggestedWidth = int(10./maxConc+.5);
    if(suggestedWidth>MAX_GEN_PIXMAP_WIDTH) suggestedWidth = MAX_GEN_PIXMAP_WIDTH;
    int width = QInputDialog::getInt(this, "Bitmap width", "Enter desired bitmap width in pixels",
                                     suggestedWidth, 1, MAX_GEN_PIXMAP_WIDTH, 1, &ok);
    if (!ok) return;

    // create an 8-bit-deep bitmap
    QImage image(width, n, QImage::Format_Indexed8);

    // set the first two colors of the 8-bit look-up table
    QRgb value;
    value = qRgb(0, 255, 0);
    image.setColor(0, value);
    value = qRgb(100, 100, 100);
    image.setColor(1, value);

    // fill the image with the first color
    image.fill(0);

    // at each depth, set randomly to the second color a number of pixels proportionnal to the concentration
    for(i=0; i<n; i++) {
        // concentration at depth i multiplied by max random number
        // (so the maximum concentration is now RAND_MAX)
        int concentration = int(d->yp[i]*RAND_MAX+0.5);
        for(j=0; j<width; j++) {
            if(qrand()<=concentration)
                image.setPixel(j, i, 1);
        }
    }

    lineEditPixelPerAngs->setText(QString("%1").arg(1./deltaDepth));

    QString s = QString("CorteoBitmap_") + QDateTime::currentDateTime().toString("yyyyMMddhhmmss")  + ".png";
    image.save(s); // save it (might be useful if user wants to edit it)...
    loadPixels(s); // and load it so post-treatment is done as when loading an image from a file

}


void CorteoParamWidget::showCorteoLog() {

  QFile file("CorteoUi.log");
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text)){
    QMessageBox::warning( 0, "Corteo user interface", "Can't read 'CorteuUi.log'.");
    return;
  }

  QTextStream in(&file);
  QDialog * showLogDialog = new QDialog(this);
  showLogDialog->resize(600,500);

  // create a QTextEdit that contains the content of CorteuUi.log
  // ans whos parent is showLogDialog
  QTextEdit * te = new QTextEdit();

  QHBoxLayout *layout = new QHBoxLayout(showLogDialog);
  layout->addWidget(te);

  te->setLineWrapMode(QTextEdit::NoWrap);
  te->insertPlainText(in.readAll());
  showLogDialog->show();
  te->ensureCursorVisible();  // to make the QTextEdit scroll to the end

  showLogDialog->exec();
  delete showLogDialog;
}

int CorteoParamWidget::numberSimul() {
    int numberSim = 0;
    for(int ilayer=0; ilayer<numberLayer(); ilayer++)
        for(int ielement=0; ielement<numberElement(ilayer); ielement++)
            if(isSimulatedElement(ilayer, ielement))
                numberSim+=numberIsotopesToSimulate(ilayer, ielement);

    return numberSim;

}

void CorteoParamWidget::aboutCorteoUi() {
  QSplashScreen *splash = new QSplashScreen(QPixmap(":/images/splash.png") ) ;
  splash->setWindowFlags(Qt::Tool) ;
  splash->setWindowTitle("About Corteo user interface") ;
  splash->show() ;
}

