/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <QtGui>
#include <QDockWidget>
#include <QMessageBox>

#include "CorteoUi.h"
#include "CScorrInterp.h"
#include "CalibrationForm.h"
#include "CorteoParamWidget.h"
#include "../corteo/techniques.h"
#include "constants.h"
#include "colors.h"

#include "XYQt/FSDoubleValidator.h"

using namespace std;

// Class that contains a pointer to a calibration form
// and a method to compute channel from energy
class Calibration {

    CalibrationForm * calibrationForm;
    double zero;
    double slope;
    double quad;
    double resolution;

public:
    unsigned int maxchannel;

    Calibration();

    bool setCalibrationForm(CalibrationForm * cf, double maxEnergy);
    unsigned int channel(double E, bool *ok = NULL);
    double energy(unsigned int channel);
    double getResolution() {return resolution;}
    double getQuad() {return quad;}
    double getSlope() {return slope;}
    double getZero() {return zero;}
};

Calibration::Calibration() {
    calibrationForm = NULL;
}

bool Calibration::setCalibrationForm(CalibrationForm * cf, double maxEnergy) {
        calibrationForm = cf;
        if(calibrationForm!=NULL) {
            zero = calibrationForm->lineEditZero->text().toDouble();
            slope = calibrationForm->lineEditSlope->text().toDouble();
            quad = calibrationForm->lineEditQuad->text().toDouble();
            resolution = calibrationForm->lineEditResolution->text().toDouble();
            if(slope==0 && quad==0) return false;
            bool ok;
            maxchannel = channel(maxEnergy, &ok);
            return ok;
        }
        zero = 0;
        slope = 0;
        quad = 0;
        resolution = 0;
        maxchannel = 0;

        return false;
}

double Calibration::energy(unsigned int channel) {
    return zero+slope*channel+quad*channel*channel;
}


// compute channel for energy E
unsigned int Calibration::channel(double E, bool *ok) {
    double returnval;
    if(quad==0.0) {
        // linear calibration
        if(slope == 0) {
            if(ok)
                *ok = false;
            return 0;
        } else {
            returnval = (E-zero)/slope;
            if(returnval < 0) {
                if(ok)
                    *ok = false;
                return 0;
            }
            if(ok)
                *ok = true;
            return (unsigned int) returnval;
        }
    }

    // quadratic calibration
    double quad1 = 1/quad; // (already checked that quad!=0)
    double min = - 0.5*slope*quad1;
    double a = E*quad1 + min*min - zero*quad1;
    if(a<0) {
        if(ok)
            *ok = false;
        return 0;
    }
    a = sqrt(a);
    if(quad>0)
        returnval = min + a;
    else
        returnval = min - a;

    if(returnval < 0) {
        if(ok)
            *ok = false;
        return 0;
    }
    if(ok)
        *ok = true;
    return (unsigned int) returnval;
}


CorteoUi::CorteoUi(QWidget *parent, int maxlayer, int maxelement, QSplashScreen * splash)
    : QMainWindow(parent)
{

    qApp->setWindowIcon(QIcon(":/images/Corteo.png") );

    // create graphs and tab containing them
    tab = new QTabWidget(this);
    setCentralWidget(tab); // tab is the central widget

    mainGraph = new FSGraph("corteo.cgr",tab);
    angleGraph = new FSGraph("angle.cgr",tab);
    angleGraph->x->title = QString("Main collision depth [A]");
    angleGraph->y->title = QString("After-main-collision-to-detector angle [deg.]");
    angleGraph->setToolTip("Graph of the angle between the trajectory of the\n"
                           "detected particle at the exit of the main scattering event\n"
                           "and the detector axis as a function of main collision depth.\n"
                           "Useful to figure out the appropriate cone angle value.\n"
                           "Should be determined using a simulation with a large cone angle.\n"
                           "At least a few thousand particles should be detected. Simulation\n"
                           "can be accelerated by using a relatively large mean free path.");

    tab->addTab(mainGraph, "Spectra/profiles");
    tab->addTab(angleGraph, "Angle to detector");
    tab->setTabEnabled (1, false);

    // create a parameter toolbox
    paramUi = new CorteoParamWidget(this, maxlayer, maxelement, splash);

    // initialize progress dialog pointer
    progress = new QProgressDialog(this);
    progress->setCancelButton ( new QPushButton ("Cancel", progress) );
    progress->cancel();

        // toolbox size defines dock size, a bit larger though
    QSize s = paramUi->toolBox->size();
    int w = s.width()+5;
    int h = s.height();
    paramUi->setFixedWidth(w+5);
    paramUi->setMinimumHeight(h);

    // create an unclosable dock widget that holds the parameter toolbox
    QDockWidget *dock = new QDockWidget("Parameters", this);
    dock->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    dock->setWidget(paramUi);
    addDockWidget(Qt::RightDockWidgetArea, dock);
    dock->setFixedWidth(w);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);

    // graph will appear square
    resize(h+w,h);

    connect( paramUi->pushButtonApplyCalib,   SIGNAL(clicked()),   SLOT(applyCalib())      );
    connect( paramUi->pushButtonGo,           SIGNAL(clicked()),   SLOT(go())              );
    connect( paramUi->comboBoxTechnique,      SIGNAL(currentIndexChanged(int)),   SLOT(UpdateConnectCorteoLoadPushbutton()) );
    connect( qApp, SIGNAL(lastWindowClosed()), paramUi,  SLOT(saveParam()));

    connect( paramUi->pushButtonSaveCorteoIn,  SIGNAL(clicked()),   SLOT(writeCorteoIn())  );
    UpdateConnectCorteoLoadPushbutton();

    corteoProcess = new QProcess( this );

}


void CorteoUi::UpdateConnectCorteoLoadPushbutton() {
    paramUi->pushButtonLoadCorteoDetect->disconnect();
    int tech = paramUi->technique();


    switch(tech) {
    case TECH_IMPLANT :
    case TECH_IMPREC :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoOut())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load implantation simulation output\nWarning: assumes current CorteoUI parameters are correct for this file."));
      break;
    case TECH_ENERDEPTHMAP:
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readEnerDepthMap())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load Energy-depth map\nWarning: assumes current CorteoUI parameters are correct for this file."));
      break;
    case TECH_ENERDEPTHLIST :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readEnerDepthList())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load Energy-depth list\nWarning: assumes current CorteoUI parameters are correct for this file."));
      break;
    case TECH_RBS_TOF :
    case TECH_RBS :
    case TECH_ERD_TOF :
    case TECH_ERD :
    case TECH_COINC :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoDetect())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load corteo.detect for normal simulation.\nWarning: assumes current CorteoUI parameters are correct for this file."));
        break;
    case TECH_FULLRBS :
    case TECH_FULLBACK :
        connect( paramUi->pushButtonLoadCorteoDetect,   SIGNAL(clicked()),   SLOT(readCorteoDetectFullRBS())      );
        paramUi->pushButtonLoadCorteoDetect->setToolTip(
                    tr("Load corteo.detect for full simulation.\nWarning: assumes current CorteoUI parameters are correct for this file."));
        break;
    default:
        QMessageBox::warning(this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
        break;
    }

}

// few checks and progress bar setup, then start simulation
bool CorteoUi::go() {

    // check if there is any element to used
    bool isUsedElement = false;
    for(int ilayer=0; ilayer<paramUi->numberLayer() && !isUsedElement; ilayer++)
        for(int ielement=0; ielement<paramUi->numberElement(ilayer) && !isUsedElement; ielement++)
            if(paramUi->isUsedElement(ilayer, ielement))
                isUsedElement = true;

    // any element in the target?
    if(!isUsedElement ) {
        QMessageBox::warning( this, "CorteoUi", "No 'used' element.");
        return false;
    }

    // check incident charge
    if(paramUi->lineEditCharge->text().toDouble()==0) {
        QMessageBox::warning( this, "CorteoUi", "Incident charge is 0.");
        return false;
    }

    // anything to simulate?
    int technique = paramUi->technique();
    if( ( technique>0 || technique == TECH_FULLRBS) &&
            paramUi->numberSimul()==0) {
              QMessageBox::warning( this, "CorteoUi", "No element selected for simulation.\n(Check 'Simulate this element' boxes in layers.)");
              return false;
    }

    // prepare for simulations: disable 'go' button and create a progress bar
    paramUi->pushButtonGo->setEnabled(false);
    progress->setLabelText("Simulation starting...");
    progress->show();

    // same Corteo.cml just in case something crashes
    paramUi->saveParam("lastsim.cml");

    if(technique==TECH_FULLBACK) {
        // run a full RBS simulation, but upon results loading, readFullRBS() will cut-off events for which
        // the largest scattering event is ABOVE cut-off, therefore retaining only background, below cut-off events
        if(!runCorteo(TECH_FULLRBS)) {
            paramUi->pushButtonGo->setEnabled(true);
            return false;
        }
    } else {
        if(!runCorteo(technique)) {
            paramUi->pushButtonGo->setEnabled(true);
            return false;
        }
    }
    return true;
}


void CorteoUi::endSimulations() {

    progress->hide();
    progress->disconnect();
    progress->setLabelText("Simulation ...");

    corteoProcess->disconnect();
    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    if(corteoProcess->state()==QProcess::Running)
        corteoProcess->kill();

    printLog();

    paramUi->pushButtonGo->setEnabled(true);
    QApplication::beep();

}


// write input parameters to corteo.in
bool CorteoUi::writeCorteoIn(int technique) {
    // get the corteo input parameters
    bool error;
    QString s = paramUi->corteo(&error, technique);

    if(error) {
        // something's wrong, show error message contained in corteo input and stop
        QMessageBox::warning(this, "CorteoUi", s);
        return false;
    }

    QFile file( "corteo.in" );
    if( !file.open( QIODevice::WriteOnly ) ) {
        QMessageBox::warning( 0, "CorteoUi", "Can't open 'corteo.in'.");
        return false;
    }
    QTextStream ts( &file );
    ts << s << endl;
    file.close();
    return true;
}

// overload of the previous where the technique is determined from
// the value of the technique combobox
bool CorteoUi::writeCorteoIn() {
    int technique = paramUi->technique();
    if(technique==TECH_FULLBACK)
        technique = TECH_FULLRBS;
    return writeCorteoIn(technique);

}


// run a Corteo simulation
bool CorteoUi::runCorteo(int technique) {

    if(corteoProcess->state()!=QProcess::NotRunning) {
        QMessageBox::warning( 0, "CorteoUi", "Corteo already running.");
        return false;
     }

    if(writeCorteoIn(technique)==false) {
        return false;
    }

    // remove old output files so we are not confused reading old results
    if(technique>0 || technique == TECH_FULLRBS || technique == TECH_ENERDEPTHLIST)
        QFile::remove("corteo.detect");
    else if(technique == TECH_ENERDEPTHMAP)
        QFile::remove("ener_depth_map.out");
    else
        QFile::remove("impl_beam.out");

    // create a process to run corteo    
#ifdef _WINDOWS_
    corteoProcess->start("corteo.exe");
#else
    corteoProcess->start("./corteo");
#endif

    if(corteoProcess->waitForStarted() == false) {
        QMessageBox::warning( 0, "Corteo error", "Corteo could not start.");
        corteoProcess->kill();
        return false;
    }

    // erase content of last output, if any
    corteoOutput  = "";
    corteoError   = "";

    // when corteo outputs something, call progress update
    connect(corteoProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(updateProgress()));

    // when corteo finishes, call the corresponing output read function
    if(technique == TECH_FULLRBS)
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readFullRBS(int)));
    else if(technique==TECH_ENERDEPTHMAP || technique==TECH_ENERDEPTHLIST)
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readEnerDepthOut(int)));
    else if(technique<0)
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readImplOutput(int)));
    else
        connect(corteoProcess, SIGNAL(finished(int ,QProcess::ExitStatus)),
                this, SLOT(readIBAOutput(int)));

    connect(progress, SIGNAL(canceled()), this, SLOT(endSimulations()));
    progress->exec();

    return true;
}


// slot called whenever some output comes out of corteo
void CorteoUi::updateProgress() {

    // append output
    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    // figure out the number of threads and simulations
    int i = corteoOutput.indexOf("thread");
    int numthreads = i>0?corteoOutput.left(i).toInt():0;

    i = corteoOutput.indexOf("simulation");
    int numSimul = i>0?corteoOutput.mid(i-4,3).toInt():0;

    // number of "_" tels us the progress of the simulation
    int count = corteoOutput.count('_');

    // at the end of the simulation, there are 10 "_" per simulation and per threads
    if(progress->isVisible()) {
        if(numthreads) {
            progress->setRange(0, 10*numSimul*numthreads);
            progress->setValue(count);
            progress->setLabelText(QString("Simulation %1 of %2").arg(int(count/10/numthreads)+1).arg(numSimul));
        } else {
            progress->setLabelText("Can't show simulation progress, number of threads apparently 0.");
        }
    }

}

void CorteoUi::readImplOutput(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","There were warnings or errors, please check log.");
        return;
    }

    readCorteoOut();
    QApplication::beep();

}

// read output related to implantation simulation
bool CorteoUi::readCorteoOut() {

    MyColors my;

    // when also loading displacement profile, let make implantation profile thicker
    double linewidth = paramUi->technique()==TECH_IMPREC?2:1;

    // simulation successful: create a dataset, load resulting profile into it, and add it to the graph
    FSDataSet * d = new FSDataSet( paramUi->beam().symbol, QPen(my.colors[paramUi->beam().Z], linewidth) );
    if(d->loadfile("impl_beam.out",1,2,1)==false) { // read column 1 (depth) as x, column 2 as y, skipping line 1
        QMessageBox::warning( this, "CorteoUi", "Can't open depth profile 'impl_beam.out'. Stopping here.");
        return false;
    }

    double qq = paramUi->lineEditCharge->text().toDouble();

    for(int i=0; i<d->getndata(); i++)
        d->yp[i]*=1e8*qq ; // at/A/ion to [at/cm3]/[at/cm2] * [at/cm2]
    mainGraph->appendDataSet(d);
    tab->setTabEnabled (1, false);

    // in case of full cascade, also load displacements profiles
    // (but Corteo generates only one file per element, even if present in many layer, so load them only once)
    if(paramUi->technique()==TECH_IMPREC) {
        QList<int> Z;
        // loop through all layers and elements
        for(int ilayer=0; ilayer<paramUi->numberLayer(); ilayer++) {
            for(int ielement=0; ielement<paramUi->numberElement(ilayer); ielement++) {
                // if element is used
                if(paramUi->isUsedElement(ilayer, ielement)) {
                    Atom a = paramUi->element(ilayer, ielement);
                    if(!Z.contains(a.Z)) {
                        FSDataSet * d = new FSDataSet( QString("%1 disp").arg(a.symbol), QPen(my.colors[a.Z], 1) );
                        QString filename = QString("impl%1.out").arg(a.Z,3,10,QChar('0'));
                        if(d->loadfile(filename,1,2,1)==false) { // read column (energy) 1 as x, column 2 as y, skipping line 1
                            QMessageBox::warning( this, "CorteoUi", QString("Can't open file %1. Stopping here.").arg(filename));
                            return false;
                        }
                        for(int i=0; i<d->getndata(); i++)
                            d->yp[i]*=1e8 * qq; // at/A/ion to [at/cm3]/[at/cm2] * [at/cm2] * [at/cm2]
                        mainGraph->appendDataSet(d);
                        Z.append(a.Z);
                    }
                }
            }
        }
    }

    if(qq==1) mainGraph->y->title = QString("Concentration/fluence [at/cm^{3}]/[at/cm^{2}]");
    else mainGraph->y->title = QString("Concentration/fluence [at/cm^{3}]");
    mainGraph->x->title = QString("Depth [A]");
    return true;

}



// read simulation screen output then call readEnerDepthMap() or readEnerDepthList()
void CorteoUi::readEnerDepthOut(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","Corteo simulation ended with error. Please check log.");
        return;
    }

    switch (paramUi->technique()) {
    case TECH_ENERDEPTHLIST:
        readEnerDepthList();
        break;

    case TECH_ENERDEPTHMAP:
        readEnerDepthMap();
        break;

    default:
        QMessageBox::warning(this,"Corteo","Wrong function call for selected technique.");
        return;

    }

}

void CorteoUi::readEnerDepthList() {
    FSDataSet * d = new FSDataSet("Energy-Depth data", QPen(QColor("black"), 0), 12);
    if(d->loadfile("corteo.detect", 2, 1)) {
        mainGraph->appendDataSet(d);
        mainGraph->x->title = QString("Depth [Angs]");
        mainGraph->y->title = QString("Energy [eV]");
    }
}

// read energy-depth map resulting from simulation, make a PNG out of it, and show it.
void CorteoUi::readEnerDepthMap() {

    // simulation successful: produce a histogram as a new dataset, and add it to the graph
    QFile file("ener_depth_map.out");
    if(!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'ener_depth_map.out'.");
        return;
    }

    QTextStream in(&file);
    int nlines = 1;
    QString line = in.readLine();

    while (!line[0].isNumber()) {
        // comment line begin with any symbol not involve in number representation
        line = in.readLine();
        nlines++;
    }

    bool ok;
    QStringList param = line.simplified().split(" ");
    if(param.count()!=3) {
        QMessageBox::warning( this, "CorteoUi", "Expected first data line to contain 3 numbers in 'ener_depth_map.out': dimension, E0, target thickness");
        return;
    }

    int n = param[0].toInt(&ok);
    if(!ok || n<0 || n>1000000 ) {
        QMessageBox::warning( this, "CorteoUi", "Apperent error with number representing lateral dimension when reading 'ener_depth_map.out'.");
        return;
    }

    QImage im(n, n, QImage::Format_RGB32);
    im.fill(Qt::white);
    for(int i=0; i<n; i++) {
        QString line = in.readLine();
        QStringList val = line.simplified().split(" ");
        nlines++;

        if(val.count()!=n) {
            QMessageBox::warning( this, "CorteoUi", QString("At line %1 of 'ener_depth_map.out', expected to read %2 values but got %3.")
                                  .arg(nlines).arg(val.count()).arg(n));
            return;
        }
        for(int j=0; j<n; j++) {
            unsigned int v = val[j].toInt(&ok);
            if(v>0) im.setPixel(i,j,v);
            if(!ok) {
                QMessageBox::warning( this, "CorteoUi", "Error converting %1th value at line %2 of 'ener_depth_map.out'.");
                return;
            }
        }
    }

    QString filename = QString("energy_depth_map_%1").arg(paramUi->beam().symbol);
    double E0 = param[1].toDouble(&ok);
    if(ok) {
        if(E0<1e3)
            filename += QString("_%1eV").arg(E0);
        else if(E0<1e6)
            filename += QString("_%1keV").arg(E0/1e3);
        else
            filename += QString("_%1MeV").arg(E0/1e6);
    } else {
        filename += QString("_") + param[1];
    }
    double thick = param[2].toDouble(&ok);
    if(ok) {
        if(thick<1e4)
            filename += QString("_%1nm").arg(thick/10);
        else if(thick<1e7)
            filename += QString("_%1um").arg(thick/1e4);
        else
            filename += QString("_%1mm").arg(thick/1e7);
    } else {
        filename += QString("_") + param[2];
    }

    filename += QString(".png");

    if(im.save(filename))
        QMessageBox::information( this, "CorteoUi", QString("Map saved in '%1'.").arg(filename));
    else
        QMessageBox::warning( this, "CorteoUi", QString("Unable to save '%1'.").arg(filename));

}

void CorteoUi::readFullRBS(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","Corteo simulation ended with error. Please check log.");
        return;
    }

    readCorteoDetectFullRBS();
    QApplication::beep();

}


bool CorteoUi::readCorteoDetectFullRBS() {

    // simulation successful: produce a histogram as a new dataset, and add it to the graph
    QFile file("corteo.detect");
    if(!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'corteo.detect'.");
        return false;
    }
    bool ok;
    int numproj = paramUi->lineEditNumberProjectiles->text().toInt(&ok);
    if(!ok || numproj<=0) {
        QMessageBox::warning( this, "CorteoUi", "Please set the number of projectiles properly.");
        return false;
    }

    // each detected particle in full RBS mode contributes the indicated incident charge (in atoms)
    // divided by the number of simulated particles
    double qq = paramUi->lineEditCharge->text().toDouble()/ELEMENTARY_CHARGE/numproj;

    int Z1 = paramUi->beam().Z;
    int Zdet = Z1;
    int technique = paramUi->technique();

    double cosCutOffAngle = paramUi->lineEditCutoff->text().toDouble(&ok);
    if(paramUi->lineEditCutoff->text().isEmpty() || !ok || cosCutOffAngle<=0 || cosCutOffAngle >= 180) {
        QMessageBox::warning( this, "CorteoUi", "Problem with cut-off angle, must be set above 0 and below 180 degrees");
        return false;
    }
    cosCutOffAngle = cos(cosCutOffAngle*PI/180.);

    PeriodicTable pt;
    MyColors my;

    // get calibration for beam
    Calibration calib;
    if(calib.setCalibrationForm(
                paramUi->getCalibrationForm(Zdet),
                paramUi->lineEditEnergy->text().toDouble()) == false) {
        QMessageBox::warning(this,
                             "CorteoUi",
                             QString("Calibration of element %1 not set properly. Correct and reload data.")
                             .arg(pt.returnAtom(Z1).symbol) );
        return false;
    }

    // create and initialize a dataset in which the histogram of detected particles will be stored
    FSDataSet * d = new FSDataSet( "Full RBS", QPen(my.colors[mainGraph->getndatasets()+1], 1, Qt::DashLine) );
    for(unsigned int iE=0; iE<calib.maxchannel; iE++)
        d->append( calib.energy(iE), 0);

    FSDataSet * angleData = new FSDataSet("max scat. angle (Full RBS)", QPen(), 11, 5, QPen(QColor("magenta")));
    angleData->linePen.setWidth(0); // no line, only dots

    QTextStream in(&file);
    int conversionProblems = 0;
    while(!in.atEnd()) {
        QString line = in.readLine();
        if(line[0].isNumber()) {  // comment line begin with any symbol not involve in number representation

            QStringList val = line.simplified().split(" ");

            // increment histogram (yp)
            bool ok, ok2;
            double E = val[0].toDouble(&ok); // sometimes, we get things such as 'nan', so lets validate the conversion
            if(ok && (isnan(E) || isinf(E))) ok = false; // guess what: in C/C++, inf, -inf and nan are valid float values!
            unsigned int ch;
            if(ok) {
                ch = calib.channel(E, &ok); // get channel for E (column 0)
                double cosAngle = val[1].toDouble(&ok2);
                if(ok2 && (isnan(cosAngle) || isinf(cosAngle)))
                    ok2 = false;

                if(ok && (technique != TECH_FULLBACK || (technique == TECH_FULLBACK && ok2 && cosAngle>cosCutOffAngle)))
                    // incremeted if not background simulation or if it is and angle is below cut-off angle (i.e. cos larger)
                    d->yp[ch] += qq;
            }

            if(!ok)
                conversionProblems++;


            // accumulate scattering angle data
            if(paramUi->checkBoxGraphAngle->isChecked()  && val.count()>1)
                angleData->append(val[0].toDouble(),acos(val[1].toDouble())*180/PI);
        }
    }
    if(conversionProblems)
        QMessageBox::warning(this, "CorteoUi",
                         QString("%1 lines in corteo.detect contained values not convertible to numbers and were ignored.").arg(conversionProblems) );

    d->gaussConvol( calib.getResolution()/calib.getSlope() );
    mainGraph->appendDataSet(d);

    if(paramUi->lineEditCharge->text().toDouble()==1) {
        // charge set to 1 coulomb, normalized spectrum
        mainGraph->y->title = QString("counts/coulomb");
    } else {
        // charge set to some value
        mainGraph->y->title = QString("counts");
    }
    mainGraph->x->title = QString("Energy [eV]");


    if(angleGraph->getndatasets()>0) {
        tab-> setCurrentIndex(1);
        if(QMessageBox::question(this, "CorteoUi", "This graph contains data from previous simulation. Clear them?",
                                 QMessageBox::Yes, QMessageBox::No )==QMessageBox::Yes) {
            angleGraph->clearDataSets();
        }
    }

    addAngleData(angleData);
    return true;

}


void CorteoUi::readIBAOutput(int exitStatus) {

    corteoOutput  += corteoProcess->readAllStandardOutput();
    corteoError   += corteoProcess->readAllStandardError();
    // QMessageBox::warning( 0, "CorteoUi", corteoOutput);

    endSimulations();

    if(exitStatus>0) {
        QMessageBox::warning(this,"Corteo","There were warnings or errors, please check log.");
        return;
    }

    readCorteoDetect();
    QApplication::beep();

}


// Read the corteo engin output file corteo.detect for IBA simulation involving a main collision (i.e. exclusing namely Full RBS)
// Return false if:
// - file cannot be oppened
// - it contains too few columns
// - calibration is not set for one of the detected elements
// - ther is an error in one of the values read (e.g. +/-infinity, NAN)
// - output looks empty (i.e. no spectra generated after reading the file)
// Return true otherwise
bool CorteoUi::readCorteoDetect() {

    MyColors my;
    QFile file("corteo.detect");
    if(!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'corteo.detect'.");
        return false;
    }
    int technique = paramUi->technique();

    bool doLoadEachZ = paramUi->checkBoxLoadSeparateSpectra->isChecked();
    bool isBeam = paramUi->isBeamDetectionTechnique();   // true if the spectrum is made of scattered beam atoms
    bool isTOF = paramUi->isTOF();
    double q = paramUi->lineEditCharge->text().toDouble()/ELEMENTARY_CHARGE; // number of incident atoms
    int Z1 = paramUi->beam().Z, Z2;

    // list of Z of elements already simulated and there corresponding dataset and calibration
    QMap<int, Calibration *> ZCalibration;  // map of Calibrations for each Z
    QMap<int, FSDataSet *> ZDataSet;  // map of datasets for each Z
    QMap<QString, CScorrInterp *> CScorrInterpSet;  // map of CScorrInterp for each CS stamp

    PeriodicTable pt;

    FSDataSet * angleData = new FSDataSet("angle to det. at coll.", QPen(), 2, 2, QPen(QColor("blue")));
    angleData->linePen.setWidth(0);

    int returnInterpValAccept = paramUi->checkBoxBelowMinAngle->isChecked() ? -2 : -1; // return value above which an event is accepted when interp() fails.
    // interp() returns 0 if interpolation OK, -1 if energy was below minimum of table (accept anyway the unmodified, (screened) Rutherford cross-section),
    // -2 if angle was below minimum of table but the user can decide to accept these events (whatever the energy is) by checking checkBoxBelowMinAngle
    // The interp() function can also return -3 if energy is above max, but then it is clearly unsafe to accept the value.

    QTextStream in(&file);
    bool ok;
    int conversionProblems = 0, CStableInterpProblem = 0;
    double cosCutoff = paramUi->lineEditCutoff->text().toDouble(&ok);
    if(ok)
        cosCutoff = cos(cosCutoff*PI/180.);
    else
        cosCutoff = 0;

    while(!in.atEnd()) {
        QString line = in.readLine();
        if(line[0].isNumber()) {  // comment line begins with any symbol not involve in number representation

            // in case of coincidence,the first value is the energy of the 2nd detected particle
            // in case of RBS-TOF or ERD-TOF, the first value is the time of flight (TOF) in ns
            int i0 = (technique==TECH_COINC || technique==TECH_RBS_TOF || technique==TECH_ERD_TOF) ? 1 : 0;

            QStringList val = line.simplified().split(" ");
            if( val.count()<10+i0 ) {
                QMessageBox::warning(this,"Corteo",QString("Too few columns in 'corteo.detect' file. Expecting at least %1 columns.").arg(10+i0));
                return false;
            }

 //values in corteo.detect: [E_2nd_part_coincidence] E crosssect Nt_Omega/Num_det_part E_before_coll cos_coll_angle cos_angle_det0 colldepth Z2 collision_layer CS_stamp

            Z2 = val[ 7+i0 ].toInt(&ok); // histograms generated in different datasets in ZDataSet (index Z2)
            int Zdet;
            double maxEnergy = paramUi->lineEditEnergy->text().toDouble();
            if(ok) {
                Zdet = isBeam ? Z1 : Z2;  // detected particle is beam or recoil
                if(isTOF) {
                    // In the case of TOF, we need a maximum time t_max = sqrt(0.5 m L^2/E_min) where L is the camera lenght and
                    // E_min is the minimum energy to detect. Lets assume that E_min is E_beam/100, that L=1m, and that m = 2*Zdet*1amu
                    // t_max = sqrt(0.5*2 * Zdet*1.66E-27 kg (1 m)^2/(E_beam[eV]*1.6E-19/100)) seconds =  sqrt(Zdet/E_beam[eV])*1E-3 seconds
                    // (since 1.66/1.602 is about 1) so t_max = sqrt(Zdet/E_beam[eV])*1E6 nanoseconds
                    maxEnergy = 1e6*sqrt(Zdet/maxEnergy); // now maxEnergy is a maximum time
                    // the sqrt() factor gives sqrt(2)/1000 for 1 MeV/amu atoms that indeed have m = 2Z so that's tmax = 707 ns max

                    Zdet = 0; // if TOF, use TOF calibration (form with index 0)
                }

                if(ZCalibration.contains(Zdet)==false) {
                    ZCalibration.insert(Zdet, new Calibration());
                    if(ZCalibration[Zdet]->setCalibrationForm(
                                paramUi->getCalibrationForm(Zdet),
                                maxEnergy) == false) {
                        if(isTOF){
                          QMessageBox::warning(this,
                               "CorteoUi",
                               QString("TOF calibration not set properly. \nCorrect and reload data. \n(No need to run the simulation again.)")   );
                        } else {
                            QMessageBox::warning(this,
                               "CorteoUi",
                               QString("Calibration of element %1 not set properly. \nCorrect and reload data. \n(No need to run the simulation again.)")
                               .arg(pt.returnAtom(Zdet).symbol) );
                        }
                        return false;
                    }
                }
                if(ZDataSet.contains(Z2)==false) {
                    // create dataset to store histogram and initialize it to 0
                    ZDataSet.insert(Z2, new FSDataSet(pt.returnAtom(Z2).symbol, QPen(my.colors[Z2], 1)) );
                    for(unsigned int iE=0; iE<ZCalibration[Zdet]->maxchannel; iE++)
                        ZDataSet[Z2]->append( ZCalibration[Zdet]->energy(iE), 0);
                }
            }

            double EE = 0, E=0, CS=0, F=0;
            if(i0 && ok) EE = val[0].toDouble(&ok); // 2nd detected particle energy in case of coincidence, TOF in the case of RBS-TOF or ERD-TOF
            if(ok) E  = val[0+i0].toDouble(&ok); // detected particle energy
            if(ok) CS = val[1+i0].toDouble(&ok); // (possibly screened) Rutherford cross-section
            if(ok) F  = val[2+i0].toDouble(&ok); // Nt * Omega / Number of detectable particles (see doc)

            // check if we got inf or nan
            if(isnan(E) || isnan(EE) || isnan(CS) || isnan(F) || // guess what: in C/C++, inf, -inf and nan are valid float values! So we need to test...
               isinf(E) || isinf(EE) || isinf(CS) || isinf(F))  // ... and isnormal(x) returns false for x=0... come on...
                ok = false;

            double angle, Elab = 0;
            if(ok) Elab = val[ 3+i0 ].toDouble(&ok); // Energy just before collision, lab frame
            if(ok) angle = val[ 4+i0 ].toDouble(&ok); // cos(collision angle)

            if(!isTOF && ok && angle > cosCutoff ) ok = false; // in RBS, check if the main collision had an angle smaller than the specified cutoff, i.e. cos(angle) > cos(cutoff angle)

            if(ok) {
                angle = acos(angle)*180./PI;
                if(isinf(angle) || isnan(angle)) ok = false;
            }
            if(ok) {

                // cross section correction
                QString CSstamp = val[ 9+i0 ]; // cross-section stamp
                // The CS stamp is simply 0 if there is no correction
                // Otherwise, it is in the format AB.MMMZZmmmzz where A is the simulation type, B is the correction type (1 for Ratio to Rutherfor, 2 for value from table)
                // MMM is the beam mass, ZZ is the beam atomic number, mmmm is the recoil mass, zz is the recoil atomic number
                // (The stamp has been gererated in this format so it looks like a number and there are no problems reading Corteo output with Python, Matlab or other software)
                if(CSstamp != "0") {
                    if( CScorrInterpSet.contains(CSstamp)==false) {
                        // CScorrInterp for this interaction does not exist in CScorrInterpSet yet, so add it
                        CScorrInterp * CScorr = new CScorrInterp(CSstamp);
                        if(!CScorr->isDataLoaded()) {
                            QMessageBox::warning(this,
                                                 "Corteo User Interface",
                                                 QString("Could not create cross-section interpolation class due to the following error :\n%1.")
                                                     .arg(CScorr->errorMessage()));
                            return false;
                        }
                        CScorrInterpSet.insert(CSstamp, CScorr);
                    }
                    if(ok) {
                        if( CScorrInterpSet[CSstamp]->interp(Elab, angle, &CS) < returnInterpValAccept ) {
                            CStableInterpProblem++;
                            ok = false;
                        }
                    }
                }
            }

            // increment histogram (yp) by cross-section * Nt * Omega / Number of detectable particles
            unsigned int ch;
            if(ok) {
                switch(technique) {
                case TECH_RBS_TOF :
                case TECH_ERD_TOF :
                    ch = ZCalibration[Zdet]->channel(EE, &ok); // get channel for time of flight (extra column 0)
                    break;
                case TECH_RBS :
                case TECH_ERD :
                    ch = ZCalibration[Zdet]->channel(E, &ok); // get channel for energy (column 0)
                    break;
                case TECH_COINC :
                    ch = ZCalibration[Zdet]->channel(E+EE, &ok); // get channel for E+EE (extra columns 0 + 1)
                    break;
                default:
                    QMessageBox::warning(this, "Corteo user interface", QString("Technique code %1 undefined when loading 'corteo.detect'.").arg(technique));
                    return false;
                }
                if(ok) ZDataSet[Z2]->yp[ch] += q * CS * F;
            }

            // accumulate angle-with-detector data
            if(ok && paramUi->checkBoxGraphAngle->isChecked()) {
                double colldepth, detangle;
                detangle = val[ 5+i0 ].toDouble(&ok); // cos(angle of particle exiting collision wrt detector)
                if(ok) detangle = acos(detangle)*180./PI;
                if(isinf(detangle) || isnan(detangle)) ok = false;

                if(ok) colldepth = val[ 6+i0 ].toDouble(&ok);

                if(ok) angleData->append(colldepth, detangle);
            }

            if(!ok) conversionProblems++;
        }
    }

    if(CStableInterpProblem)
        QMessageBox::warning(this, "CorteoUi",
                         QString("%1 interpolations in cross-section table went wrong; events were ignored.").arg(CStableInterpProblem) );

    if(conversionProblems)
        QMessageBox::warning(this, "CorteoUi",
                         QString("%1 line(s) in corteo.detect contain value(s) not convertible to numbers and were ignored.").arg(conversionProblems) );

    if(ZDataSet.empty()) {
        QMessageBox::warning(this,"Corteo","Corteo output looks empty!");
        return false;
    }

    // add datasets to graph
    QMapIterator<int, FSDataSet *> i(ZDataSet); // iterator for FSDataSets in ZDataSet

    // convolution of first spectrum for resolution (i.key() = Z2)
    i.next();
    Z2 = i.key();
    Calibration * cal = ZCalibration.value(isTOF? 0 : (isBeam ? Z1 : Z2) );
    i.value()->gaussConvol( cal->getResolution()/cal->getSlope() );

    // in the case of RBS, produce a total spectrum
    FSDataSet * sumSpectrum = NULL;
    if(isBeam) {
        // create a new dataset to store the sum, as a copie of the first dataset in ZDataSet
        sumSpectrum = new FSDataSet(i.value());
        sumSpectrum->legend = paramUi->beam().symbol; // legend = beam name
        sumSpectrum->linePen = QPen(my.colors[mainGraph->getndatasets()+1], doLoadEachZ ? 2:1 ); // thicker line of sumSpectrum is not alone
    }

    if(doLoadEachZ || !isBeam)
        mainGraph->appendDataSet(i.value());

    while (i.hasNext()) {
        i.next();
        Z2 = i.key();

        // convolution of spectrum for resolution
        cal = ZCalibration.value(isTOF ? 0 : (isBeam ? Z1 : Z2) );
        i.value()->gaussConvol( cal->getResolution()/cal->getSlope() );

        // for RBS, add data to total spectrum
        if(isBeam)
            for (int j = 0; j<sumSpectrum->getndata(); j++)
                sumSpectrum->yp[j] += i.value()->yp[j];

        // add to graph or free memory
        if(doLoadEachZ || !isBeam)
            mainGraph->appendDataSet(i.value());

    }
    // for RBS, add total spectrum to graph
    if(isBeam)
        mainGraph->appendDataSet(sumSpectrum);

    if(q==1) {
        // charge not set, normalized spectrum
        mainGraph->y->title = QString("counts/coulomb");
    } else {
        // charge set to some value
        mainGraph->y->title = QString("counts");
    }
    if(isTOF) mainGraph->x->title = QString("TOF [ns]");
    else mainGraph->x->title = QString("Energy [eV]");

    if(paramUi->checkBoxGraphAngle->isChecked() && angleData->getndata()>1 ) {
        if(angleGraph->getndatasets()>0) {
            tab-> setCurrentIndex(1);
            if(QMessageBox::question(this, "CorteoUi", "This graph contains data from previous simulation. Clear them?",
                                     QMessageBox::Yes, QMessageBox::No )==QMessageBox::Yes) {
                angleGraph->clearDataSets();
            }
        }
        addAngleData(angleData);
    }

    return true;

}



// if user requested an angle-to-detector graph, append dataset to the graph and show it
// also show a line delimiting 99% of the largest angles
void CorteoUi::addAngleData(FSDataSet * angleData) {

    if(paramUi->checkBoxGraphAngle->isChecked()) {

        // append angle datasets
        angleGraph->appendDataSet(angleData);

        // generate a curve delimiting the 99th percentile of angles to detector
        // we will generate 10 such points, each for 1/10th of the data
        // as we need the 99th percentile 10 times, we need at least 1000 points
        angleData->sort(); // in depth order
        int n = angleData->getndata()/10;
        FSDataSet * percentile99 = new FSDataSet(QString("%1 99%").arg(angleData->getname()),
                                                 QPen(QColor("red"), 2));
        if(n>100) {
            FSDataSet * temp = new FSDataSet();
            temp->grow(n);
            double xmoy = 0;
            for(int j=0; j<10; j++) {
                temp->clear();
                // create a dataset containing a subset (1/10) of the points
                for(int k=j*n; k<(j+1)*n; k++) {
                    temp->append(angleData->xp[k],angleData->yp[k]);
                    xmoy += angleData->xp[k];
                }
                xmoy /= n;

                // swap x and y so we can use the sort function on y data
                temp->swap();
                temp->sort();

                // add the point that is the 99th/100 of the sorted data
                percentile99->append(xmoy,temp->xp[int(temp->getndata()*0.99+.5)]);
            }
            angleGraph->appendDataSet(percentile99);
        }

        tab->setTabEnabled(1, true);
    } else {
        // no angle data, disable tab
        if(angleGraph->getndatasets()==0) {
            tab-> setCurrentIndex(0);
            tab->setTabEnabled(1, false);
        }
    }

}

// Apply currently displayed calibration to currently selected dataset
// (so data loaded in channels can be converted to energy)
bool CorteoUi::applyCalib() {
    FSDataSet * s = mainGraph->selectedDataSet;
    if(s==NULL) return false; // no currently selected dataset

    // get current calibration
    CalibrationForm * currentCal = (CalibrationForm *)paramUi->tabWidgetCalibrations->widget(paramUi->tabWidgetCalibrations->currentIndex());
    double zero =  currentCal->lineEditZero->text().toDouble();
    double slope = currentCal->lineEditSlope->text().toDouble();
    double quad =  currentCal->lineEditQuad->text().toDouble();

    if(slope==0 && quad==0) return false; // applying the calibration with only zero defined would put all channels to zero

    int i, n = s->getndata();
    for(i=0;i<n; i++) s->xp[i] = zero + i*slope + i*quad*i;

    mainGraph->updateGraph();
    return true;
}


bool CorteoUi::printLog() {
    // open log file
    QFile logfile("CorteoUi.log");
    if (!logfile.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
        QMessageBox::warning( this, "CorteoUi", "Can't open 'CorteoUi.log'.");
        return false;
    }

    QTextStream logstream(&logfile);
    QDateTime dt = QDateTime::currentDateTime();
    logstream << endl << endl << "======= " << dt.toString("yyyy/MM/dd hh:mm:ss") << " =======" << endl;
    logstream << corteoOutput << endl;
    if(!corteoError.isEmpty()) {
        logstream << "--- Warning and Error(s) ---" << endl;
        logstream << corteoError << endl ;
    }
    logstream << endl;
    logfile.close();

    return true;
}



