
/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <QMessageBox>
#include <math.h>

#include "XYQt/FSDoubleValidator.h"
#include "DetectorForm.h"
#include "constants.h"

DetectorForm::DetectorForm(QString name, QWidget * parent) : QWidget(parent)
{
  detectorName = name;

  setupUi(this);

  FSDoubleValidator * fsdValidator = new FSDoubleValidator(this);
  FSDoubleValidator * fsdPosValidator = new FSDoubleValidator(this, 0.0); // validator positive parameters

  // validators
  lineEditAnnulusRadius->setValidator(fsdPosValidator);
  lineEditZ0distance->setValidator(   fsdPosValidator);
  lineEditZ0width->setValidator(      fsdPosValidator);
  lineEditZ0height->setValidator(     fsdPosValidator);
  lineEditZ0angle->setValidator(      fsdValidator   );
  lineEditZ0angleAzim->setValidator(  fsdValidator   );

  lineEditP3z->setValidator(fsdValidator);
  lineEditP2x->setValidator(fsdValidator);
  lineEditP2y->setValidator(fsdValidator);
  lineEditP2z->setValidator(fsdValidator);
  lineEditP3x->setValidator(fsdValidator);
  lineEditP1y->setValidator(fsdValidator);
  lineEditP1x->setValidator(fsdValidator);
  lineEditP1z->setValidator(fsdValidator);
  lineEditP3y->setValidator(fsdValidator);

  connect( tabWidgetGeometry,  SIGNAL(currentChanged(int)), SLOT(updateGeometryTab(int)) ); 

}

// control geometry update actions when geometry tab 'itab' is changed
void DetectorForm::updateGeometryTab(int itab) {
  if(itab==0) {
    // tell user that setting detector parameters in 'Z=0' tab override any changes in 'arbitrary' tab
    if(QMessageBox::warning(this, "Corteo user interface", "'Z=0' geometry parameters will overwrite 'arbitrary' parameters.\nProceed anyway?",
                            QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape)
       == QMessageBox::No) {
      tabWidgetGeometry->setCurrentIndex(1); // set tab back to 'arbitrary'
    } else { 
      updateDetectorPoints(); // 'Z=0' is selected so set 'arbitrary' parameters according to those in 'Z=0'
    } 
  } else {
      updateDetectorPoints(); // before swithing to 'arbitrary', set parameters according to those in 'Z=0'
  }
}


// set 'arbitrary' tab parameters according to parameters in 'Z=0' tab
// do it only at certain conditions, listed below
void DetectorForm::updateDetectorPoints() {

  // update if...
  if(!lineEditZ0angle->text().isEmpty() &&     // all fields in 'Z=0' tab contain text
     !lineEditZ0angleAzim->text().isEmpty() &&
     !lineEditZ0distance->text().isEmpty() &&
     !lineEditZ0width->text().isEmpty() &&
     !lineEditZ0height->text().isEmpty() ) {

    // get angle, distance, width, height
    double a = lineEditZ0angle->text().toDouble()*PI/180;
    double az= lineEditZ0angleAzim->text().toDouble()*PI/180;
    double d = lineEditZ0distance->text().toDouble();
    double w = lineEditZ0width->text().toDouble();
    double h = lineEditZ0height->text().toDouble();

    // detector center and dimensions
    double xc = -d*cos(a);  // x>0 inside target and angle is with surface normal, which is -x
    double yc = -d*sin(a)*cos(az);
    double zc = -d*sin(a)*sin(az);
    double dx = w/2.*sin(a);
    double dy = w/2.*cos(a)*cos(az);
    double dy2 = h/2.*sin(az);
    double dz = w/2.*cos(a)*sin(az);
    double dz2 = h/2.*cos(az);

    lineEditP1x->setText(QString::number(xc+dx));
    lineEditP1y->setText(QString::number(yc - dy + dy2));
    lineEditP1z->setText(QString::number(zc - dz - dz2));

    lineEditP2x->setText(QString::number(xc+dx));
    lineEditP2y->setText(QString::number(yc - dy - dy2));
    lineEditP2z->setText(QString::number(zc - dz + dz2));

    lineEditP3x->setText(QString::number(xc-dx));
    lineEditP3y->setText(QString::number(yc + dy - dy2));
    lineEditP3z->setText(QString::number(zc + dz + dz2));
  }
}

// create a XML node for element parameters
QDomElement DetectorForm::XMLNode(QDomDocument &doc) { 

  QDomElement e = doc.createElement( "detector" );

  e.setAttribute( "rectangular",  radioButtonRectangular->isChecked()?QString("true"):QString("false") );

  e.setAttribute( "geometry",     tabWidgetGeometry->currentIndex() );
  e.setAttribute( "Z0_angle",     lineEditZ0angle->text() );
  e.setAttribute( "Z0_azimAngle", lineEditZ0angleAzim->text() );
  e.setAttribute( "Z0_distance",  lineEditZ0distance->text() );
  e.setAttribute( "Z0_width",     lineEditZ0width->text() );
  e.setAttribute( "Z0_height",    lineEditZ0height->text() );

  e.setAttribute( "p1x",          lineEditP1x->text() );
  e.setAttribute( "p1y",          lineEditP1y->text() );
  e.setAttribute( "p1z",          lineEditP1z->text() );
  e.setAttribute( "p2x",          lineEditP2x->text() );
  e.setAttribute( "p2y",          lineEditP2y->text() );
  e.setAttribute( "p2z",          lineEditP2z->text() );
  e.setAttribute( "p3x",          lineEditP3x->text() );
  e.setAttribute( "p3y",          lineEditP3y->text() );
  e.setAttribute( "p3z",          lineEditP3z->text() );

  e.setAttribute( "is_annulus",   checkBoxAnnular->isChecked()?QString("true"):QString("false") );
  e.setAttribute( "annulus_radius", lineEditAnnulusRadius->text() );

  return e;
}


bool DetectorForm::setFromNode(QDomNode &n) {

  QDomElement e = n.toElement();
  if( !e.isNull() ) {
    if( e.tagName() == "detector" ) {
  
      if(e.attribute("rectangular","false")=="true") radioButtonRectangular->setChecked(true);
      else radioButtonElliptical->setChecked(true);

      tabWidgetGeometry->setCurrentIndex(e.attribute("geometry", "").toInt() );
      lineEditZ0angle->setText(    e.attribute("Z0_angle",    ""));
      lineEditZ0angleAzim->setText(e.attribute("Z0_azimAngle",""));
      lineEditZ0distance->setText( e.attribute("Z0_distance", ""));
      lineEditZ0width->setText(    e.attribute("Z0_width",    ""));
      lineEditZ0height->setText(   e.attribute("Z0_height",   ""));

      lineEditP1x->setText( e.attribute("p1x", ""));
      lineEditP1y->setText( e.attribute("p1y", ""));
      lineEditP1z->setText( e.attribute("p1z", ""));
      lineEditP2x->setText( e.attribute("p2x", ""));
      lineEditP2y->setText( e.attribute("p2y", ""));
      lineEditP2z->setText( e.attribute("p2z", ""));
      lineEditP3x->setText( e.attribute("p3x", ""));
      lineEditP3y->setText( e.attribute("p3y", ""));
      lineEditP3z->setText( e.attribute("p3z", ""));

      checkBoxAnnular->setChecked((e.attribute("is_annulus", "false")=="true")?true:false );
      lineEditAnnulusRadius->setText(e.attribute("annulus_radius", ""));

      QDomNode nn = n.firstChild();
      if(nn.isNull() ) return false;

      return true;
    }
  }
  return false;
}

bool DetectorForm::emptyFields() {
  return lineEditP1x->text().isEmpty() ||
         lineEditP1y->text().isEmpty() ||
         lineEditP1z->text().isEmpty() ||
         lineEditP2x->text().isEmpty() ||
         lineEditP2y->text().isEmpty() ||
         lineEditP2z->text().isEmpty() ||
         lineEditP3x->text().isEmpty() ||
         lineEditP3y->text().isEmpty() ||
         lineEditP3z->text().isEmpty();

}

QString DetectorForm::corteo(bool *error) {
  updateDetectorPoints();

  double p1x = lineEditP1x->text().toDouble()*1e7;
  double p1y = lineEditP1y->text().toDouble()*1e7;
  double p1z = lineEditP1z->text().toDouble()*1e7;
  double p2x = lineEditP2x->text().toDouble()*1e7;
  double p2y = lineEditP2y->text().toDouble()*1e7;
  double p2z = lineEditP2z->text().toDouble()*1e7;
  double p3x = lineEditP3x->text().toDouble()*1e7;
  double p3y = lineEditP3y->text().toDouble()*1e7;
  double p3z = lineEditP3z->text().toDouble()*1e7;

  *error = false;
  QString serror;

  if(emptyFields()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror += "point entries of " + detectorName + " detector empty";
  } else {

    double normp2p1 = sqrt((p2x-p1x)*(p2x-p1x)+(p2y-p1y)*(p2y-p1y)+(p2z-p1z)*(p2z-p1z));
    double normp3p2 = sqrt((p3x-p2x)*(p3x-p2x)+(p3y-p2y)*(p3y-p2y)+(p3z-p2z)*(p3z-p2z));

    if( normp2p1<=0 ) {
      if(*error) serror+= ", ";
      else *error = true;
      serror += detectorName + " length of P2-P1 is null";
    }
    if( normp3p2<=0 ) {
      if(*error) serror+= ", ";
      else *error = true;
      serror += detectorName + " length of P3-P2 is null";
    }
    double vv = fabs((p2x-p1x)*(p3x-p2x)+(p2y-p1y)*(p3y-p2y)+(p2z-p1z)*(p3z-p2z))/normp2p1/normp3p2;
    if( vv>1e-3 ) {
      if(*error) serror+= ", ";
      else *error = true;
      serror += detectorName + QString(" P2-P1 not perpendicular to P3-P2 %1").arg(vv);
    }
  }

  if(*error) return serror;

  QString s = QString("%1  %2 %3 %4  %5 %6 %7  %8 %9 %10  %11")
                 .arg(radioButtonElliptical->isChecked()?1:0)
                 .arg(p1x).arg(p1y).arg(p1z)
                 .arg(p2x).arg(p2y).arg(p2z)
                 .arg(p3x).arg(p3y).arg(p3z)
                 .arg(checkBoxAnnular->isChecked()?lineEditAnnulusRadius->text().toDouble()*1e7:0);

  return s;

}

