/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "ElementForm.h"
#include <math.h>
#include <QMessageBox>

ElementForm::ElementForm(QWidget * parent) : QWidget(parent)
{
  // set default as carbon
  setupUi(this);
  setElement( 6 ); 

  // connect element button and useElement checkbox
  connect( pushButtonElement,   SIGNAL(clicked()), SLOT(setElement()) ); 
  
  // validator for double parameters >=0
  FSDoubleValidator * fsdPosValidator = new FSDoubleValidator(parent, 0.0); 
  lineEditConcentration->setValidator(fsdPosValidator);

}

// set layer parameters and its related elements from a XML node 
bool ElementForm::setFromNode(QDomNode &n) {

  PeriodicTable periodicTable;
  QDomElement e = n.toElement();
  if( !e.isNull() ) {
    if( e.tagName() == "element" ) {
      checkBoxUseElement->setChecked(     (e.attribute("use_element",   "no")=="yes")?true:false );
      setElement(periodicTable.returnAtom( e.attribute("Z",             ""     ).toInt() ));
      comboBoxElementMass->setCurrentIndex(e.attribute("selected_mass", ""     ).toInt() );
      lineEditConcentration->setText(      e.attribute("concentration", ""     ));
      checkBoxSimulate->setChecked(       (e.attribute("simulate",      "no")=="yes")?true:false );
      checkBoxIsotopes->setChecked(       (e.attribute("isotopes",      "no")=="yes")?true:false );
      comboBoxCStable->setCurrentIndex(    e.attribute("CStable",       "0" ).toInt() );
      lineEditEd->setText(                 e.attribute("Edisp",         "20"   ));
      lineEditEb->setText(                 e.attribute("Ebind",         "2"    ));
      return true;
    }
  }
  return false;
}


// create a XML node for element parameters
QDomElement ElementForm::XMLNode(QDomDocument &doc) { 

  QDomElement e = doc.createElement( "element" );
  e.setAttribute( "use_element",   checkBoxUseElement->isChecked()?QString("yes"):QString("no") );
  e.setAttribute( "Z",             _element.Z );
  e.setAttribute( "selected_mass", comboBoxElementMass->currentIndex() );
  e.setAttribute( "concentration", lineEditConcentration->text() );
  e.setAttribute( "simulate",      checkBoxSimulate->isChecked()?QString("yes"):QString("no"));
  e.setAttribute( "isotopes",      checkBoxIsotopes->isChecked()?QString("yes"):QString("no"));
  e.setAttribute( "CStable",       comboBoxCStable->currentIndex() );
  e.setAttribute( "Edisp",         lineEditEd->text() );
  e.setAttribute( "Ebind",         lineEditEb->text() );

  return e;
}


// fire up a periodic table dialog from which the user will select an element
// WARNING: can be used externally, does not refer to any specific member of this class
Atom ElementForm::selectElement(QWidget * parent) {

  // Create a modal dialog window ...
  QDialog * periodicTableDialog = new QDialog(parent);
  periodicTableDialog->setFixedSize(QSize(500,255));

  // ...that will containt a periodic table widget to select a single element
  PeriodicTable * periodicTable = new PeriodicTable(500,255, true, periodicTableDialog);

  // when the widget tels that an element is selected, close the dialog
  connect(periodicTable, SIGNAL(elementSelected()), periodicTableDialog, SLOT(accept()) );

  periodicTableDialog->exec();

  Atom selectedElement = periodicTable->selectedElement();
  delete periodicTable;
  delete periodicTableDialog;

  // return the selected element info
  return selectedElement;

}

// return the element selected from the periodic table, also setting the specified pushbutton and combobox
// WARNING: static member, can be used externally, does not refer to any specific member of this class
Atom ElementForm::selectElement(QWidget * parent, QPushButton * pb, QComboBox * cb) {
  Atom a = selectElement(parent);
  if(a.Z!=0) setElement(a, pb, cb);
  return a;
}


// return Atom information structure corresponding to atomic symbol
// if a button and a mass list box are specified, set them
// WARNING: static member, can be used externally, does not refer to any specific member of this class
Atom ElementForm::setElement(QString atomicSymbol, QPushButton* pb, QComboBox *cb, double mass) {
    PeriodicTable periodicTable;
    Atom a = periodicTable.returnAtom(atomicSymbol);
    if(a.Z!=0) setElement(a, pb, cb, mass);
    return a;
}


// return Atom information structure corresponding to atomic number Z
// set the specified button and a mass list box
// WARNING: static member, can be used externally, does not refer to any specific member of this class
Atom ElementForm::setElement(int Z, QPushButton* pb, QComboBox *cb, double mass) {
    PeriodicTable periodicTable;
    Atom a = periodicTable.returnAtom(Z);
    if(a.Z!=0) setElement(a, pb, cb, mass);
    return a;
}

// set element button and mass list of the indicated pushbutton pb and combobox cb
// the isotope mass list has an icon which darkness is proportional to the isotope abundance 
// (if mass!=0, the mass box selection is set to the closest isotope mass, otherwise it is set to MAI)
// WARNING: can be used externally, does not refer to any specific member of this class
Atom ElementForm::setElement(Atom a, QPushButton * pb, QComboBox * cb, double mass) {
  pb->setText(a.symbol);

  // set mass list
  int i, minDigit = 20;
  double totalAbundance = 0;
  cb->clear();
  for(i=0; i<a.nIsotopes(); i++) {
    QPixmap pm(16,16);
    int ia = int((1.0-a.isotopeAbundance[i])*255);
    pm.fill(QColor(ia,ia,255)); 
    cb->addItem(QIcon(pm), a.isotopeMass[i]); // add mass next to a color square as blue as mass abundance

    // number of significant digits, given the abundance, for the average mass
    int numDigit = a.isotopeMass[i].length()-a.isotopeMass[i].indexOf(".")
                    -int(log10(a.isotopeAbundance[i]>0?a.isotopeAbundance[i]:1e-100));
    if(numDigit>=0 && numDigit<minDigit) minDigit=numDigit;

    totalAbundance+=a.isotopeAbundance[i];
  } 
  // add average mass only if there is abundance (i.e. there are naturally present isotopes)
  if(totalAbundance>0) cb->addItem(QIcon(":/images/averageMass.png"), QString("%1").arg(a.averageMass, 0, 'f', minDigit-1));

  // select mass closest to "mass" argument, will be MAI if no mass was specified
  int select = 0;
  if(mass==0.) mass=a.isotopeMass[a.MAI].toDouble(); // if no mass specified, consider MAI
  for(i=0; i<a.nIsotopes(); i++)
    if( fabs(a.isotopeMass[i].toDouble()-mass)<fabs(a.isotopeMass[select].toDouble()-mass) )
      select = i;

  cb->setCurrentIndex(select);

  return a;
}


// update the enable/disable state of widgets
// set isDetection true if there can be detection (no detection for implantation)
// set isFullCascade  true if we compute the full cascade in the case of implantation, and we need Ed and Eb
bool ElementForm::update(bool isDetection, bool isFullCascade) {

  bool b = checkBoxUseElement->isChecked();

  // if element is used, enable the frame containting the other widgets 
  frameElement->setEnabled(b); 

  // we can simulated isotopes and use tabulated cross-sections only if the element itself is simulated
  checkBoxIsotopes->setEnabled(checkBoxSimulate->isChecked()); 
  comboBoxCStable->setEnabled(checkBoxSimulate->isChecked());

  // if simulating isotopes, this is according to their natural abundance,
  // so mass selection is pointless and perhaps misleading, lets disable it
  comboBoxElementMass->setDisabled(isDetection && checkBoxSimulate->isChecked() && checkBoxIsotopes->isChecked());

  if(isDetection) {
    stackedWidget->setCurrentIndex(1);
  } else {
    if(isFullCascade) {
      stackedWidget->setCurrentIndex(2);
    } else {
      stackedWidget->setCurrentIndex(0);
    }
  }  
  return checkBoxUseElement->isChecked();
}

// set this form element (including its pushbutton and combobox) to element 'a'
void ElementForm::setElement(Atom a, double mass) {
  _element = setElement(a, pushButtonElement, comboBoxElementMass, mass);
}


// set this form element to an element selected from the periodic table
Atom ElementForm::setElement() {
  Atom a = selectElement(this);
  if(a.Z!=0) setElement(a);
  return a;
}

// set this form element to the atom corresponding to atomic symbol
Atom ElementForm::setElement(QString atomicSymbol, double mass) {
    PeriodicTable periodicTable;
    Atom a = periodicTable.returnAtom(atomicSymbol);
    if(a.Z!=0) setElement(a, mass);
    return a;
}

// set this form element (including its pushbutton and combobox) to the atom corresponding to atomic number Z
Atom ElementForm::setElement(int Z, double mass) {
    PeriodicTable periodicTable;
    Atom a = periodicTable.returnAtom(Z);
    if(a.Z!=0) setElement(a, mass);
    return a;
}


// return the mass currently selected
// return -1 upon conversion error
double ElementForm::mass() {
    bool ok;
    double val = comboBoxElementMass->currentText().toDouble(&ok);
    if(ok)
       return val;

    return -1;

}

// return a string containing Z and mass
// if isotope < 0 or isotope>=nIsotopes(), return isotope mass in comboBoxElementMass, 
// otherwise return isotope mass
QString ElementForm::ZMass(int isotope) {
  if(isotope < 0 || isotope>=_element.nIsotopes()) 
    return QString("%1 %2").arg(_element.Z).arg(comboBoxElementMass->currentText());
  else
    return QString("%1 %2").arg(_element.Z).arg(_element.isotopeMass[isotope]);
}

// if element is used and concentration>0, return the concentration
// otherwise, return 0
double ElementForm::concentration() {
  double conc = lineEditConcentration->text().toDouble();
  if(checkBoxUseElement->isChecked() && conc>0) return conc;
  return 0;
}


// if element is used and concentration>0, return a string containing Z, mass and concentration,
// and the number and list of abudance, mass of the isotopes to simulate
// otherwise, return a null string
// Note: if "simulate isotopes" checkbox is checked, sets 0 for the mass, so Corteo will use the average mass (bug in previous versions)
QString ElementForm::corteo() {

    if(checkBoxUseElement->isChecked() && lineEditConcentration->text().toDouble()>0) {

        bool simulateIsotopes = checkBoxIsotopes->isChecked();
        QString s = QString("%1 %2 %3 %4 %5")
                .arg(_element.Z)
                .arg(simulateIsotopes ? QString("0") : comboBoxElementMass->currentText())
                .arg(lineEditConcentration->text())
                .arg(lineEditEd->text())
                .arg(lineEditEb->text());

        s += (simulateIsotopes ? QString(" 1") :  QString(" 0"));

        unsigned int nisotopes = numberIsotopesToSimulate();
        switch (nisotopes) {
        case 0:
            s+= QString(" 0");
            break;
        case 1:
            s+= QString(" 1 1 %1 %2").arg(comboBoxElementMass->currentText()).arg(comboBoxCStable->currentIndex());
            break;
        default:
            s+= QString(" %1").arg(nisotopes);
            for(unsigned int i=0; i<nisotopes; i++)
                s+= QString(" %1 %2 %3").arg(_element.isotopeAbundance[i]).arg(_element.isotopeMass[i]).arg(comboBoxCStable->currentIndex());
        }

        return s;

      }
      return QString(); // null string

}

bool ElementForm::isUsed() {
  return checkBoxUseElement->isChecked();
}


int ElementForm::numberIsotopesToSimulate() {
  if(!isUsed()) return 0; // not used, no simulation
  if(!checkBoxSimulate->isChecked()) return 0;   // no simulation of this element
  if(!checkBoxIsotopes->isChecked()) return 1;   // no isotope simulation, only one simulation
  else return _element.nIsotopes();    
}
