/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "LayerForm.h"
#include <QMessageBox>

LayerForm::LayerForm(QWidget * parent, int maxelement) : QWidget(parent)
{
  setupUi(this);

  green = QIcon(":/images/green.png");
  gray  = QIcon(":/images/gray.png");

  // validator for double parameters >=0
  FSDoubleValidator * fsdValidator =    new FSDoubleValidator(this);
  FSDoubleValidator * fsdPosValidator = new FSDoubleValidator(this, 0.0);
  QIntValidator     * intValidator =    new QIntValidator(this);
  intValidator->setBottom(0);

  lineEditDensity->setValidator(fsdPosValidator);
  lineEditThickness->setValidator(fsdPosValidator);
  lineEditRoughness->setValidator(fsdValidator);
  lineEditNprojectile->setValidator(intValidator);

  // set element tabs in layers
  tabWidgetElements->removeTab(0); // remove default element tab
  for(int ielement=0;ielement<maxelement; ielement++) {
    // create a widget that contains an element parameter tab form
    ElementForm * ef = new ElementForm(parent);
    element.append(ef);
    tabWidgetElements->addTab(ef, gray, "");
    connect( ef->checkBoxUseElement,  SIGNAL(clicked()), SLOT(useElementChanged()) );
  }

  connect( comboBoxMaterial,  SIGNAL(currentIndexChanged(QString)), SLOT(showHideThickness( QString )) );

}

void LayerForm::showHideThickness( QString text ) {
    if(text == "Substrate") {
        labelThickness->show();
        lineEditThickness->show();
    } else {
        labelThickness->hide();
        lineEditThickness->hide();
    }

}


// create a XML node for specified layer and its elements
QDomElement LayerForm::XMLNode(QDomDocument &doc) {

    QDomElement e = doc.createElement( "layer" );
    e.setAttribute( "use_layer", checkBoxUseLayer->isChecked()?QString("true"):QString("false") );
    e.setAttribute( "density",   lineEditDensity->text() );
    e.setAttribute( "thickness", lineEditThickness->text() );
    e.setAttribute( "roughness", lineEditRoughness->text() );
    e.setAttribute( "nprojectile", lineEditNprojectile->text() );
    e.setAttribute( "mat_color", comboBoxMaterial->itemData( comboBoxMaterial->currentIndex() ).toUInt() );

    for(int ielement=0; ielement<element.count(); ielement++) {
      e.appendChild( element[ielement]->XMLNode(doc) );
    }
    return e;
}


// set layer parameters and its related elements from an XML node
void LayerForm::setFromNode(QDomNode &n) {

  QDomElement e = n.toElement();
  if( !e.isNull() ) {
    if( e.tagName() == "layer" ) {
      checkBoxUseLayer->setChecked((e.attribute("use_layer", "false")=="true")?true:false );
      lineEditDensity->setText(     e.attribute("density",   ""));
      lineEditThickness->setText(   e.attribute("thickness", ""));
      lineEditRoughness->setText(   e.attribute("roughness", ""));
      lineEditNprojectile->setText( e.attribute("nprojectile", ""));
      loadedMaterialColor = e.attribute( "mat_color", "1" ).toUInt();

      QDomNode nn = n.firstChild();
      int ielement = 0;
      bool tooMany = false;
      while( !nn.isNull() ) {
        if(ielement>=tabWidgetElements->count()) {
          tooMany = true;
          break;
        } else {
          element[ielement]->setFromNode(nn);
        }
        nn = nn.nextSibling();
        ielement++;
      }
      if(tooMany) QMessageBox::warning( this, "Corteo user interface",
            QString("Too many elements in layer, only loading first %1").arg(tabWidgetElements->count()) );
    }
  }
}

void LayerForm::useElementChanged() {
  int ielement = tabWidgetElements->currentIndex();
  bool be = element[ielement]->checkBoxUseElement->isChecked();

  // icon green or gray whether element is used or not
  tabWidgetElements->setTabIcon(ielement, be?green:gray);
}


// update the enable/disable state of widgets
// set isElementDetection true if there can be detection (no detection for implantation)
// set isFullCascade  true if we compute the full cascade in the case of implantation, thus we need Ed and Eb
bool LayerForm::update(bool isElementDetection, bool isFullCascade) {

  bool bl = checkBoxUseLayer->isChecked();

  // if layer is used, enable the frame containting the other widgets
  frameLayer->setEnabled(bl);
  lineEditNprojectile->setEnabled(isElementDetection);
  label_NProjectile->setEnabled(isElementDetection);

  // update element tabs
  for(int ielement=0; ielement<element.count(); ielement++) {
    bool b = element[ielement]->update(isElementDetection, isFullCascade);
    // icon green or gray whether element is used or not
    tabWidgetElements->setTabIcon(ielement, b?green:gray);
    element[ielement]->labelNumber->setText(QString("Elem. #%1").arg(ielement));
  }

  return bl;
}


QString LayerForm::corteo(QString name, unsigned long numproj, bool *error) {

  *error = false;
  QString serror;

  if(!checkBoxUseLayer->isChecked()) {
    *error = true;
    serror = name + " not used";
    return serror;
  }

  // count elements
  int numberElements = 0;
  for(int j=0; j<element.count(); j++) {
    if(element[j]->checkBoxUseElement->isChecked()) {
      if(element[j]->concentration()==0) {
        if(*error) serror += ", ";
        else *error = true;
        serror += element[j]->element().symbol + " concentration in " + name;
      } else {
        numberElements++;
      }
    }
  }

  if(numberElements==0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror += "no elements are used or have concentration>0 in " + name;
  }

  if(lineEditThickness->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror += name + " tickness";
  }

  if(lineEditDensity->text().isEmpty() || lineEditDensity->text().toDouble()==0.0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror += name + " density";
  }

  if(*error) return serror;

  // check if there is a number of projectiles
  if(lineEditNprojectile->text().toUInt()>0)
      numproj = lineEditNprojectile->text().toUInt();

  QString s = QString("%1 %2 %3 %4 %5\n% Z, mass, proportion, Edisp, Ebind, simNaturalAbundance, N isotopes to simulate, N*[fraction, mass, use CS table]")
                  .arg(lineEditThickness->text())
                  .arg(lineEditRoughness->text().isEmpty()?QString("0"):lineEditRoughness->text())
                  .arg(lineEditDensity->text().toDouble()*1e-24)
                  .arg(numberElements)
                  .arg(numproj);

  for(int j=0; j<element.count(); j++) {
    QString ss = element[j]->corteo();  // will be null if not used or with no concentration
    if(!ss.isNull()) s += QString("\n %1").arg(ss);
  }


  return s;
}

// return true if layer's "used" checkbox is checked
// as well as the "used" checkbox of at least one of its elements
// return false otherwise
bool LayerForm::isUsed() {
    if(checkBoxUseLayer->isChecked())
        for(int j=0; j<element.count(); j++)
          if(element[j]->isUsed())
              return true;

    return false;
}
