/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <QRegExp>

#include "FSAxis.h"
#include "FSEditAxis.h"

//
// class initialization
//
FSAxis::FSAxis(bool iisHorizontal, QString ttitle)
{
    isHorizontal = iisHorizontal;
    logAxis = false;

    min = 0.0;
    max = 1.0;
    minC = 0.0;
    maxC = 1.0;

    framePos = 3;
    framePen = QPen(Qt::black, 2);

    tickPos = 1;
    tickPen = QPen(Qt::black, 1);

    grid = true;
    gridPen = QPen(Qt::black, 1, Qt::DotLine);

    title = (ttitle.length()>0)?ttitle:isHorizontal?QString("X"):QString("Y");
    titleFont = QFont("times",12);
    titleColor = QColor(Qt::black);
    titlePos = 1;

    labelColor = QColor(Qt::black);
    labelFont = QFont("times",10);
    labelPos = 1;
    ntick = 5;

}

//
// destructor
//
FSAxis::~FSAxis()
{

}


void FSAxis::toggle( void )
{
    logAxis = !logAxis;
}

//
// draw the axis along r
// REP stands for Roundoff Error Protection, i.e., numbers are adjusted to prevent roundoff error
//
void FSAxis::draw(QPainter *dc, double fsize, QRect r)
{
    bool h=isHorizontal;
    int li, fudge = int(fsize/5);

    if(!formCalc(r)) return; // problems calculating world transform

    dc->save();

    for(double i=min; i<=max+major/1.e10; i+=major) // REP: major/1.e10
    {
        // grid
        int size1 = 0, size2 = 0;
        li = logic(i);
        dc->setPen(gridPen);
        if(grid && i>min && i<max) dc->drawLine(h?li:r.left(),  h?r.bottom():li, h?li:r.right(),  h?r.top():li );

        // major ticks
        int l1 = (tickPos & 1)?int(fsize/2.+.5):0;
        int l2 = (tickPos & 2)?int(fsize/2.+.5):0;
        dc->setPen(tickPen);
        if(framePos&1 && i>min && i<max) dc->drawLine(h?li:r.left()+l1,  h?r.bottom()-l1:li, h?li:r.left()-l2,  h?r.bottom()+l2:li );
        if(framePos&2 && i>min && i<max) dc->drawLine(h?li:r.right()-l1, h?r.top()+l1:li,    h?li:r.right()+l2, h?r.top()-l2:li    );

        if(framePos&1) size1+=l2;
        if(framePos&2) size2+=l2;

        // minor ticks
        l1/=2;
        l2/=2;
        if(logAxis)
            for(double j=pow(10.,i)*2.; j<=pow(10.,i)*9.00001 && j<=pow(10.,max); j+=pow(10.,i)) // REP: 9.00001 instead of 9. and j<=pow(10.,max)
            {
                int lj = logic(log10(j));
                if(framePos&1) dc->drawLine(h?lj:r.left()+l1,  h?r.bottom()-l1:lj, h?lj:r.left()-l2,  h?r.bottom()+l2:lj );
                if(framePos&2) dc->drawLine(h?lj:r.right()-l1, h?r.top()+l1:lj,    h?lj:r.right()+l2, h?r.top()-l2:lj    );
            }
        else
            for(double j=i+minor; j<=i+major-minor+major/1.e10 && j<=max-minor+major/1.e10; j+=minor) // REP: major/1.e10 and  j<=max-minor
            {
                int lj = logic(j);
                if(framePos&1) dc->drawLine(h?lj:r.left()+l1,  h?r.bottom()-l1:lj, h?lj:r.left()-l2,  h?r.bottom()+l2:lj );
                if(framePos&2) dc->drawLine(h?lj:r.right()-l1, h?r.top()+l1:lj,    h?lj:r.right()+l2, h?r.top()-l2:lj    );
            }

        //labels
        QFont font(labelFont);
        int ffs = int(font.pixelSize()*fsize/FS_SCALE);
        if (ffs> 0) font.setPointSize(ffs);
        dc->setFont(font);
        dc->setPen(labelColor);
 //        char slabel[100];  //  min/1e10 or max/1e10 is to prevent roundoff error near 0 (e.g. 1.543E-14 instead of 0)
 //       QString label(gcvt(logAxis?expt((int)i):(fabs(i)>fabs(min)/1e10||fabs(i)>fabs(max)/1e10)?i:0.,logAxis?4:labelPrec,slabel));
        QString label = QString("%1").arg(logAxis?expt((int)i):(fabs(i)>fabs(min)/1e10||fabs(i)>fabs(max)/1e10)?i:0.,0,'G'); 
        //remove unuseful character in exponent
        label.replace(QString("E+0"), QString("E"));
        label.replace(QString("E-0"), QString("E-"));
        label.replace(QString("E+"), QString("E"));

        if(label.right(1)==QString(".")) label.remove( label.length()-1,1);
        QRect rtext(0,0,dc->fontMetrics().width(label)+fudge,dc->fontMetrics().height());
        if(labelPos&1)
        {
            if(h)rtext.moveCenter(QPoint(li, r.bottom()+size1+rtext.height()));
            else rtext.moveBottomRight(QPoint(r.left()-size1-int(fsize/5), li+rtext.height()/2));
            dc->drawText(rtext,Qt::AlignVCenter|Qt::AlignRight,label);
        }
        if(labelPos&2)
        {
            if(h)rtext.moveCenter(QPoint(li, r.top()-size2-rtext.height()));
            else rtext.moveBottomLeft(QPoint(r.right()+size2+int(fsize/5),li+rtext.height()/2));
            //rtext.moveCenter(QPoint(h?li:r.right()+size2+rtext.height(), h?r.top()-size2-rtext.height():li));
            dc->drawText(rtext,Qt::AlignVCenter|Qt::AlignLeft,label);
        }
    }

    // frame
    dc->setPen(framePen);
    if(framePos&1) dc->drawLine(h?logic(min):r.left(),  h?r.bottom():logic(min), h?logic(max):r.left(),  h?r.bottom():logic(max) );
    if(framePos&2) dc->drawLine(h?logic(min):r.right(), h?r.top():logic(min),    h?logic(max):r.right(), h?r.top():logic(max)    );

    // title
    QFont font(titleFont);
    int ffs = int(font.pixelSize()*fsize/FS_SCALE);
    if(ffs>0) font.setPointSize(ffs);
    dc->setFont(font);
    dc->setPen(titleColor);

    QString ttitle = title;
    ttitle.replace(QRegExp("[\\^\\_\\\\\\{\\}]"),""); // we want the size of title without control characters
    QRect rtext(0,0,dc->fontMetrics().width(ttitle)+fudge,dc->fontMetrics().height());
    QSize sizeall = getSize(dc, fsize);

    if(isHorizontal)
    {
        if(titlePos&1)
        {
            rtext.moveBottomLeft( QPoint(r.center().x()-rtext.width()/2,r.bottom()+sizeall.width()+rtext.height()/2));
	          drawTextExt(dc, rtext.topLeft(), title); //+" ("+QString::number(font.pixelSize())+")"
        }
        if(titlePos&2)
        {
            rtext.moveTopLeft( QPoint(r.center().x()-rtext.width()/2,r.top()-sizeall.height()+rtext.height()));
	          drawTextExt(dc, rtext.topLeft(), title);
        }
    } else {
        if(titlePos&1)
        {
            dc->save();
            dc->translate( r.left()-sizeall.width()+rtext.height(), r.center().y()+rtext.width()/2 );
            dc->rotate( -90 );
	          drawTextExt(dc, QPoint(0,0), title);
            dc->restore();
        }
        if(titlePos&2)
        {
            dc->save();
            dc->translate( r.right()+sizeall.height()-rtext.height(), r.center().y()-rtext.width()/2 );
            dc->rotate( 90 );
	          drawTextExt(dc, QPoint(0,0), title);
            dc->restore();
        }
    }

    dc->restore();
    return;
}

//
// how many pixels thick is the axis labeling and title
// (qsize.width() corresponds to the space at bottom or left, qsize.height() at top or right)
//
QSize FSAxis::getSize(QPainter *dc, double fsize)
{
    QSize qsize((int)fsize,(int)fsize);
    if(tickPos&2) qsize += QSize((framePos&1)?int(fsize/2):0, (framePos&2)?int(fsize/2):0);  // if outside ticks

    dc->save();
    QFont font(titleFont);
    int ffs = int(font.pixelSize()*fsize/FS_SCALE);
    if(ffs>0) font.setPointSize(ffs);
    dc->setFont(font);
    qsize+=QSize((titlePos&1)?dc->fontMetrics().height():0, (titlePos&2)?dc->fontMetrics().height():0);  // title

    font = labelFont;
    ffs = int(font.pixelSize()*fsize/FS_SCALE/2);
    if(ffs>0) font.setPointSize(ffs);
    dc->setFont(font);
    if(isHorizontal)
        qsize+=QSize((labelPos&1)?dc->fontMetrics().height():0, (labelPos&2)?dc->fontMetrics().height():0);  // horizotal labels
    else {
        int maxLabelWidth = 0;
        for(double i=min; i<=max; i+=major) // for vertical axis, go through all labels and find the whidest one
        {
            char slabel[100]; //  |i|>|min|/1e10 or |i|>|max|/1e10 are there to prevent roundoff error near 0 (e.g. 1.543E-17 instead of 0)
            QString label(gcvt(logAxis?expt((int)i):(fabs(i)>fabs(min)/1e10||fabs(i)>fabs(max)/1e10)?i:0.,logAxis?4:labelPrec,slabel));
            int labelWidth = dc->fontMetrics().width(label);
            if(labelWidth>maxLabelWidth) maxLabelWidth = labelWidth;
        }
        qsize+=QSize((labelPos&1)?maxLabelWidth:0, (labelPos&2)?maxLabelWidth:0);  // horizotal labels
    }

    dc->restore();
    return qsize;

}

//
// special function that interprete \{} for symbols, ^{} for superscript, _{} for subscript
//
void FSAxis::drawTextExt(QPainter * dc, QPoint p, QString s)
{
    dc->save();

    QString ss = QString("");
    QFont ffont = dc->font();
    QFont backupfont = dc->font();
    //ffont.setCharSet(QFont::AnyCharSet);

    double delta = 0., pointsize = ffont.pointSize(), height = p.y();
    int n = s.length();
    int stack[100], istack = 0; // stacks what was the last { for. ( 1=\, 2=^, 3=_ )
    bool flush;

    dc->setFont(ffont);
    for(int i=0; i<n; i++)
    {
        flush = true;
        if(s.mid(i,2)==QString("\\{"))
        {
            ffont.setFamily("symbol");
            stack[istack++]=1;
            i++;
        }
        else if(s.mid(i,2)==QString("^{"))
        {
            pointsize/=1.5;
            delta-=pointsize/2;
            stack[istack++]=2;
            i++;
        }
        else if(s.mid(i,2)==QString("_{"))
        {
            pointsize/=1.5;
            delta+=pointsize/2;
            stack[istack++]=3;
            i++;
        }
        else if(s[i]=='}')
        {
            if(istack>0)
            {
                switch(stack[--istack]) {
                case 1:
                    ffont.setFamily(backupfont.family());
                    break;
                case 2:
                    delta+=pointsize/2;
                    pointsize*=1.5;
                    break;
                case 3:
                    delta-=pointsize/2;
                    pointsize*=1.5;
                    break;
                }
            }
        }
        else
        {
            ss += s[i];
            flush = false;
        }

        if(flush)
        {
            dc->drawText(p, ss);
            p.setX(p.x()+dc->fontMetrics().width(ss));
            p.setY(int(height+delta));
            //ffont.setPointSizeFloat(pointsize);
            ffont.setPointSize(int(pointsize+.5));
            dc->setFont(ffont);
            ss = QString("");
        }
    }
    dc->drawText(p, ss);
    dc->restore();
}


//
// axis majorant rounding (Method borowed to GEMS I)
//
void FSAxis::round()
{
  double f;

  if(minC>maxC) // dumb proof!
  {
       double temp = minC;
       minC = maxC;
       maxC = temp;
  }

  if(logAxis)
  {
	 min=floor(log10(minC<=0.?minCgt0:minC));
	 max= ceil(log10(maxC<=0.?minCgt0:maxC));
	 major = 1;
	 if(max == min) max++;
  }
  else if(maxC==0.0e0 && minC==0.0e0)
  {
	 min=-1.e0;
	 max=1.e0;
	 major=0.2e0;
	 minor=0.1e0;
  }
  else if (maxC==minC)
  {
	 major = pow(10.e0, floor(log10(fabs(maxC)) - 1.e0));
	 min = floor(minC/major) * major;
	 max = ceil(maxC/major) * major;
	 major/=10.e0;
	 minor=major/2.e0;
  }
  else
  {
	 f = nicenum(maxC-minC, false);
	 major = nicenum(f/(ntick-1), true);
	 min = floor(minC/major)*major;
	 max = ceil(maxC/major)*major;
	 if (major/expt(int(log10(major))) == 5.)	// 1, 2 or 5
		minor = major/5.;
	 else
		minor = major/2.;
  }
  labelPrec = int(log10((fabs(max)+fabs(min))/major)) + 4;
}

//
// find a "nice" number below x (b=true) or above x (b=false)
//
double FSAxis::nicenum(double x, bool b)
{
  int expv;			/* exponent of x */
  double f;			/* fractional part of x */
  double nf;			/* nice, rounded fraction */

  expv = int(floor(log10(x)));
  f = x/expt(expv);		/* between 1 and 10 */
  if (b)
	 if (f<1.5)
		nf = 1.;
	 else if (f<3.)
		nf = 2.;
	 else if (f<7.)
		nf = 5.;
	 else
		nf = 10.;
  else
	 if (f<=1.)
		nf = 1.;
	 else if (f<=2.)
		nf = 2.;
	 else if (f<=5.)
		nf = 5.;
	 else
		nf = 10.;

  return nf*expt(expv);
}


//
// reliable calculation of 10^n
//
double FSAxis::expt(int n)
{
  double x = 1.;

  if (n>0)
    for (; n>0; n--)
      x *= 10.;
  else
    for (; n<0; n++)
      x /= 10.;

  return x;
}


//
// calculate the world transform
//
bool FSAxis::formCalc(int MIN, int MAX)
{
    if(max==min) return false;

    if(!isHorizontal)
    {
        int temp = MIN;
        MIN = MAX;
        MAX = temp;
    }
    factor = double(MAX-MIN)/(max-min);
    Origin = double(MIN) - factor * min;
    return true;
}

//
// calculate the world transform from the display rectangle,
// choosing (left,right) for horizontal axis or (top,bottom) for vertical axis 
//
bool FSAxis::formCalc(QRect r)
{
    bool h=isHorizontal;
    return formCalc(h?r.left():r.top(), h?r.right():r.bottom());
}



//
// calculate the value of p in logical units according to world transform
//
int FSAxis::logic(double p, bool linlog)
{
  if(linlog && logAxis) p = log10(p>0.?p:minCgt0 );
  return int(Origin + factor*p);
}

//
// calculate the value of p in graph's units according to world transform
//
double FSAxis::real(int p, bool linlog)
{
  if(factor)
  {
      double pp = (double(p)-Origin)/factor;
      if(linlog && logAxis) pp = pow(10.,pp);
      return pp;
  }

  return HUGE_VAL;
}


//
// open the edit axis dialog; return true Ok is pressed
// (if a variable extremaChanged is supplied, it is set to true if min or max values are changed)
//
bool FSAxis::edit(bool *extremaChanged )
{
    FSEditAxis * axis = new FSEditAxis(this);
    if( axis->exec() )
    {
        title = axis->title();
        settitlePos(axis->titlePos());
        titleFont = axis->titleFont();
        titleColor = axis->titleColor();
        setlabelPos(axis->labelPos());
        labelFont = axis->labelFont();
        labelColor = axis->labelColor();
        framePen = axis->framePen();
        setframePos(axis->framePos());
        tickPen = axis->tickPen();
        settickPos(axis->tickPos());
        gridPen = axis->gridPen();
        grid = axis->drawGrid();
        logAxis = axis->isLog();
        minC = axis->min();
        maxC = axis->max();
        setntick(axis->ntick());

        if(minC>maxC)
        {
            double temp = minC;
            minC = maxC;
            maxC = temp;
        }

        if(extremaChanged)
            *extremaChanged = axis->extremaChanged();

        return true;
    }

    return false;
}

