/* This file is part of the Corteo project
   Copyright (C) 2004-2015 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

// Widget that holds two axes and the address of the first dataset to plot if any.
// Also manages all user interactions

#include <QMenu>
#include <QClipboard>
#include <QFileDialog>
#include <QMessageBox>
#include <QRegExp>
#include <QCursor>
#include <QPolygon>
#include <QSet>

#include "FSGraph.h"
#include "FSEditLegend.h"
#include "FSOpDialog.h"
#include "FSFitDialog.h"
#include "FSFileDialog.h"
#include "JamaQRDecomposition.h"
#include "spline.h"

void FSGraph::init()
{
    setWindowTitle( "XYQt" );
    
    QPalette pal = palette ();
    pal.setBrush(QPalette::Window, QBrush(QColor(Qt::white)) );
    setPalette ( pal );

    setFocusPolicy(Qt::StrongFocus);

    legendFont = QFont("helvetica",12);
    legendPos = 2;
    legendPen = QPen(Qt::black,1);

    resize(1024,768 );

    x = new FSAxis(true,  "X");
    y = new FSAxis(false, "Y");

    _isZoomed = false;
    _isZooming = false;
    _isMarking = false;

    firstDataSet = NULL;
    ndatasets = 0;
    selectedDataSet = NULL;


    firstMarker = NULL;
    nmarkers = 0;

    readOnlyGraph = false;

    allowSpots = false;
    spots = new FSDataSet("", QPen(QColor(0,200,0),1), 2, 3); // green filled circles joined by solid lines
    spots->symbolPen = QPen(QColor(0,200,0),1);

    hasChanged = true;
}

FSGraph::FSGraph(QWidget * parent) : QWidget(parent)
{
    init();
}


FSGraph::FSGraph(QString conffile, QWidget * parent): QWidget(parent)
{
    qint8  c;
    qint32 i;

    init();    // execute usual constructor

    confFile = conffile;

    QFile f(confFile);
    if(!f.exists()) return;    
    
    f.open( QIODevice::ReadOnly );              // open file for reading
    QDataStream s( &f );                        // serialize using f

    s >> graphName;    setGraphName( graphName );

    // legend    
    s >> c;            setlegendPos((char)c);
    s >> legendPen;
    s >> legendFont;

    // x axis
    s >> x->title;
    s >> x->titleFont;
    s >> x->titleColor;
    s >> c;            x->settitlePos((char)c);

    s >> x->labelFont;
    s >> x->labelColor;
    s >> c;            x->setlabelPos((char)c);

    s >> c;            x->logAxis = c!=0;
    s >> c;            x->setframePos((char)c);
    s >> i;            x->setntick((int)i);
    s >> c;            x->settickPos((char)c);
    s >> c;            x->grid = c!=0;

    // y axis
    s >> y->title;
    s >> y->titleFont;
    s >> y->titleColor;
    s >> c;            y->settitlePos((char)c);

    s >> y->labelFont;
    s >> y->labelColor;
    s >> c;            y->setlabelPos((char)c);

    s >> c;            y->logAxis = c!=0;
    s >> c;            y->setframePos((char)c);
    s >> i;            y->setntick((int)i);
    s >> c;            y->settickPos((char)c);
    s >> c;            y->grid = c!=0;

    f.close();

}



FSGraph::~FSGraph()
{

    if(!confFile.isEmpty()) {
        // write graph parameters in specified file
        QFile f(confFile);
        f.open( QIODevice::ReadOnly  );                    // open file for writing
        QDataStream s( &f );                        // serialize using f

        s << graphName;

        // legend    
        s << (qint8) getlegendPos();
        s << legendPen;
        s << legendFont;

        // x axis
        s << x->title;
        s << x->titleFont;
        s << x->titleColor;
        s << (qint8)x->gettitlePos();

        s << x->labelFont;
        s << x->labelColor;
        s << (qint8)x->getlabelPos();

        s << (qint8)(x->logAxis?1:0);
        s << (qint8)x->getframePos();
        s << (qint32)x->getntick();
        s << (qint8)x->gettickPos();
        s << (qint8)(x->grid?1:0);

        // y axis
        s << y->title;
        s << y->titleFont;
        s << y->titleColor;
        s << (qint8)y->gettitlePos();

        s << y->labelFont;
        s << y->labelColor;
        s << (qint8)y->getlabelPos();

        s << (qint8)(y->logAxis?1:0);
        s << (qint8)y->getframePos();
        s << (qint32)y->getntick();
        s << (qint8)y->gettickPos();
        s << (qint8)(y->grid?1:0);

        f.close();
    
    }
    
    delete x;
    delete y;
    delete spots;

    //clearMarkers(); // removed, as it creates a deallocation conflict in MSVC
    clearDataSets();
    
}


// append ddataset to the datasets in the graph.
// if differentColor is true, change the color in linePen to the most different color in RGB_ordCIE94 colors
void FSGraph::appendDataSet(FSDataSet* ddataset)
{
    if(firstDataSet)  // if there are already some datasets
    {
        // go through datasets of this graph to find last
        FSDataSet* pdataset = firstDataSet;
        while(pdataset->nextDataSet) {
            pdataset = pdataset->nextDataSet;
        }
        pdataset->nextDataSet = ddataset;  // tell the last dataset who is now its next dataset
        ddataset->prevDataSet = pdataset;  // tell the new dataset who is its previous dataset

    } else {
        firstDataSet = ddataset;
        selectedDataSet = ddataset;
       // ... and since this is the first dataset in the graph, we are sure no other have the same color
    }

    ndatasets++;
    updateGraph();
}


void FSGraph::saveSelectedDataSet()
{
    if(selectedDataSet) {
        QString fileName( QFileDialog::getSaveFileName() );
        if ( fileName.isNull() )
            return;

        selectedDataSet->savefile(fileName);
        updateGraph();
    }

}


void FSGraph::copySelectedDataSet( void )
{
    if(selectedDataSet)
    {
        FSDataSet* pdataset = new FSDataSet(selectedDataSet);
        pdataset->legend = selectedDataSet->legend;

        // find if the legend ends by _???, where ??? is an int
        bool isUInt;  // will tell us if ??? is actually an int
        QString s = pdataset->legend;
        int n1 = s.lastIndexOf("_");
        int n2 = s.right(s.length()-n1-1).toUInt(&isUInt);
        if(n1!=-1 && isUInt)
        {
            // increment ??? by 1
            s.truncate(n1+1);
            pdataset->legend =  s + QString::number(n2+1);
        }
        else
            // append _0 to the legend
            pdataset->legend = pdataset->legend + QString("_0");

        pdataset->dataHaveChanged = true;
        appendDataSet(pdataset);
        updateGraph();
    }

}


int FSGraph::deleteSelectedDataSet( void )
{
    if (selectedDataSet)
    {
        if(selectedDataSet==firstDataSet)
            firstDataSet=selectedDataSet->nextDataSet;

        FSDataSet * dset = (selectedDataSet->nextDataSet?selectedDataSet->nextDataSet:selectedDataSet->prevDataSet);

        delete selectedDataSet;

        selectedDataSet=dset; // will be NULL if we deleted the last dataset
        updateGraph();
        return --ndatasets;
    }
    return 0;
}

void FSGraph::clearDataSets()
{
    if(firstDataSet)
    {
        FSDataSet* pdataset = firstDataSet;
        while(pdataset->nextDataSet)
        {
            pdataset = pdataset->nextDataSet;
            delete pdataset->prevDataSet;  // ask iteratively to each dataset to delete its previous one
        }
        delete pdataset; // delete the last remaining dataset
        updateGraph();
    }
    firstDataSet = NULL;
    selectedDataSet = NULL;
    ndatasets = 0;
}

void FSGraph::clipboardSelectedDataSet(void) {
  if(selectedDataSet) {
    QString s;
    for(int i=0; i<selectedDataSet->getndata();i++)
      s += QString("%1\t%2\n").arg(selectedDataSet->xp[i]).arg(selectedDataSet->yp[i]);

    QApplication::clipboard()->setText(s);
  }

}

void FSGraph::clipboardGraph(void) {
    QApplication::clipboard()->setPixmap(pm); // private member 'pm' contains the last drawn graph 
}


void FSGraph::appendMarker(double x1, double y1, double x2, double y2)
{
    FSMarker* dmarker = new FSMarker(x1,y1,x2,y2);
    if(firstMarker)
    {
        FSMarker* pmarker = firstMarker;
        while(pmarker->nextMarker) pmarker = pmarker->nextMarker;  // search for the last
        pmarker->nextMarker = dmarker;  // tell the last marker who is now its next marker
        dmarker->prevMarker = pmarker;  // tell the new marker who is its previous marker
    } else {
        firstMarker = dmarker;
    }
    nmarkers++;
    updateGraph();
}

void FSGraph::clearMarkers()
{

    bool doUpdate = false;
    if(firstMarker)
    {
        FSMarker* pmarker = firstMarker;
        while(pmarker->nextMarker)
        {
            pmarker = pmarker->nextMarker;
            delete pmarker->prevMarker;  // ask iteratively to each marker to delete its previous one
        }
        delete pmarker; // delete the last remaining marker
        doUpdate = true;
    }

    if(spots->getndata())
    {
        spots->clear();
        doUpdate = true;
    }

    if (doUpdate) updateGraph();
    firstMarker = NULL;
    nmarkers = 0;

}


//
// Iterates through the datasets to find the extrema majorants.
//
void FSGraph::majorant()
{
    int n = ndatasets;

    if(n == 0)
    {
        x->minC  = 0.1; y->minC = 0.1;   // set defaults
        x->maxC  = 10.; y->maxC = 10.;
        x->setminCgt0(0.1); y->setminCgt0(0.1);
    }
    else
    {
        FSDataSet* d = firstDataSet;

        // initial values
        d->extrema();
        x->minC = d->getxMin();
        y->minC = d->getyMin();
        x->maxC = d->getxMax();
        y->maxC = d->getyMax();
        x->setminCgt0(d->getxMingt0());
        y->setminCgt0(d->getyMingt0());

        // Iterates trough the datasets
        while(d->nextDataSet)
        {
            d = d->nextDataSet;
            d->extrema();    // calculate the extrema of that data set
            if(d->getxMin()    < x->minC   ) x->minC=d->getxMin();
            if(d->getxMax()    > x->maxC   ) x->maxC=d->getxMax();
            if(d->getyMin()    < y->minC   ) y->minC=d->getyMin();
            if(d->getyMax()    > y->maxC   ) y->maxC=d->getyMax();
            if(d->getxMingt0() < x->getminCgt0()) x->setminCgt0(d->getxMingt0());
            if(d->getyMingt0() < y->getminCgt0()) y->setminCgt0(d->getyMingt0());
        }
    }

}

void FSGraph::drawMarkers(QPainter * dc, QRect r )
{
    if(!firstMarker)
        return;

    dc->save();
    dc->setPen(QColor(223,223,223));
    dc->setBrush(QBrush(QColor(223,223,223)));//,Qt::BDiagPattern));

    dc->setClipRect(r);

    FSMarker * pmarker = firstMarker;
    while(pmarker)
    {
        dc->drawRect(QRect( QPoint(x->logic(pmarker->x1,true), y->logic(pmarker->y1,true)),
                            QPoint(x->logic(pmarker->x2,true), y->logic(pmarker->y2,true)) ));
        pmarker = pmarker->nextMarker;
    }
    dc->restore();
}


void FSGraph::paintEvent( QPaintEvent * )
{
    if( hasChanged || previousSize != size()) {
        pm = QPixmap(size());            // create a pixmap the size of the window
        pm.fill( palette().color(QPalette::Window) );

        QPainter paint;
        paint.begin(&pm);
        drawIt( &paint );            // paint the graph on the pixmap (without legend or zooming rectangles)
        paint.end();
        hasChanged = false;
        previousSize = size();
    }

    QPainter dc(this);
    dc.drawPixmap(0,0, pm); // copy (previous or new) pixmap to window (without legend or zooming rectangles)

    dc.save();
    dc.setPen( QPen(QColor(128,128,128), 1, Qt::DotLine) );
    if(selectedDataSet)  dc.drawRect(selectedDataSet->legendRect); // draw the selected dataset rectangle
    if(_isZooming) dc.drawRect( _zoomRect);      // draw zoom rectangle
    if(_isMarking) dc.drawRect(_markRect);       // draw current rectangle for new mark
    dc.restore();

}


void FSGraph::drawIt( QPainter *dc )
{
    // get drawing rectangle
    QSize s = size();
    fsize = pow(double(s.height())*double(s.width()),0.5)/30.;

    // set extrema
    if(!_isZoomed)
        majorant();
    x->round();
    y->round();

    // find how much space we need for the axis title and numbering
    QSize sx = x->getSize(dc,fsize);
    QSize sy = y->getSize(dc,fsize);

    // rectangle along which we draw the axes
    QRect r(sy.width()+10,sx.height()+10, s.width()-sy.width()-sy.height()-20, s.height()-sx.width()-sx.height()-20);

    // legend size
    legendSize(dc, r);
    if(legendPos==3 && legendWidth)
        r.setWidth(r.width()-legendWidth-int(2*fsize)); // make room for outside legend

    // draw markers
    x->formCalc(r);    // make sure world transform is OK before drawing markers
    y->formCalc(r);    
    drawMarkers(dc, r);

    // draw axes (and calculate world tranformation)
    x->draw(dc,fsize,r);
    y->draw(dc,fsize,r);

    // calculate the legend position
    switch (legendPos) {
    case 0:  // no legend
        break;
    case 1: // inside left
        xLeg = x->logic(x->getmin()) + int(fsize*1.5);
        yLeg = y->logic(y->getmax()) + int(fsize);
        break;
    case 2: // inside right
        xLeg = x->logic(x->getmax()) - legendWidth - int(fsize*1.5);
        yLeg = y->logic(y->getmax()) + int(fsize);
        break;
    case 3: // outside right
        xLeg = x->logic(x->getmax())+ y->getSize(dc,fsize).height() + int(fsize*1.5);
        yLeg = y->logic(y->getmax());
        break;
    default:
        break;
    }

    // draw the data
    FSDataSet* pdataset = firstDataSet;
    dc->setClipping(true);
    dc->setClipRect(r);
    while(pdataset)
    {
        pdataset->draw(dc,x,y,fsize, r);
        pdataset = pdataset->nextDataSet;
    }
    if(allowSpots) spots->draw(dc,x,y,fsize);

    dc->setClipping(false);

    // draw the legend
    if(legendPos)
    {
        dc->save();
        QFont font(legendFont);
        int ffs = int(font.pixelSize()*fsize/FS_SCALE);
        if(ffs>0) font.setPixelSize(ffs);
        dc->setFont(font);

        // clipping region: legend rectangle
        dc->setClipping(true);
        r = QRect(int(xLeg-fsize/2),yLeg,legendWidth+int(fsize),legendHeight+int(fsize)) ;
        dc->setClipRect( r.adjusted(-1,-1,3,3) );

        // box around legend
        dc->setPen(legendPen);
        dc->setBrush(QBrush(QColor(Qt::white)));
        dc->drawRect( r );

        // content of the legend
        int lpos = yLeg+int(fsize/2.);
        int fh = dc->fontMetrics().height();
        pdataset = firstDataSet;
        while(pdataset)
        {
            // line
            if(pdataset->linePen.width()>0)
            {
                dc->setPen(pdataset->linePen);
                dc->drawLine(xLeg, lpos+fh/2, xLeg+int(fsize*1.8),lpos+fh/2);
            } else {
                dc->setPen(pdataset->symbolPen);
            }

            // symbol
            QPolygon temparray;
            temparray.setPoints(1, xLeg+int(fsize*0.9), lpos+fh/2);
            pdataset->drawSymbol(dc, &temparray, 1, fsize);

            // text
            QRect rtext(xLeg+int(fsize*2), lpos,legendWidth, fh);
            dc->drawText(rtext, Qt::AlignVCenter|Qt::AlignLeft, pdataset->getlegend());

            // rectangle in which we must click to select the dataset
            pdataset->legendRect = QRect(xLeg, lpos-(legendSpacing-fh)/2, legendWidth, legendSpacing);
            pdataset = pdataset->nextDataSet;
            lpos+=legendSpacing;
        }

        dc->setClipping(false);
        dc->restore();
    }
}

//
// Calculate the size of the legend and reduce the spacing until all the legends fit in the window height.
// (updates legendWidth, legendHeight and legendSpacing)
//
bool FSGraph::legendSize(QPainter *dc, QRect r)
{
    bool reducedSpacing = false;
    if(legendPos==0) return false;

    dc->save();
    QFont font(legendFont);
    int ffs = int(font.pixelSize()*fsize/FS_SCALE);
    if(ffs>0) font.setPixelSize(ffs);
    dc->setFont(font);
    legendWidth  = 0;
    legendHeight = 0;
    legendSpacing = dc->fontMetrics().height() + int(fsize/2);
    if(ndatasets > 0)
    {
        for(;;)  // loop until the legend is smaller that the y axis lenght-fs
        {
            FSDataSet* pdataset = firstDataSet;
            legendHeight = 0;
            while(pdataset)
            {
                if(legendWidth<dc->fontMetrics().width(pdataset->getlegend()))
                    legendWidth = dc->fontMetrics().width(pdataset->getlegend());

                legendHeight += legendSpacing;
                pdataset = pdataset->nextDataSet;
            }
            if(legendHeight < r.height()-int(fsize))
                break;
            else {
                legendSpacing-= int(fsize/10.)>0?int(fsize/10.):1;
                reducedSpacing = true;
            }
        }
    }
    legendWidth+=int(fsize*2); // room to draw the line and symbol

    dc->restore();
    return reducedSpacing;
}


void FSGraph::mousePressEvent( QMouseEvent * e )
{
    bool isDataSet = false;

    // if we are cliking a legend item, select it
    FSDataSet* pdataset = firstDataSet;
    while(pdataset)
    {
        if(pdataset->legendRect.contains(e->pos()))
        {
            selectedDataSet = pdataset;
            isDataSet = true;
            update(); // ... so a rectangle is draw around this dataset 
            break;
        }
        pdataset = pdataset->nextDataSet;
    }


    if(e->button()==Qt::RightButton)
    {
        QMenu * menu = new QMenu( this );

        // we are clickin on a dataset legend
        if(isDataSet)
        {
            if(readOnlyGraph) {
              menu->addAction("Read-Only graph,")->setEnabled(false);
              menu->addAction("no data modification")->setEnabled(false);
              menu->addAction("&To clipboard",   this, SLOT(clipboardSelectedDataSet())   );
            } else {
                menu->addAction(selectedDataSet->legend)->setEnabled(false);
                menu->addSeparator();
                menu->addAction("&Save",   this, SLOT(saveSelectedDataSet()) );
                menu->addAction("&Copy",   this, SLOT(copySelectedDataSet())   );
                menu->addAction("&Delete", this, SLOT(deleteSelectedDataSet()) );
                menu->addAction("&Edit",   this, SLOT(editSelectedDataSet())   );
                menu->addSeparator();
                menu->addAction("So&rt",      this, SLOT(sortSelectedDataSet()) );
                menu->addAction("S&wap",      this, SLOT(swapSelectedDataSet()) );
                menu->addSeparator();
                menu->addAction("&Operation", this, SLOT(slotOp()) );
                menu->addAction("&Fit",       this, SLOT(fit()) );
                menu->addSeparator();
                menu->addAction("&To clipboard",   this, SLOT(clipboardSelectedDataSet())   );
            }
        }

        // we are clicking somewhere on the y axis (left or right)
        else if(x->real(e->x())<x->getmin() || x->real(e->x())>x->getmax())
        {
            QString s = y->logAxis?QString("Linear &Y axis"):QString("Logarithmic &Y axis");
            menu->addAction(s, this, SLOT( toggleYAxis() ));
            menu->addAction("&Edit axis", this, SLOT( editYAxis() ));
        }

        // we are clicking somewhere on the x axis (bottom or top)
        else if(y->real(e->y())<y->getmin() || y->real(e->y())>y->getmax())
        {
            QString s = x->logAxis?QString("Linear &X axis"):QString("Logarithmic &X axis");
            menu->addAction(s, this, SLOT( toggleXAxis() ));
            menu->addAction("&Edit axis", this, SLOT( editXAxis() ));
        }

        // we are clicking inside the graph frame
        else
        {
          if(readOnlyGraph) {
            menu->addAction("Read-Only graph,")->setEnabled(false);
            menu->addAction("no data modification")->setEnabled(false);
          } else {
            menu->addAction("&Load file", this, SLOT(loadFile()))->setEnabled(!readOnlyGraph);
          }
          menu->addSeparator();
          menu->addAction("&Unzoom", this, SLOT(unZoom()))->setEnabled(_isZoomed);
          menu->addAction("Clear &markers", this, SLOT(clearMarkers()))->setEnabled(nmarkers>0 || spots->getndata()>0);
          menu->addSeparator();
          menu->addAction("Edit &legend",  this, SLOT(editLegend()))->setEnabled(!readOnlyGraph);
          menu->addAction("&Image to clipboard", this, SLOT(clipboardGraph()) );
        }

        menu->exec(e->globalPos());

    }

    // clicking with left button in the graph but outside the legend: zoom
    if(e->button()==Qt::LeftButton && e->modifiers()==Qt::NoModifier && !isDataSet)
    {
        _zoomRect = QRect(e->x(), e->y(), 0, 0);
        _isZooming = true;
    }

    // clicking with left button + CTRL in the graph but outside the legend: marker
    if(e->button()==Qt::LeftButton && e->modifiers()==Qt::ControlModifier && !isDataSet)
    {
        _markRect = QRect(e->x(), e->y(), 0, 0);
        _isMarking = true;
    }

    // clicking with left button + Shift in the graph but outside the legend: spots
    if(allowSpots && e->button()==Qt::LeftButton && e->modifiers()==Qt::ShiftModifier && !isDataSet)
    {
        spots->append(x->real(e->x(), true), y->real(e->y(), true));
        emit spotCreated();
        updateGraph();
    }

}


void FSGraph::mouseMoveEvent( QMouseEvent * e )
{
    if(_isZooming){
        _zoomRect = QRect(_zoomRect.topLeft(), e->pos());
        update();
    } 
    else if(_isMarking) {
        _markRect = QRect(_markRect.topLeft(), e->pos());
        update();
    }
    
}


void FSGraph::mouseReleaseEvent( QMouseEvent * e )
{
    if (_isZooming)
    {
        _zoomRect = QRect(_zoomRect.topLeft(), e->pos()).normalized();
        if( _zoomRect.width() > 5 && _zoomRect.height() > 5 &&
           // no infinite zooming allowed except around the origin!!
           (x->getmax()-x->getmin())/(fabs(x->getmax())+fabs(x->getmin()))>1e-10  &&
           (y->getmax()-y->getmin())/(fabs(y->getmax())+fabs(y->getmin()))>1e-10  )
        {
            x->minC = x->real(_zoomRect.bottomLeft().x(),true);
            y->minC = y->real(_zoomRect.bottomLeft().y(),true);
            x->maxC = x->real(_zoomRect.topRight().x(),true);
            y->maxC = y->real(_zoomRect.topRight().y(),true);
            _isZoomed = true;

            updateGraph();
        }
        _isZooming = false;
    }
    if(_isMarking) {
        _markRect = QRect(_markRect.topLeft(), e->pos()).normalized();

        if( _markRect.width() > 0 && _markRect.height() > 0)
        {
            FSMarker m(x->real(_markRect.bottomLeft().x()),
                       y->real(_markRect.bottomLeft().y()),
                       x->real(_markRect.topRight().x()),
                       y->real(_markRect.topRight().y())   );
            if(x->logAxis) {
                m.x1 = pow(10.,m.x1);
                m.x2 = pow(10.,m.x2);
            }
            if(y->logAxis) {
                m.y1 = pow(10.,m.y1);
                m.y2 = pow(10.,m.y2);
            }
            appendMarker(m.x1,m.y1,m.x2,m.y2);
            updateGraph();
        }

        _isMarking = false;

    }
}

void FSGraph::keyPressEvent( QKeyEvent * e )
{
    if(readOnlyGraph) {
        switch (e->key()) {
        case Qt::Key_M:
            clearMarkers();
            break;
        case Qt::Key_U:
            unZoom();
            break;            
        case Qt::Key_X:
            toggleXAxis();
            break;
        case Qt::Key_Y:
            toggleYAxis();
            break;
        case Qt::Key_T:
            if(selectedDataSet) clipboardSelectedDataSet();
            break;
        case Qt::Key_I:
            clipboardGraph();
            break;
        default:
            e->ignore();
            return;
        }
        e->accept();
        return;
    }

    int sign = (e->modifiers()&Qt::ControlModifier)?-1:1;
    //QPoint p;

    switch (e->key()) {
    case Qt::Key_Plus:
        if(selectedDataSet) operation(sign*1);
        break;
    case Qt::Key_Minus:
        if(selectedDataSet) operation(sign*2);
        break;
    case Qt::Key_Asterisk:
        if(selectedDataSet) operation(sign*3);
        break;
    case Qt::Key_Slash:
        if(selectedDataSet) operation(sign*4);
        break;
    case Qt::Key_C:
        if(selectedDataSet) copySelectedDataSet();
        break;
    case Qt::Key_T:
        if(selectedDataSet) clipboardSelectedDataSet();
        break;
    case Qt::Key_I:
        clipboardGraph();
        break;
    case Qt::Key_E:
        if(selectedDataSet) editSelectedDataSet();
        break;
    case Qt::Key_R:
        if(selectedDataSet) sortSelectedDataSet();
        break;
    case Qt::Key_W:
        if(selectedDataSet) swapSelectedDataSet();
        break;
    case Qt::Key_F:
        if(selectedDataSet) fit();
        break;
    case Qt::Key_L:
        editLegend();
        break;
    case Qt::Key_U:
        unZoom();
        break;
    case Qt::Key_M:
        clearMarkers();
        break;
    case Qt::Key_X:
        toggleXAxis();
        break;
    case Qt::Key_Y:
        toggleYAxis();
        break;
    case Qt::Key_Delete:
    case Qt::Key_Backspace:
        if(selectedDataSet) deleteSelectedDataSet();
        break;
    case Qt::Key_Up:
        if(selectedDataSet && selectedDataSet->prevDataSet) {
            selectedDataSet = selectedDataSet->prevDataSet;
            update(); // ... so a rectangle is draw around this dataset 
        }
        break;
    case Qt::Key_Down:
        if(selectedDataSet && selectedDataSet->nextDataSet) {
            selectedDataSet = selectedDataSet->nextDataSet;
            update(); // ... so a rectangle is draw around this dataset 
        }
        break;
    default:
        e->ignore();
        return;
    }
    e->accept();
}


void FSGraph::keyReleaseEvent( QKeyEvent * e )
{
    e->ignore();
}

void FSGraph::zoom(double xmin, double xmax, double ymin, double ymax) {
    if(xmin<xmax && ymin<ymax) {
        x->minC = xmin;
        x->maxC = xmax;
        y->minC = ymin;
        y->maxC = ymax;
        _isZoomed = true;
        updateGraph();
    }
}



void FSGraph::unZoom( void )
{
    _isZoomed = false;
    updateGraph();
}

void FSGraph::editLegend( void )
{
    FSEditLegend * legEdit = new FSEditLegend(legendPen, legendFont, legendPos);
    if( legEdit->exec() )
    {
        legendPos = legEdit->pos();
        legendPen = legEdit->pen();
        legendFont = legEdit->font();
        updateGraph();
    }
}

void FSGraph::loadFile()
{
    loadFile(false);

}


void FSGraph::loadFile( bool clearAllDatasets )
{

    QString fileName;
    QStringList temp;

//  QString fileName( QFileDialog::getOpenFileName() );

    // fire up a FSFileDialog to get a file name and colums to read
    FSFileDialog fd(this);
    fd.setDirectory(QDir::current());
    if (fd.exec()) temp = fd.selectedFiles();
    if(temp.count()==0) return;
    fileName = temp[0];

    QString idName = fileName;
    int slash = fileName.lastIndexOf(QRegExp("[\\/\\\\]"));
    if(slash!=-1)
        idName = fileName.right( fileName.length()-slash-1 );

    FSDataSet* ds = new FSDataSet(idName, QPen(Qt::black,1));
    if(ds->loadfile(fileName, fd.Xcolumn(), fd.Ycolumn(), fd.skipLines())) {
        if(clearAllDatasets) clearDataSets();
        appendDataSet(ds);
    }
    updateGraph();
}



void FSGraph::editXAxis( void )
{
    bool extremaChanged;
    if(x->edit(&extremaChanged ))
    {
        if(extremaChanged)
            _isZoomed = true;
        updateGraph();
    }
}

void FSGraph::editYAxis( void )
{
    bool extremaChanged;
    if(y->edit(&extremaChanged ))
    {
        if(extremaChanged)
            _isZoomed = true;
        updateGraph();
    }
}

void FSGraph::editSelectedDataSet( void )
{
    if(selectedDataSet->edit())
        updateGraph();

}


void FSGraph::sortSelectedDataSet( void )
{
    selectedDataSet->sort();
    updateGraph();
}

void FSGraph::swapSelectedDataSet( void )
{
    selectedDataSet->swap();
    updateGraph();
}

void FSGraph::toggleXAxis( void )
{
    x->toggle();
    updateGraph();
}

void FSGraph::toggleYAxis( void )
{
    y->toggle();
    updateGraph();
}

void FSGraph::slotOp()
{
    operation(1);
}

//
// find which points of selectedDataSet are within markers
// insideMark[] must me already allocated to selectedDataSet->ndata
//
int FSGraph::isInsideMarker(bool *insideMark)
{
    int nInsideMark = 0, n = selectedDataSet->getndata();

    if(firstMarker) {
        for(int i=0; i<n; i++)
        {
            // go through all markers
            insideMark[i]=false;
            FSMarker* pmarker = firstMarker;
            while(pmarker)
            {
                if( selectedDataSet->xp[i] > pmarker->x1 &&
                    selectedDataSet->xp[i] < pmarker->x2 &&
                    selectedDataSet->yp[i] > pmarker->y1 &&
                    selectedDataSet->yp[i] < pmarker->y2 &&
                    !(x->logAxis && selectedDataSet->xp[i]<=0.) &&  // rejects values lower than 0 in log mode
                    !(y->logAxis && selectedDataSet->yp[i]<=0.) )

                {
                    insideMark[i] = true;  // inside at least on marker
                    nInsideMark++;
                    break; // one marker is enough, dont want to count a point twice in nInsideMark
                }
                pmarker = pmarker->nextMarker;
            }
        }
    }
    else {
        // all data if no marker defined
        for(int i=0; i<n; i++)
            insideMark[i]=true;

	 nInsideMark=n;
    }
    return nInsideMark;
}


void FSGraph::operation(int op)
{
    FSOpDialog * opDialog = new FSOpDialog(op, firstDataSet);
    if( opDialog->exec() )
    {
        FSDataSet *pdataset[1000];  // array of selected datasets for an operation with datasets
        opDialog->selectedDatasets(pdataset);

        double v = opDialog->constant();
        double *xx = NULL, *yy = NULL;
        bool onX = opDialog->onX();
        bool onY = opDialog->onY();
        int i, n = selectedDataSet->getndata();
        bool *insideMark;
        int returnVal = 0;
        op = opDialog->operation();

        if(op<0 && !pdataset[0]) return;  // operation with dataset but no dataset selected!

        // find which points of selectedDataSet are within markers
        insideMark = new bool[n];
        int nInsideMark = isInsideMarker(insideMark);
        if(nInsideMark==0) {
            delete [] insideMark;
            return;
        }

        // for operations with dataset, set a xx to interpolate
        int iInsideMark = 0;
        if(op<0)
        {
            xx = new double[nInsideMark]; // working variables
            yy = new double[nInsideMark];
            for(i=0; i<n; i++)
                if(insideMark[i])
                    xx[iInsideMark++]=selectedDataSet->xp[i];
        }

        int nd = 0;
        switch(op)
        {

        // operations with constant
        case 1:
            if(onX) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->xp[i]+=v;
            if(onY) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->yp[i]+=v;
            break;
        case 2:
            if(onX) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->xp[i]-=v;
            if(onY) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->yp[i]-=v;
            break;
        case 3:
            if(onX) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->xp[i]*=v;
            if(onY) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->yp[i]*=v;
            break;
        case 4:
            if(v!=0.) {
                if(onX) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->xp[i]/=v;
                if(onY) for(i=0; i<n; i++) if(insideMark[i]) selectedDataSet->yp[i]/=v;
            }
            break;
        // operations with datasets
        case -1:
            while (pdataset[nd]) {
               returnVal = pdataset[nd]->interpolate(nInsideMark, xx, yy, 0.);
               if(returnVal & 0x18) break;
               iInsideMark = 0;
               for(i=0; i<n; i++) {
                   if(insideMark[i]) selectedDataSet->yp[i]+=yy[iInsideMark];
                   iInsideMark++;
               }
               nd++;
            }
            break;
        case -2:
            while (pdataset[nd]) {
               returnVal = pdataset[nd]->interpolate(nInsideMark, xx, yy, 0.);
               if(returnVal & 0x18) break;
               iInsideMark = 0;
               for(i=0; i<n; i++) {
                   if(insideMark[i]) selectedDataSet->yp[i]-=yy[iInsideMark];
                   iInsideMark++;
               }
               nd++;
            } 
            break;
        case -3:
            while (pdataset[nd]) {
               returnVal = pdataset[nd]->interpolate(nInsideMark, xx, yy, 1.);
               if(returnVal & 0x18) break;
               iInsideMark = 0;
               for(i=0; i<n; i++) {
                   if(insideMark[i]) selectedDataSet->yp[i]*=yy[iInsideMark];
                   iInsideMark++;
               }
               nd++;
            } 
            break;
        case -4:
            while (pdataset[nd]) {
               returnVal = pdataset[nd]->interpolate(nInsideMark, xx, yy, 1.);
               if(returnVal & 0x18) break;
               iInsideMark = 0;
               for(i=0; i<n; i++) {
                   if(insideMark[i] && yy[iInsideMark]!=0.) selectedDataSet->yp[i]/=yy[iInsideMark];
                   iInsideMark++;
               }
               nd++;
            } 
            break;
        }

        if(returnVal)
        {
            QString s; // = "Return value: " + QString::number(returnVal) + "\n";
            if(returnVal & 0x01) s+= "\""+pdataset[nd-1]->legend + "\" sorted prior to interpolation\n";
            if(returnVal & 0x02) s+= "\""+selectedDataSet->legend + "\" has data defined outside \""+pdataset[nd-1]->legend + "\" range\n";
            if(returnVal & 0x04) s+= "\""+selectedDataSet->legend + "\" has no data\n";
            if(returnVal & 0x08) s+= "\""+pdataset[nd-1]->legend + "\" has no data\n";
            if(returnVal & 0x10) s+= "interpolation function called with bad parameters";
            QMessageBox::warning( 0, "Interpolation error", s);
        }

        updateGraph();
        if(op<0)
        {
            delete [] yy;
            delete [] xx;
        }

        delete insideMark;
        delete opDialog;
        selectedDataSet->dataHaveChanged = true;

    }
}

void FSGraph::fit( void )
{
    FSFitDialog * fitDialog = new FSFitDialog(nmarkers!=0);
    if( fitDialog->exec() )
    {
        int order = fitDialog->order();
        if(order<0) return;

        bool isPoly = fitDialog->fitType()==2;
        int n = selectedDataSet->getndata();
        bool *insideMark;
        insideMark = new bool[n];
        int nInsideMark = isInsideMarker(insideMark);
        if(nInsideMark==0)
        {
            QMessageBox::warning( 0, "fit error", "no data to fit");
            delete [] insideMark;
            return;
        }

        double *xx, *yy;
        int i, k;

        switch(fitDialog->fitType()) {

        // boxcar
        case 1:
            yy = new double[n]; // working variable
            for(i=0; i<n; i++)
                if(insideMark[i])
                {
                    int nav = 0;
                    yy[i] = 0.;
                    for(int j=-order/2; j<=(order+1)/2; j++)
                        if(i+j>=0 && i+j<n)
                        {
                            yy[i]+= selectedDataSet->yp[i+j];
                            nav+=1;
                        }
                     yy[i]/=(nav>0)?nav:1;
                }
                else
                    yy[i] = selectedDataSet->yp[i];

            for(i=0; i<n; i++)
                selectedDataSet->yp[i] = yy[i];

            delete[] yy;
            break;

        // spline
        case 3:
            if(order<4)
            {
                QMessageBox::warning( 0, "Spline fit error", "Order (number of nodes) must be larger than 3");
                delete [] insideMark;
                return;
            }
            // NO break

        // polynomial & spline
        case 2:
            if(isPoly) order++; // in the case of polynomial fit order means order+1 parameters
            if(nInsideMark<=order) {
                QMessageBox::warning( 0, "Fit error", "Number of points smaller or equal to fit order");
               delete [] insideMark;
            return;
            }

            // (x,y) values used for regression
            xx = new double[nInsideMark];
            yy = new double[nInsideMark];
            int iInsideMark=0;
            for(i=0; i<n; i++)
            {
                if(insideMark[i])
                {
                    xx[iInsideMark]=selectedDataSet->xp[i];
                    yy[iInsideMark]=selectedDataSet->yp[i];
                    if(x->logAxis) xx[iInsideMark] = log10(xx[iInsideMark]); // fit made on log of data if log axis
                    if(y->logAxis) yy[iInsideMark] = log10(yy[iInsideMark]);
                    iInsideMark++;
                }
            }

            //normalize the problem
            double xmin = xx[0], xmax = xx[0];
            double ymin = yy[0], ymax = yy[0];
            for(i=1; i<nInsideMark; i++)
            {
                if(xx[i]<xmin) xmin=xx[i];
                if(xx[i]>xmax) xmax=xx[i];
                if(yy[i]<ymin) ymin=yy[i];
                if(yy[i]>ymax) ymax=yy[i];
            }
            if(xmin==xmax) {
                QMessageBox::warning( 0, "Fit error", "All data have the same x value");
                delete [] insideMark;
                return;
            }
            xmax*=1.0001;
            for(i=0; i<nInsideMark; i++)
            {
                xx[i] = (xx[i]-xmin)/(xmax-xmin);
                yy[i] = (yy[i]-ymin)/(ymax-ymin);
            }

            // create and initialize arrays for QR decomposition
            double **QR, *B, *X;
            QR = new double*[order];
            B = new double[order];
            for(i=0; i<order; i++) {
                QR[i] = new double[order];
                B[i] = 0.;
                for(int j=0; j<order; j++)
                        QR[i][j] = 0.;
            }

            // set array values
            for(k=0; k<nInsideMark; k++)
            {
                if(isPoly)
                    for(int i=0; i<order; i++)
                    {
                        B[i] += yy[k]*pow(xx[k],i);
                        for(int j=0; j<order; j++)
                            QR[i][j] += pow(xx[k],i+j); // polynomial
                    }
                else
                {
                    double s[4];
                    int si = spline(order-2, xx[k], s);
                    for(int i=si; i<si+4; i++)
                    {
                        B[i] += yy[k]*s[i-si];
                        for(int j=si; j<si+4 && j<order; j++)
                            QR[i][j] += s[i-si]*s[j-si]; // spline
                    }
                }
            }
            delete[] xx;
            delete[] yy;

            // solve QR X = B for X
            QRDecomposition *qrd = new QRDecomposition(QR, order);
            X = qrd->solve(B);

            // little clean up
            for(i=0; i<order; i++)
                delete[] QR[i];
            delete[] QR;
            delete[] B;

            // create a new dataset containing the fit
            FSDataSet* ds = new FSDataSet(selectedDataSet->legend + "_fit", QPen(Qt::blue,1));
            bool in = fitDialog->isInterpolate();
            bool ex = fitDialog->isExtrapolate();
            double x1 = 0;
            double x2 = 0;

            bool isMark = nmarkers>0;
            if(isMark) {
                // find the markers x extrema
                x1 = firstMarker->x1; // initial values
                x2 = firstMarker->x1;

                FSMarker* pmarker = firstMarker;
                do {
                    if(x1>pmarker->x1) x1 = pmarker->x1;
                    if(x1>pmarker->x2) x1 = pmarker->x2;
                    if(x2<pmarker->x1) x2 = pmarker->x1;
                    if(x2<pmarker->x2) x2 = pmarker->x2;
                    pmarker = pmarker->nextMarker;
                } while(pmarker);
            }

            for(k=0; k<n; k++)
            {
              double xxx = selectedDataSet->xp[k];
              if(!isMark || ex || (in && xxx>=x1 && xxx <= x2) || insideMark[k]) {
                if(x->logAxis) xxx = log10(xxx); // fit made on log of data if log axis
                xxx = (xxx-xmin)/(xmax-xmin);

                double val = 0.;
                if(isPoly)
                    for(int i=0; i<order; i++)
                        val+=X[i]*pow(xxx,i);
                else
                {
                    double s[4];
                    int si = spline(order-2, xxx, s);
                    for(int i=si; i<si+4; i++)
                        val+=X[i]*s[i-si];
                }
                xxx = xxx*(xmax-xmin)+xmin; // de-normalize
                val = val*(ymax-ymin)+ymin;
                if(x->logAxis) xxx = pow(10.,xxx); // fit was made on log data if axis is log
                if(y->logAxis) val = pow(10.,val);

                ds->append(xxx,val);
              }
            }
            appendDataSet(ds);

            delete[] X;

            break;
        }

        updateGraph();
        delete [] insideMark;
    }
}

FSDataSet * FSGraph::yProjectionOfSelectedDataset(double binSize)
{
    int n = selectedDataSet->getndata();

    // find which points of selectedDataSet are within markers
    bool * insideMark = new bool[n];
    int nInsideMark = isInsideMarker(insideMark);

    // if nothing return nothing
    if(nInsideMark==0) {
        delete [] insideMark;
        return NULL;
    }


    // create bins to accumulate counts
    selectedDataSet->extrema();  // make sure we know the extrema
    double max = selectedDataSet->getyMax();
    double min = selectedDataSet->getyMin();
    if(min>=max) {
        delete [] insideMark;
        return NULL;  // this should never append, just in case
    }

    if(binSize<=0) binSize = 1;
    int nbins = int((max-min)/binSize);  // calculate the number of bins
    int * bins = new int[nbins+1];  // one more bin than intervals

    int i;
    // initialize bins
    for(i=0; i<=nbins; i++)
        bins[i] = 0;

    // compute frequecy
    for(i=0; i<n; i++)
        if(insideMark[i])
            bins[int((selectedDataSet->yp[i]-min)/binSize+0.5)]++;

    delete [] insideMark;

    // create a new dataset that will contain the projection
    FSDataSet* ds = new FSDataSet(selectedDataSet->getlegend()+" Y proj.", QPen(Qt::red,1));
    for(i=0; i<nbins; i++)
        ds->append(i*binSize+min, bins[i]/binSize);

    delete [] bins;
    return ds;

}

// call a graph update with graph repaint
void FSGraph::updateGraph() {

    hasChanged = true;
    update();    // will call paintEvent
}
