/*
    Copyright 2007-2015, Francois Schiettekatte

    This file is part of Corteo.

    Corteo is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Corteo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Corteo.  If not, see <http://www.gnu.org/licenses/>.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#ifdef __APPLE__
    #include <sys/malloc.h>
#else
    #include <malloc.h>
#endif

#include "constants.h"
#include "corteoindex.h"
#include "corteomatrix.h"
#include "corteoutil.h"
#include "randomx.h"
#include "techniques.h"

#ifdef MAX_THREAD
  #if defined(_WIN64) || defined(_WIN32)
    //#include "winpthreads.h"
    #include <pthread.h>

  #else
    #include <pthread.h>
  #endif
    pthread_mutex_t mut;
#endif

#define MAXLAYERS  20   // number of target layers; absorber is an extra layer
#define MAXELEMENT 10   // number of elements in each layer

#define NUMDET     2    // number of detectors
#define NUMBINPROF 1000 // number of bins to store implantation and recoils depth profiles, and energy-depth map
#define NUMMINCOSSCAT 5 // number of largest scattering events (smallest cos(theta_scat)) to store during a slowdown (full RBS)

// prime numbers for lists length
#define MAXLOGLIST 104729
#define MAXAZILIST 72211
#define MAXRANLIST 100003

// inverf list loaded from file, fixed length
#define MAXERFLIST 79999

// size of file names and other char buffers
#define FILENAMESIZE 1000

// global variables (faster than passing arguments)

// layer parameters
unsigned int numberLayers;                   // number of layers (or materials in case of bitmap)
unsigned int numberElements[MAXLAYERS+1];    // number of elements in each layer
unsigned int isBitmap;                       // 1 if target is a described by a bitmap, 0 if it is a layer structure

float       layerThickness0[MAXLAYERS+1];    // nominal thickness of each layer [A]
float       layerThickDist[MAXLAYERS+1];     // width of thickness distribution of each layer [A]
float       layerDensity[MAXLAYERS+1];       // density of each layer [A^{-3}]
float       interatomicDist[MAXLAYERS+1];    // (3/(4 pi density))^(1/3) [A]
float       sqrtLayerDensityPi[MAXLAYERS+1]; // (density*pi))^(1/2) [A]
float       targetBottom;

float       elementFraction[MAXLAYERS+1][MAXELEMENT];    // relative concentration of element in layer
float       elementFractionSum[MAXLAYERS+1][MAXELEMENT]; // cumulative concentration of elements in layer

unsigned long xDim, yDim, zDim; // dimensions of bitmap (in pixels)
float pixelsPerAngs, AngsPerPixel;  // conversion factors between pixels and Angstroms
float bitmapThichness;          // thichness (x direction) of the bitmap in Angstroms
int substrateLayer;             // layer that is the substrate on top of which the bitmap sits, -1 if none
unsigned char *bitmap;          // bitmap
char *filenameWithoutExt;       // filename to which .in, .bitmap, .detect, etc. will be added to read the input or save the results

// rather than calling lengthy functions, corteo uses large lists of precalculated values for certain functions
float       randomlist[MAXRANLIST];     // list of evenly distributed but randomly ordered values between 0 and 1
float       sqrtrandomlist[MAXRANLIST]; // sqrt of randomlist
float       sqrtloglist[MAXLOGLIST];    // list of evenly distributed but randomly ordered values of sqrt of -log of 1/MAXLOGLIST to 1
float       sqrtloglist1[MAXLOGLIST];   // 1/sqrtloglist
float       sinAzimAngle[MAXAZILIST];   // list cos and sin components of angles...
float       cosAzimAngle[MAXAZILIST];   //   ...this angle are evenly distributed but randomly ordered between 0 and 2*PI
float       erflist[MAXERFLIST];        // list of evenly distributed but randomly ordered values of inverse erf of -1+1/MAXERFLIST to 1-1/MAXERFLIST

float       *hsp[MAXLAYERS+1];    // hydrogen stopping power table in each layer

// initial beam energy, minimum energy to follow a projectile (eV) 
float E0, minE0;

// sqrt of mean free path ([A]^0.5)
float sqrtMeanFreePath[MAXLAYERS+1];
float sinMinDeflAngle; // sin of minimum deflection angle
int doRotate;

// Factor by which the histogram is mltiplied: density*thickness*solidAngle/number_of_detected_particle, with  
//   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle, 
//   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)  
// The spectrum then needs (externally) to be multipllied by the number of incident ions q.
// See documentation for a more detailed explanation.
float factorNtOmegaByNumdetpart;

// projectile incident directional cosines
float l0, m0, n0;

// spot size (sigma of a gaussian distribution)
float ySpotSize, zSpotSize;

unsigned int  enerStragType; // type of energy straggling computation: 0=none, 1=Brag, 2=Chu, 3=Wang

// counters
unsigned long detectedParticles, exitBottom, belowCutoff, exitUndetected, badRescale, collVacuum,
              exitUndetectedFinal, exitTop, exitMinEnergy, badKinematics, moreThanTwoMaterials;

unsigned long numProjectile; // global number of projectiles (for ion implantation, full RBS)

// parameters for the series of IBA simulations of a target
unsigned long numProjectileLayer[MAXLAYERS]; // number of projectiles for each layer
unsigned int nsim[MAXLAYERS][MAXELEMENT]; // number of isotopes to simulate for a layer
float abundance[MAXLAYERS][MAXELEMENT][MAXELEMENT]; // relative abundance of an isotope (numProjectileLayer*abundance ions simulated)
float massIsotope[MAXLAYERS][MAXELEMENT][MAXELEMENT]; // mass of the simulated isotope

int CScorrType[MAXLAYERS][MAXELEMENT][MAXELEMENT]; // cross-section correction type: 0 = none, 1 = ratio to Rutherford, 2 = value from table
int currentCScorrType; // variable in which the current value in the table above is stored for the current element/layer simulation

unsigned int natural[MAXLAYERS][MAXELEMENT]; // =1 if simulating natural isotopic abundance of an element, 0 if the user specified the abundance(s)
unsigned int isNatural; // variable in which the current value of table above is stored

unsigned int nthread;

// coordinates of 3 points defining a detector in space
float       p1x[NUMDET], p1y[NUMDET], p1z[NUMDET];
float       p2x[NUMDET], p2y[NUMDET], p2z[NUMDET];
float       p3x[NUMDET], p3y[NUMDET], p3z[NUMDET];

// coordinates of points 1 and 3 in detector reference frame
float       p1yp[NUMDET], p1zp[NUMDET];
float       p3yp[NUMDET], p3zp[NUMDET]; 

// coordinates of detector center, factor d in plane equation
float       pcx[NUMDET], pcy[NUMDET], pcz[NUMDET], d[NUMDET];
float       pcnormx[NUMDET], pcnormy[NUMDET], pcnormz[NUMDET];

// components of normal vectors xp, yp, zp defining detectors reference frame 
float       xpx[NUMDET], xpy[NUMDET], xpz[NUMDET];
float       ypx[NUMDET], ypy[NUMDET], ypz[NUMDET]; 
float       zpx[NUMDET], zpy[NUMDET], zpz[NUMDET];


// variables indicating if detector is an ellipse
int         isEllipse[NUMDET]; 

// inverse of detector half-size in each direction, annulus radius squared
// (in case of ellipse, this is the inverse of major and minor axis, a & b)
float       ainv[NUMDET], binv[NUMDET], r2annular[NUMDET]; 

// factor in width (W: parallel to p3-p2) and height (H: parallel to p2-p1) 
// by which a virtual detector is larger than the real detector
float       Wf, Hf;

// angle of the cone around detector axis in which scattered event are generated
float       cosAngSpread0; 

// cos of the minimum angle beyond which collision is assumed to have missed
// (required because collision near 0 degrees contribute an enormous cross section in RBS)
float       coscutoffAngle;

// cross section screening: 0: none, 1: Andersen, 2:considering Universal potential, 3: Lenz-Jensen
unsigned int screen;

// simulation type
int simulationType;

// follow (!=0) or do not follow (==0) recoils during implantation simulation
int followRecoils;

FILE *ofdetect;
int writeDetails;  // if writeDetails!=0, spits lots of info about main collision and detection
                   //  and in the case of full RBS, the NUMMINCOSSCAT smallest values of cos(scat_angle) and corresp. depth

// absorber thickness, 0 is no absorber
float absorberThick;

// parameters for Chu straggling correction
float Chu[100][4];

unsigned int *energy_depth_map;

// thread-related parameters
typedef struct ThreadParams {
  int id;
  unsigned long numProjectile;
  unsigned long totalnumbercoll;
  unsigned int collisionLayer;
  //float * histogram;
} thread_params;


// ion parameters
typedef struct IonParams {
  unsigned int Z;
  float mass;
  int roundMass;

  float a[MAXLAYERS+1][MAXELEMENT];            // screening length
  float f[MAXLAYERS+1][MAXELEMENT];            // reduced energy conversion factor
  float mr[MAXLAYERS+1][MAXELEMENT];           // mass ratio
  float K[MAXLAYERS+1][MAXELEMENT];            // kinematic factor
  float smax[MAXLAYERS+1][MAXELEMENT];         // maximum reduced impact parameter
  float smaxE[MAXLAYERS+1][MAXELEMENT][DIME];  // maximum reduced impact parameter vs epsilon ...
       // ... i.e. reduced impact parameter that produce the minimum deflection angle at a given reduced energy

  // pointers to tables of cos and sin of scattering angle
  float *CosScat[MAXLAYERS+1][MAXELEMENT];     
  float *SinScat[MAXLAYERS+1][MAXELEMENT];

  float  *sp[MAXLAYERS+1];        // stopping power table (computed according to Bragg's rule with compound correction)
  float  *sqrtStrag[MAXLAYERS+1]; // sqrt of energy straggling * sqrt(2) (computed according to Bragg's rule)
  float compoundCorr[MAXLAYERS+1];

  int tablesComputed;

} ion_params;

ion_params beam;   // beam parameters

ion_params *recoil;   // collision partner parameters
// VERY IMPORTANT: *recoil is only a pointer to an element of the target. Do not allocate.

// pointer to element parameters (implantation following recoils, same address for same Z, mass)
ion_params *element[MAXLAYERS+1][MAXELEMENT]; 

double *beamImplHist;  // pointer to beam implantation profile
double *implHist[NUMBERELEMENTS];  // pointer to implantation profile of recoils per element
double *dispHist[NUMBERELEMENTS];  // pointer to displacements profile of recoils per element

// displacement and binding energy (for implantation, when following recoils)
// these are declatated outside ion_params structure because they can be different for different layers
// while the rest of the structure is identical for elements with same Z and mass whatever the layer is
float Ed[MAXLAYERS][MAXELEMENT]; 
float Eb[MAXLAYERS][MAXELEMENT]; 

// fill the table of cos and sin of the scattering angle in the lab frame given a mass ratio
void fillCosSinTable(float * cosTable, float * sinTable, float mr) {

  double sin2thetaby2, costheta, costhetaLab, sinthetaLab;
  unsigned long i, j;

  // compute scattering angle components
  for(i=0; i<DIME*DIMS; i++) {
    sin2thetaby2 = Matrix(i);
    costheta = 1.-2.*sin2thetaby2;

    if(costheta==-1.0 && mr==1.0) {
      costhetaLab = 0.0;  // peculiar case of head-on collision of identical masses
      sinthetaLab = 1.0;
    } else {
      costhetaLab = (costheta+mr)/sqrt(1.+2.*mr*costheta+mr*mr);
      if(costhetaLab<-1.0) costhetaLab = -1.0; //roundoff errors some times...
      if(costhetaLab>1.0) costhetaLab = 1.0;  // causes sinthetaLab to be NAN
      sinthetaLab = sqrt(1.-costhetaLab*costhetaLab);
    }
    cosTable[i] = d2f(costhetaLab);
    sinTable[i] = d2f(sinthetaLab);
  }

}


// load a list of inverse error function erfinv(x) values that contains MAXERFLIST elements
// the list is such that x is uniformly distributed between -1 and 1 excluding these boundaries
// i.e. x = -1+dx, -1+2dx, ...1-2dx, 1-dx, with 1/2dx = MAXERFLIST
// return 0 on error, 1 otherwise
int loadErfInv() {
  unsigned int k;
  char erfval[FILENAMESIZE];
  FILE * ifp;

  ifp = fopen("data/erfinv.dat", "r");
  if(ifp==NULL) {
    fprintf(stderr,"\nCan't open data/erdinv.dat ");
    return 0;
  }

  for(k=0; k<MAXERFLIST; k++) {
    fscanf(ifp,"%s", erfval);
    erflist[k] = a2f(erfval);
  }
  fscanf(ifp,"%s", erfval);
  fclose(ifp);

  // control value at the end should be the number of elements in the list
  if(atoi(erfval)!=MAXERFLIST) return 0; 
  return 1;
}


// load Chu straggling correction parameters
// return 0 on error, 1 otherwise
int loadChu() {
  unsigned int k, l, Z;
  char Chuval[FILENAMESIZE];
  FILE * ifp;

  ifp = fopen("data/chu.dat", "r");
  if(ifp==NULL) {
    fprintf(stderr,"\nCan't open data/chu.dat ");
    return 0;
  }

  ignoreline(ifp); // skip header line
  for(k=0; k<92; k++) {
    fscanf(ifp,"%u", &Z);
    for(l=0; l<4; l++) {
      fscanf(ifp,"%s", Chuval);
      Chu[Z][l] = a2f(Chuval);
    }
    if(Z!=k+2) // no data for Z=0,1, table starts at Z=2
      return 0;
  }
  fclose(ifp);
  return 1;
}

// - load dE/dx data for projectile particleZ in each element present in layer ilayer; 
//     - file is presumed to be in the right format, written by generateDedxTable with the right parameters
//     - control data should be wrong if something is wrong
// - compute stopping accoring to Bragg's rule;
// - result is stored in spp[]
// - return the adress of allocated memory if successfull; 
// - return NULL on error: stopping already set, can't open file or bad control data
// IMPORTANT: stopping for the most abundant isotope (energy for index calculation must be corrected for mass)
float * fillDedxTable(unsigned int ilayer, unsigned int projectileZ, float projectileMass, float dens, float compoundCorr) {
  char filename[FILENAMESIZE];
  unsigned int i, k, elemZ;
  float f, buf[DIMD+1];
  FILE *ifp;
  char *mode = "rb";

  float * spp = calloc(DIMD,sizeof(float));  // allocate while initializing to 0
  if(spp == NULL) {
    fprintf(stderr,"\nStopping power allocation error");
    return NULL;
  }

  sprintf(filename, "data/%u.asp",projectileZ);

  for(i=0; i<numberElements[ilayer]; i++) {
    elemZ = element[ilayer][i]->Z;
    
    // open file and move to record position
    ifp = fopen(filename, mode);
    if(ifp==NULL) {
      fprintf(stderr,"\nCan't open %s ", filename);
      return NULL;
    }
    fseek(ifp,sizeof(float)*(DIMD+1)*(elemZ-1),1);
    fread((void*)buf, sizeof(float), DIMD+1, ifp); // read dedx data plus control data
    fclose(ifp);
    
    if(buf[DIMD]!=elemZ) { // each record must end with the element number as a control data
      fprintf(stderr, "Error reading stopping power for particle Z=%u in material Z=%u\n",projectileZ, elemZ);
      return NULL; 
    }

    for(k=0; k<DIMD; k++) {
      // electronic stopping of index k; factor 10 is convertion from eV/(1E15 at/cm2) to eV/(at/A2), density is in at/A3
      spp[k]+=buf[k]*elementFraction[ilayer][i]*10.0f*dens; // Bragg's rule
    }
  }


  // compound correction according to Zeigler & Manoyan NIMB35(1998)215, Eq.(16) (error in Eq. 14)
  if(compoundCorr!=1.0f)
    for(k=0; k<DIMD; k++) {
      f = d2f( 1.0/(1.0+exp( 1.48*( sqrt(2.*Dval(k)/projectileMass/25e3) -7.0) )) );
      spp[k]*=f*(compoundCorr-1.0f)+1.0f;
    }
  return spp;
}

// compute energy straggling of projectile (to be multiplied by distance in Angstroems during calculations)
// - compute straggling accoring to Bragg's rule;
// - compute for most abundant isotope mass
// - when stragType 
//          = 0: straggling = 0
//          = 1: Bohr's straggling
//          = 2: Chu correction
//          = 3: Chu + Wang correction
// - return the allocated table or NULL
// IMPORTANT: stopping for the most abundant isotope (energy for index calculation must be corrected for mass)
float * fillStragglingTable(unsigned int ilayer, unsigned int projectileZ, float stopping[], float hstopping[], int stragType) {
  unsigned int i, targetZ;
  unsigned long ii, k;
  double chargestate, E, EE, S;
  double omegaBohr2, chucorr, epsi, gamma, extra, str, sstrag; 
  double MAImass = MostAbundantIsotope[projectileZ];

  float * strag = calloc(DIMD,sizeof(float));  // allocate while initializing to 0
  if(strag == NULL) {
    fprintf(stderr,"\nStopping power allocation error");
    return NULL;
  }

  for(k=0; k<DIMD; k++) {
    sstrag=0.;

    for(i=0; i<numberElements[ilayer]; i++) {
      targetZ = element[ilayer][i]->Z;

      S = stopping[k];
      E = Dval(k); // energy of index k

      // effective charge state squared is S(H)/S(ion) Z_ion^2 for S at same speed
      ii = Dindex(d2f(E/MAImass));
      chargestate = S/(hstopping[ii]*projectileZ*projectileZ);

      omegaBohr2 = 4.*PI*projectileZ*projectileZ*E2*E2*targetZ*layerDensity[ilayer];

      EE = E/MAImass/1.e6;
      chucorr = 1./(1.+Chu[targetZ][0]*pow(EE,Chu[targetZ][1])+Chu[targetZ][2]*pow(EE,Chu[targetZ][3]));

      epsi = EE/((double)projectileZ*sqrt((double)projectileZ*(double)targetZ)); 

      if(projectileZ==1)
        gamma = 2.522*(1.-exp(-1.040*EE));
      else
        gamma = 0.3931*(1.-exp(-3.812*epsi));

      if(projectileZ==1) 
        extra = 0.1955*gamma/(pow(EE-0.6941,2)+gamma*gamma);
      else
        extra = (double)projectileZ*pow((double)projectileZ/(double)targetZ,1./3.)*1.273e-2*gamma/(pow(epsi-3.458e-2,2)+gamma*gamma);

      switch(stragType) {
      case 0: // no straggling
      default:
        str = 0.;
        break;
      case 1: // Bohr's
        str = omegaBohr2;
        break;
      case 2: // Chu correction only
        str = omegaBohr2*chucorr;
        break;
      case 3: // Chu + Wang correction
        str = omegaBohr2*(chargestate*chucorr+extra);
      }
      sstrag += str*elementFraction[ilayer][i]; // Bragg's rule;
    }
    strag[k] = sqrtf(sstrag)*sqrtf(2.);
  }
  return strag;
}



int fillTables(ion_params *ion) {
  unsigned int ilayer, ielem, i, j, ii;

  if(ion==NULL) return 2;
  if(ion->tablesComputed) return 1;

  for(ilayer=0; ilayer<=MAXLAYERS; ilayer++) {
    for(ielem=0; ielem<MAXELEMENT; ielem++) {
      if(element[ilayer][ielem]) {

        // screaning length
        ion->a[ilayer][ielem] = d2f(SCREENCONST/(pow(element[ilayer][ielem]->Z,0.23)+pow(ion->Z,  0.23)));

        // mass ratio
        ion->mr[ilayer][ielem] = ion->mass/element[ilayer][ielem]->mass;  // m1/m2 
      
        // mass part of the kinematic factor
        ion->K[ilayer][ielem] = 4.0f*ion->mr[ilayer][ielem]/((1.0f+ion->mr[ilayer][ielem])*(1.0f+ion->mr[ilayer][ielem]));

        // reduced energy conversion factor
        ion->f[ilayer][ielem]  = ion->a[ilayer][ielem]/((ion->mr[ilayer][ielem]+1.0f) * element[ilayer][ielem]->Z*ion->Z *E2);

        // cos and sin components of Matrix  
        if(ion->CosScat[ilayer][ielem]==NULL) ion->CosScat[ilayer][ielem] = (float *) malloc(DIME*DIMS*sizeof(float));
        if(ion->SinScat[ilayer][ielem]==NULL) ion->SinScat[ilayer][ielem] = (float *) malloc(DIME*DIMS*sizeof(float));

        if(ion->CosScat[ilayer][ielem]==NULL || ion->SinScat[ilayer][ielem] == NULL) {
          fprintf(stderr, "Matrices memory allocation error.\n");
          exit(1);
        }  
        fillCosSinTable(ion->CosScat[ilayer][ielem], ion->SinScat[ilayer][ielem], ion->mr[ilayer][ielem]);

        // maximum reduced impact parameter
        ion->smax[ilayer][ielem] = 1.0f/(ion->a[ilayer][ielem]*sqrtf(PI*layerDensity[ilayer])*sqrtMeanFreePath[ilayer]);

        // find reduced impact parameter producing a [ minimum deflection angle | minimum energy transfer ] at each reduced energy
        if(sinMinDeflAngle) {
            for(i=0; i<DIME; i++) { // for each reduiced energy...
                for(j=DIMS-1; j>0; j--) {  // start from the largest reduced impact parameter...
                    //if(sinTable[i*DIMS+j]>=sinMinDeflAngle) // minimum angle
                    // E_reduced / f * K * sin^2(th/2) = transfered energy in eV at that energy for given impact parameter
                    if(Eval(i)/ion->f[ilayer][ielem]*ion->K[ilayer][ielem]*Matrix(i*DIMS+j)>=sinMinDeflAngle)  // sinMinDeflAngle = minimum energy transfer in eV
                        break; // break when sin(angle for impact parameter and energy)>sin(minimum deflection angle)
                }
                ion->smaxE[ilayer][ielem][i] = Sval(j); // record the reduced impact parameter that produce the [ minimum deflection angle | energy transfer ] at that reduced energy
                if(ion->smaxE[ilayer][ielem][i]*ion->a[ilayer][ielem] > interatomicDist[ilayer])
                    ion->smaxE[ilayer][ielem][i] = interatomicDist[ilayer] / ion->a[ilayer][ielem];
            }
        }


      }
    }  

    if(numberElements[ilayer]) {
      // fill stopping power tables for this layer
      ion->sp[ilayer] = fillDedxTable(ilayer, ion->Z, ion->mass, layerDensity[ilayer], ion->compoundCorr[ilayer] );
      if(!ion->sp[ilayer]) {
        fprintf(stderr, "\nError: can't load stopping power table for Z=%u\n", ion->Z);
        exit(1);
      }  

      // fill straggling table for this layer
      ion->sqrtStrag[ilayer] = fillStragglingTable(ilayer, ion->Z, ion->sp[ilayer], hsp[ilayer], enerStragType);
      if(!ion->sp[ilayer]) {
       exit(1);
      } 
    }
  }
  return 0;
}
  

// generate and randomize lists randomlist & sqrtloglist, sinAzimAngle & cosAzimAngle
void computelists() {
  unsigned int irandomlist, iloglist, iazimAngle;
  // generate a uniformly spaced list of values between .5/MAXRANLIST and 1-.5/MAXRANLIST
  for(irandomlist=0; irandomlist<MAXRANLIST; irandomlist++)
    randomlist[irandomlist] = (irandomlist+0.5f)/MAXRANLIST;
  randomizelist(randomlist, MAXRANLIST); // put the list in random order
  for(irandomlist=0; irandomlist<MAXRANLIST; irandomlist++)
    sqrtrandomlist[irandomlist] = sqrtf(randomlist[irandomlist]); // also compute sqrt of these values
 
  // produce a list of MAXLOGLIST sqrt(-log(x)) values for x uniformly distributed 
  // ...between x=.5/MAXLOGLIST and x=1-.5/MAXLOGLIST
  for(iloglist=0; iloglist<MAXLOGLIST; iloglist++)
    sqrtloglist[iloglist] = sqrtf(-log((iloglist+.5)/MAXLOGLIST));
  randomizelist(sqrtloglist, MAXLOGLIST); // put the list in random order
  // precompute 1/sqrtloglist[]
  for(iloglist=0; iloglist<MAXLOGLIST; iloglist++)
    sqrtloglist1[iloglist] = 1.0f/sqrtloglist[iloglist];
 
  // produce a list uniformly distributed but randomly ordered azimutal angles  
  for(iazimAngle=0; iazimAngle<MAXAZILIST; iazimAngle++)
    // cosAzimAngle temporarly contains angles
    cosAzimAngle[iazimAngle]= 2.0f*(float)PI*(float)(iazimAngle)/(float)(MAXAZILIST);
  randomizelist(cosAzimAngle, MAXAZILIST); // put the list in random order
  for(iazimAngle=0; iazimAngle<MAXAZILIST; iazimAngle++) {
    // compute the cos and sine of these angles
    sinAzimAngle[iazimAngle] = d2f(sin(cosAzimAngle[iazimAngle]));
    cosAzimAngle[iazimAngle] = d2f(cos(cosAzimAngle[iazimAngle]));
  }

}


// Compute new directional cosines from cos & sin components of scattering angle theta
// and a random azimutal angle (no trigonometric functions called!)
//
// **** this function is critical in terms of computational efficiency ****
//
void rotate(float *l, float *m, float *n, unsigned int * iazimAngle, float costheta, float sintheta) {
  
  float k, kk, kinv;
  float ll=*l, mm=*m, nn=*n;
  float k2 = 1.0f-nn*nn;

  // random azimutal rotation angle components
  float cosomega = cosAzimAngle[*iazimAngle];
  float sinomega = sinAzimAngle[*iazimAngle];
  if(++*iazimAngle==MAXAZILIST) *iazimAngle = 0;

  if(k2==0) {
    // extreemly rare case
    *n = costheta;
    *m = sintheta*cosomega;
    *l = sintheta*sinomega;
    return;
  } 

  // usual case
  kinv = 1.0f/sqrtf(k2); // using sqrtf here almost as fast as using tables, and more precise
  k = k2*kinv;

  kk = sintheta*kinv;
  *l = ll * costheta+kk*(ll*nn*cosomega+mm*sinomega);
  *m = mm * costheta+kk*(mm*nn*cosomega-ll*sinomega);
  *n = nn * costheta-k*sintheta*cosomega;

}

// in full RBS with writeDetails != 0, find if cosScat is among the NUMMINCOSSCAT largest scattering
// event (i.e. smallest cosScat) and store it in sorted order of increasing value
//
// WARNING: if minCosScat!=NULL, CALLED FOR EACH INDIVIDUAL COLLISION, MUST BE EFFICIENT
//
int storeMinCosScat(float cosScat, float depth, float *minCosScat, float * minCosScatDepth) {
    int i,j;

    for(i = 0; i<NUMMINCOSSCAT; i++) {
        // if current value is smaller than i_th value of minCosScat...
        if(cosScat<minCosScat[i]) {
            // ... push minCosScat & minCosScatDepth values one spot up
            for(j = i+1; j<NUMMINCOSSCAT; j++) {
                minCosScat[j] = minCosScat[j-1];
                minCosScatDepth[j] = minCosScatDepth[j-1];
            }
            // and replace values at spot i by the new values
            minCosScat[i] = cosScat;
            minCosScatDepth[i] = depth;
            return 1;
        }
        // hence, at the end of the process, minCosScat contains the smallest values of cosScat in increasing order
        // and minCosScatDepth their corresponding depth
    }
    return 0;
}



// Compute *ion slowdown to 'depth' in the target, taking into account stopping power, energy straggling and multiple collisions.
// *x is the depth (*x = 0 at surface and *x>0 inside the target)
// For an incomming particle, *x < depth and its initial value should be *x<0 but very small, e.g. *x = -1e-10
// Conversely, for an outgoing particle, depth < *x and depth should be < 0 but very small, e.g. depth = -1e-10.
// isTarget !=0 for particle in target, ==0 for particle in absorber/detector (layer numberLayers)
//
// **** this function is critical in terms of computational efficiency ****
//
// Return:
//  0 for normal function end (depth>0 and *x>depth or depth<0 and *x<0)
//  1 if particle exited at bottom of target (e.g. outgoing particle reached target bottom)
// -1 if particle abnormally exited at top (e.g. incoming incident particle deflected outide target top)
// -2 if minimum energy is reached
// -3 when exiting from infinite loop (>1 000 000 collisions)
int slowdown(float *x, float *y, float *z, float *l, float *m, float *n, double *E,
             float depth, int isTarget, unsigned long *numcol, unsigned int *irandomlist,
             unsigned int *iloglist, unsigned int *ierflist, unsigned int *iazimangle,
             ion_params *ion, int level, float *layerTop, float *layerBottom, float *minCosScat, float *minCosScatDepth) {

    int thisLevel = level+1; // level of recursion in the case of full cascade
    float fl, fl2, temp, temp2; // flight length, working variables
    float sqrtfl, ran, fraction;  // sqrt(fl), random number for collision partner, fraction of fl in current layer
    float S, sstr;       // computed stopping and straggling
    float ss;            // reduced impact parrameter, working variable

    unsigned int ilayer; // current layer; ilayer = numberLayers for absorber
    unsigned int ielem;  // collision partner

    unsigned long i;     // matrix index
    unsigned long ncol = 0;   // local number of collisions
    unsigned int bin, binE, iE, isEnerDepth;

    float sin2thetaby2;
    float amuRatio = MostAbundantIsotope[ion->Z]/ion->mass; // energy/velocity correction for stopping evaluation

    // these parameters are used when following recoils during implantation
    float lr, mr, nr, xr, yr, zr; // flying direction and position
    float lbc, mbc, nbc; // ion flying direction before collision (bc)
    double Er, Erb, Edd, minE=minE0; // recoil energy, binding and displacement energy
    ion_params * elem;
    int exitStatus;
    float sqrand, sqrmaxranlist1000 = sqrtf(1000.0f/MAXRANLIST); // sqrt(random) that is 1/1000 of max value
    float ssmax, bbmax;

    isEnerDepth = (simulationType == TECH_ENERDEPTHMAP || simulationType == TECH_ENERDEPTHLIST) ? 1 : 0;
    ilayer = isTarget?0:numberLayers; // index of current layer; ilayer = numberLayers for absorber

    while(*E>minE) {
        if(isTarget) { // if it's not target, it's absorber, and ilayer = numberLayers

            // find in which layer we are
            if(*x<layerTop[ilayer]) {
                do {
                    if(ilayer==0) {
                        *numcol+=ncol;
                        return -1; // exit at top
                    }
                    ilayer--;
                } while(*x<layerTop[ilayer]);
            } else if(*x>=layerBottom[ilayer]) {
                do {
                    ilayer++;
                    if(ilayer>=numberLayers) {
                        *numcol+=ncol;
                        return 1; // exit at bottom
                    }
                } while(*x>layerBottom[ilayer]);
            }
        }

        // randomly pick a partner for collision
        if(*irandomlist>=MAXRANLIST-2) *irandomlist = 0;
        ran = randomlist[(*irandomlist)++];
        for(ielem=0;ielem<numberElements[ilayer]; ielem++)
            if(elementFractionSum[ilayer][ielem]>=ran)
                break;

        // pick an impact parameter (LSS units) within maximum value at current energy for minimum deflection
        // (if the number from sqrtrandomlist is in the 1/1000th smallest values, generate a number using randomx;
        //  needed to reproduce wide angle scattering without granularity induced by discretisation effect of the list)
        iE = Eindex(*E* ion->f[ilayer][ielem]);
        sqrand = sqrtrandomlist[(*irandomlist)++];
        if(sqrand<sqrmaxranlist1000) {
            // for the 1000 smallest values, use instead a random number generator
            sqrand = sqrtf((float)randomx())*sqrmaxranlist1000;
        }


        // (below, note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist] used to set sqrtfl)
        if(sinMinDeflAngle) {
            ssmax = ion->smaxE[ilayer][ielem][iE];
            ss = sqrand*ssmax;
            //bb = ss * ion->a[ilayer][ielem]; // impact parameter in Angs
            bbmax = ssmax * ion->a[ilayer][ielem]; // max impact parameter in Angs

            // estimate sqrt of flight length according to Poisson's probability
            // (fl0 = 1/(density * pi * smax^2), fl = ln(r)*fl0
            sqrtfl  = sqrtloglist[*iloglist]/(sqrtLayerDensityPi[ilayer]*bbmax); // or bb?
        } else {
            // estimate sqrt of flight length according to Poisson's probability
            sqrtfl  = sqrtloglist[*iloglist]*sqrtMeanFreePath[ilayer];

            // (below, note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist] used to set sqrtfl)
            ss = sqrand*sqrtloglist1[*iloglist]*ion->smax[ilayer][ielem];
            //ss = sqrtrandomlist[(*irandomlist)++]*sqrtloglist1[*iloglist]*ion->smax[ilayer][ielem]; // original version
            //ss = sqrtf(randomx())*smax/sqrtf(-log(randomx())); // using this makes program 3 times slower!
        }

        fl = sqrtfl*sqrtfl;

   /*
        // pick an impact parameter (LSS units)
        // (if the number from sqrtrandomlist is in the 1/1000th smallest values, generate a number using randomx;
        //  needed to reproduce wide angle scattering without granularity induced by discretisation effect of the list)
        sqrand = sqrtrandomlist[(*irandomlist)++];
        if(sqrand<sqrmaxranlist1000) {
            // for the 1000 smallest values, use instead a random number generator
            sqrand = sqrtf((float)randomx())*sqrmaxranlist1000;
        }
        // (below, note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist] used to set sqrtfl)
        ss = sqrand*sqrtloglist1[*iloglist]*ion->smax[ilayer][ielem];
        //ss = sqrtrandomlist[(*irandomlist)++]*sqrtloglist1[*iloglist]*ion->smax[ilayer][ielem]; // original version
        //ss = sqrtf(randomx())*smax/sqrtf(-log(randomx())); // using this makes program 3 times slower!
  */

        if(++*iloglist>=MAXLOGLIST) *iloglist=0;

        // compute collision
        if(ss * ion->a[ilayer][ielem]>=interatomicDist[ilayer]) {
            // impact parameter larger than interatomic distance: assume collision has missed
            sin2thetaby2 = 0.0f;
            Er = 0;
        } else { // compute collision

            // get matrix index
            i = iE*DIMS+Sindex(ss);
            // energy transfer to recoil
            sin2thetaby2 = Matrix(i);
            Er = *E*ion->K[ilayer][ielem]*sin2thetaby2;

            if(followRecoils) {
                // store flying direction before collision
                lbc = *l;
                mbc = *m;
                nbc = *n;

                // rotate ion trajectory
                if(doRotate) rotate(l, m, n, iazimangle, ion->CosScat[ilayer][ielem][i], ion->SinScat[ilayer][ielem][i]);

                elem = element[ilayer][ielem];
                Edd = Ed[ilayer][ielem];
                if(Er>2.0f*Edd) {  // Er > 2Ed: generate a new recoil

                    // compute recoil flying direction
                    temp  = sqrtf(*E/Er * ion->mr[ilayer][ielem]);      // *********VERIFIER SI BON RAPPORT mr ***********
                    temp2 = sqrtf((*E-Er)/ *E );
                    lr = temp*(lbc - *l*temp2);
                    mr = temp*(mbc - *m*temp2);
                    nr = temp*(nbc - *n*temp2);

                    xr = *x;
                    yr = *y;
                    zr = *z;

                    // increment displacements histogram
                    bin = (unsigned int) (xr/targetBottom*NUMBINPROF);
                    if(bin<NUMBINPROF)
                        dispHist[elem->Z][bin]++;

                    // follow the generated recoil (recursive function call!)
                    Erb = Er - Eb[ilayer][ielem]; // subtract binding energy
                    exitStatus = slowdown(&xr, &yr, &zr, &lr, &mr, &nr, &Erb, depth, 1,
                                          &ncol, irandomlist, iloglist, ierflist, iazimangle,
                                          elem, thisLevel, layerTop, layerBottom, NULL, NULL);

                    switch(exitStatus) {
                    case 1:
                        exitBottom++;
                        break;
                    case -1:
                        exitTop++;
                        break;
                    case -2:
                        //minimum energy: normal function end for implanted ions and generated recoils
                        bin = (unsigned int) (xr/targetBottom*NUMBINPROF);
                        if(bin<NUMBINPROF)
                            implHist[elem->Z][bin]++;
                        break;
                    default:
                        // Unexpected but extreemly rare outcome due to a rotate() call with [l m n] = [0 0 1]
                        fprintf(stdout, "*");
                        break;
                    }

                } else if(Er>Edd) { // Ed < Er < 2Ed
                    // ion has just enough energy to diplace the recoil, but we then assume that the ion stops here
                    // (we wont't enter the replacement collision game)
                    bin = (unsigned int) (xr/targetBottom*NUMBINPROF);
                    if(bin<NUMBINPROF) {
                        implHist[elem->Z][bin]++;
                        dispHist[elem->Z][bin]++;
                    }
                }
            } else {
                // not following recoils, just rotate ion trajectory
                if(doRotate) {
                    rotate(l, m, n, iazimangle, ion->CosScat[ilayer][ielem][i], ion->SinScat[ilayer][ielem][i]);
                    if(minCosScat) {
                        if(writeDetails) // find if in NUMMINCOSSCAT smallest values
                            storeMinCosScat(ion->CosScat[ilayer][ielem][i], *x, minCosScat, minCosScatDepth);
                        else // only keep smallest value
                            if(ion->CosScat[ilayer][ielem][i]<minCosScat[0]) {
                                minCosScat[0] = ion->CosScat[ilayer][ielem][i];
                                minCosScatDepth[0] = *x;
                            }
                    }
                }
            }
            if(doRotate) *E -= Er;
        }

        // readjust flight length so we don't go beyond depth
        temp = *l*depth;
        if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
                && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
            fl = (depth-*x)/ *l*1.001f;     // readjust flight length so projectile flights slightly over depth
        }

        // electronic stopping
        i = Dindex(*E*amuRatio); // energy corrected for velocity in dE/dx calculation
        S = ion->sp[ilayer][i]*fl;

        // reajust stopping if we cross a layer boundary
        temp = *x+*l*fl;
        if(*l>0 && temp>layerBottom[ilayer]) { // bottom-oriented projectile crossing layer bottom
            fraction = (layerBottom[ilayer]-*x)/(*l*fl);
            S *= fraction; // fraction of stopping in this layer

            if(ilayer<numberLayers-1) { // fl and stopping in layer below, if any
                fl2 = fl*(1.0f-fraction); // rest of flight length
                fl -= fl2; // from fl subtract rest of flight length
                fl2 *= layerDensity[ilayer]/layerDensity[ilayer+1]; // correct rest of flight length for density
                fl += fl2; // readd corrected rest of flight length to fl
                temp = *x+*l*fl; // we changed fl, so recompute x+l*fl
                S += ion->sp[ilayer+1][i]*fl2; // stopping along rest of flight length in layer below
            }
        } else if(*l<0 && temp<layerTop[ilayer]) { // top-oriented projectile crossing layer top
            fraction = (layerTop[ilayer]-*x)/(*l*fl);  // (l<0 and layerTop-x<0 also so signs cancel)
            S *= fraction;

            if(ilayer>0 && ilayer<numberLayers-1) { // fl and stopping in layer above, if any
                fl2 = fl*(1.0f-fraction); // rest of flight length
                fl -= fl2; // from fl subtract rest of flight length
                fl2 *= layerDensity[ilayer]/layerDensity[ilayer+1]; // correct rest of flight length for density
                fl += fl2; // readd corrected rest of flight length to fl
                temp = *x+*l*fl; // we changed fl, so recompute x+l*fl

                S += ion->sp[ilayer-1][i]*fl2; // stopping along rest of flight length in layer above
            }
        }

        // advance to collision site
        *x = temp; // +=fl* *l;
        *y+=fl* *m;
        *z+=fl* *n;

        // compute energy straggling and subtract electronic energy loss
        sstr = ion->sqrtStrag[ilayer][i]*sqrtfl*erflist[(*ierflist)++];  // energy straggling; not reajusted, already a 1st order correction
        if(*ierflist>=MAXERFLIST) *ierflist=0;
        *E -= S + sstr;

        //fprintf(stderr, "\n%g %g %g %f %f %f %f %f %f %f %f",*E,sin2thetaby2,S,fl, sstr,*x,*y,*z,*l,*m,*n);
        if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
            *numcol+=ncol;
            if(*E<minE) return -2; // last collision made energy lower than minimum energy
            return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
        }
        if( ncol++  > 1000000000 || *E>E0*1.5) {
            fprintf(stdout, "?");
            *numcol+=ncol;
            return -3;
        }

        if(isEnerDepth) {
            if(simulationType == TECH_ENERDEPTHMAP) {
                if(*x>=0 && *x<targetBottom && *E<E0 && *E>=0) {
                    bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
                    binE = (unsigned int) (*E/E0*NUMBINPROF);
                    if(bin<NUMBINPROF && binE<NUMBINPROF)
                        energy_depth_map[bin*NUMBINPROF+binE]++;
                }
            } else if(simulationType == TECH_ENERDEPTHLIST) {
#ifdef MAX_THREAD
                pthread_mutex_lock(&mut);
#endif
                fprintf(ofdetect, "%e %e", *E, *x);
                if(writeDetails) {
                    fprintf(ofdetect, " %e %e  %e %e %e", *y, *z, *l, *m, *n);
                }
                fprintf(ofdetect, "\n");
                fflush(ofdetect);

#ifdef MAX_THREAD
                pthread_mutex_unlock(&mut);
#endif
            }

        }

    }
    *numcol+=ncol;
    if(simulationType == TECH_IMPLANT) {
        bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
        if(bin<NUMBINPROF)
            beamImplHist[bin]++;
    }
    return -2; // we have reach minimum energy (normal function end for implantation)
}


// slowdown in a single, monoatomic layer
int quickslowdown(float *x, float *y, float *z, float *l, float *m, float *n, double *E,
             float depth, int isTarget, unsigned long *numcol, unsigned int *irandomlist, 
             unsigned int *iloglist, unsigned int *ierflist, unsigned int *iazimangle,
             float amuRatio, float interDist, float top, float bottom, 
             float a, float f, float K, float smax, float * CosScat, float * SinScat, 
             float *sp, float *sqrtStrag, float sqrtMFP, float layerTop, float layerBottom,
             float *minCosScat, float *minCosScatDepth ) {

  float fl, temp; // flight length, working variables
  float sqrtfl, fraction;  // sqrt(fl), fraction of fl in current layer
  float S, sstr;           // computed stopping and straggling
  float ss;        // reduced impact parrameter
  float Er = 0;
  float sqrand, sqrmaxranlist1000 = sqrtf(1000.0f/MAXRANLIST); // sqrt(random) that is 1/1000 of max value

  unsigned long i;     // matrix index
  unsigned long ncol = 0;

  float sin2thetaby2;

  unsigned int bin, binE, isEnerDepth;
  isEnerDepth = (simulationType == TECH_ENERDEPTHMAP || simulationType == TECH_ENERDEPTHLIST) ? 1 : 0;

  while(*E>minE0) {
      
    if(*x<top) {
      *numcol+=ncol;
      return -1; // exit at top
    } else if(*x>=bottom) { 
      *numcol+=ncol;
      return 1; // exit at bottom
    }

    // estimate sqrt of flight length according to Poisson's probability 
    sqrtfl  = sqrtloglist[*iloglist]*sqrtMFP;
    fl = sqrtfl*sqrtfl;

    // readjust flight length so we don't go beyond depth
    temp = *l*depth;
    if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
      && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
      fl = (depth-*x)/ *l*1.0001f;     // readjust flight length so projectile flights slightly over depth
    }

    // electronic stopping
    i = Dindex(*E*amuRatio); // energy corrected for velocity in dE/dx calculation
    S = sp[i]*fl;

    // reajust stopping if we cross a layer boundary
    temp = *x+*l*fl;
    if(temp>bottom) { // bottom-oriented projectile crossing layer bottom
      fraction = (bottom-*x)/(*l*fl);
      S *= fraction; // fraction of stopping in this layer
    } else if(temp<top) { // top-oriented projectile crossing layer top 
      fraction = (top-*x)/(*l*fl);  // (l<0 and layerTop-x<0 also so signs cancel)
      S *= fraction; 
    }

    // advance to the next collision site
    *x = temp; // +=fl* *l;
    *y+=fl* *m;
    *z+=fl* *n;

    // energy straggling
    sstr = sqrtStrag[i]*sqrtfl*erflist[(*ierflist)++];  // energy straggling; not reajusted, already a 1st order correction
    if(*ierflist>=MAXERFLIST) *ierflist=0; 

    // subtract electronic energy losses, including energy straggling
    *E -= S + sstr;

    // pick an impact parameter (LSS units) (note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist])
    sqrand = sqrtrandomlist[(*irandomlist)++];
    if(sqrand<sqrmaxranlist1000) {
        // for the 1000 smallest values, use instead a random number generator
        sqrand = sqrtf((float)randomx())*sqrmaxranlist1000;
    }
    ss = sqrand*sqrtloglist1[*iloglist]*smax;
    if(++*iloglist>=MAXLOGLIST) *iloglist=0;
    if(*irandomlist>=MAXRANLIST) *irandomlist=0;

    if(ss*a>=interDist) {
      // impact parameter larger than interatomic distance: assume collision has missed
      sin2thetaby2 = 0.0f;
      Er = 0;

    } else {      // compute collision
      // compute matrix index
      i = Eindex(*E*f)*DIMS+Sindex(ss);

      // rotate directional cosines with precomputed cos and sin values of scattering angle
      sin2thetaby2 = Matrix(i);
      Er = *E*K*sin2thetaby2;       // recoil energy

      // not following recoils, just rotate ion trajectory
      if(doRotate) rotate(l, m, n, iazimangle, CosScat[i], SinScat[i]);
      if(minCosScat) {
          if(writeDetails) // find if in NUMMINCOSSCAT smallest values
              storeMinCosScat(CosScat[i], *x, minCosScat, minCosScatDepth);
          else // only keep smallest value
              if(CosScat[i]<minCosScat[0]) {
                  minCosScat[0] = CosScat[i];
                  minCosScatDepth[0] = *x;
              }
      }
    }

    // subtract nuclear energy loss
    *E -= Er;

    if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
      *numcol+=ncol;
      if(*E<minE0) return -2; // last collision made energy lower than minimum energy
      return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
    }
    if( ncol++  > 1000000) {
        fprintf(stdout, "?");
        *numcol+=ncol;
        return -3;
    }

    if(isEnerDepth) {
        if(simulationType == TECH_ENERDEPTHMAP) {
            if(*x>=0 && *x<targetBottom && *E<E0 && *E>=0) {
                bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
                binE = (unsigned int) (*E/E0*NUMBINPROF);
                if(bin<NUMBINPROF && binE<NUMBINPROF)
                    energy_depth_map[bin*NUMBINPROF+binE]++;
            }
        } else if(simulationType == TECH_ENERDEPTHLIST) {
#ifdef MAX_THREAD
            pthread_mutex_lock(&mut);
#endif
            fprintf(ofdetect, "%e %e", *E, *x);
            if(writeDetails) {
                fprintf(ofdetect, " %e %e  %e %e %e", *y, *z, *l, *m, *n);
            }
            fprintf(ofdetect, "\n");
            fflush(ofdetect);

#ifdef MAX_THREAD
            pthread_mutex_unlock(&mut);
#endif
        }
    }
  }
  *numcol+=ncol;
  if(simulationType == TECH_IMPLANT) {
      bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
      if(bin<NUMBINPROF)
          beamImplHist[bin]++;
  }
  return -2; // we have reach minimum energy (normal function end for implantation)
}


// get pixel map indices for coords x, y, z
// return:
// 0 if particle is in bitmap
// 1 if it is exiting at top
// 2 if it is exiting at bottom
// 3 if it is not in target but directed towards it
// 4 if it is in substrate layer
// (*iy, *iz are always set to something reliable by virtue of boundary conditions, but *ix may not be usable)
int getIndices(float x, float y, float z, float l, unsigned long *ix, unsigned long *iy, unsigned long *iz) {

    // initial checks (and sanity checks to ensure conversion to 'long' are safe)
    if(x<0) {
        if(l<0) {
            return 1; // exit at top
        } else {
            return 3; // particle not in bitmap, but directed towards it: consider vaccuum
        }
    }

    if(x > bitmapThichness) {
        if(substrateLayer>=0) {
            if(x > bitmapThichness+layerThickness0[substrateLayer]) {
                if(l>0) {
                    return 2; // exit at bottom
                } else {
                    return 3; // particle not in bitmap, but directed towards it: consider vaccuum
                }
            } else {
                return 4; // in substrate layer
            }
        } else {
            if(l>0) {
                return 2; // exit at bottom
            } else {
                return 3; // particle not in bitmap, but directed towards it: consider vaccuum
            }

        }
    }

    // at this point, we are sure that 0 <= x*pixelsPerAngs <= xDim, discretisation safe
    *ix = (long)(x * pixelsPerAngs);

    // periodic boundary conditions along lateral directions
    // (conversion should be safe because we have initialized y,z with a value smaller than (yDim,zDim)/pixelsPerAngs)
    *iy = ((unsigned long)(y * pixelsPerAngs)) % yDim;

    if(zDim>1) {
        // case of a 3D pixel map
        *iz = ((unsigned long)(z * pixelsPerAngs)) % zDim; // periodic boundary conditions along lateral directions
    } else
        // case of a 2D image, and we will save a couple of operations
        *iz = 0;

    return 0;

}



// slowdown in a target described by a pixel map in which 255 = vaccum, 0 = material of layer 0, 1 = material of layer 1, etc.
// If material changes between two collison, stopping is only approximately corrected by taking
// the average stopping value of 5 points along the trajectory
int pixmap_slowdown(float *x, float *y, float *z, float *l, float *m, float *n, double *E,
             float depth, unsigned long *numcol, unsigned int *irandomlist,
             unsigned int *iloglist, unsigned int *ierflist, unsigned int *iazimangle,
             ion_params *ion, float *minCosScat, float *minCosScatDepth) {

    float fl, temp, fl_2ndPart; // flight length, working variables
    float sqrtfl, ran;  // sqrt(fl), random number for collision partner, fraction of fl in current layer
    float S, sstr;       // computed stopping and straggling
    float ss;            // reduced impact parrameter, working variable

    unsigned char imaterial, im, imm; // materials (= layer number)
    int retval;

    unsigned int ielem;  // collision partner

    unsigned long i, ii, fl_pix;  // matrix index, flight length in pixels
    unsigned long ncol = 0;   // local number of collisions
    unsigned int bin, binE, iE, isEnerDepth;

    isEnerDepth = (simulationType == TECH_ENERDEPTHMAP || simulationType == TECH_ENERDEPTHLIST) ? 1 : 0;

    double Er, minE=minE0; // recoil energy

    float sin2thetaby2;
    float amuRatio = MostAbundantIsotope[ion->Z]/ion->mass; // energy/velocity correction for stopping evaluation

    float sqrand, sqrmaxranlist1000 = sqrtf(1000.0f/MAXRANLIST); // sqrt(random) that is 1/1000 of max value
    float ssmax, bbmax;
    float dx, dy, dz, fraction;

    unsigned long ix, iy, iz, iii, bitmapDim = zDim*yDim*xDim;

    // slowdown loop
    while(*E>minE) {

        retval = getIndices(*x, *y, *z, *l, &ix, &iy, &iz);

        if(retval) {
            switch (retval) {
            case 1:
                return -1; // exit at top
            case 2:
                return 1; // exit at bottom
            case 3:
                imaterial = UCHAR_MAX; // vacuum (were are outside target but directed towards it)
                break;
            case 4:
                imaterial = substrateLayer; // we are in the substrate
                break;
            default:
                fprintf(stderr,"\nUnknown return code %u from getIndices.\n", retval);
                exit(1);
            }
        } else {
            iii = (iz*yDim+iy)*xDim+ix;
            if(iii>=bitmapDim) {
                // something went wrong...
                fprintf(stdout, "?");
                *numcol+=ncol;
                return -3;
            }
            imaterial = bitmap[iii];
        }

        //printf("\n01234 %g %g %g %g %li %li %li %u\n", *E, *x, *y, *z, ix, iy, iz, imaterial);

        if(imaterial==UCHAR_MAX) {
            dx = *l*AngsPerPixel; // step size of one pixel
            dy = *m*AngsPerPixel;
            dz = *n*AngsPerPixel;
            while(imaterial==UCHAR_MAX) {
                // we are in vacuum, advance position until we reach some material
                *x+=dx;
                *y+=dy;
                *z+=dz;
                retval = getIndices(*x, *y, *z, *l, &ix, &iy, &iz);
                switch (retval) {
                case 0:
                    imaterial = bitmap[(iz*yDim+iy)*xDim+ix];
                    break;
                case 1:
                    if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
                        *numcol+=ncol;
                        if(*E<minE) return -2; // last collision made energy lower than minimum energy
                        return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
                    }
                    return -1; // exit at top
                case 2:
                    if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
                        *numcol+=ncol;
                        if(*E<minE) return -2; // last collision made energy lower than minimum energy
                        return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
                    }
                    return 1; // exit at bottom
                case 3:
                    imaterial = UCHAR_MAX; // vacuum (were are outside target but directed towards it)
                    break;
                case 4:
                    imaterial = substrateLayer; // we are in the substrate
                    break;
                default:
                    fprintf(stderr,"\nUnknown return code %u from getIndices.\n", retval);
                    exit(1);
                }
            }
            if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
                *numcol+=ncol;
                if(*E<minE) return -2; // last collision made energy lower than minimum energy
                return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
            }

        }

        // if we reach this point, imaterial should be something else than vacuum
        if(imaterial == UCHAR_MAX)
            printf("?");

        // randomly pick a partner for collision
        if(*irandomlist>=MAXRANLIST-2) *irandomlist = 0;
        ran = randomlist[(*irandomlist)++];
        for(ielem=0;ielem<numberElements[imaterial]; ielem++)
            if(elementFractionSum[imaterial][ielem]>=ran)
                break;

        // pick an impact parameter (LSS units) within maximum value at current energy for minimum deflection
        // (if the number from sqrtrandomlist is in the 1/1000th smallest values, generate a number using randomx;
        //  needed to reproduce wide angle scattering without granularity induced by discretisation effect of the list)
        iE = Eindex(*E* ion->f[imaterial][ielem]);
        sqrand = sqrtrandomlist[(*irandomlist)++];
        if(sqrand<sqrmaxranlist1000) {
            // for the 1000 smallest values, use instead a random number generator
            sqrand = sqrtf((float)randomx())*sqrmaxranlist1000;
        }

        // (below, note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist] used to set sqrtfl)
        if(sinMinDeflAngle) {
            // in this version, find s first, with s < smax(E), so as to provide a minimum deflection
            ssmax = ion->smaxE[imaterial][ielem][iE];
            ss = sqrand*ssmax;
            //bb = ss * ion->a[ilayer][ielem]; // impact parameter in Angs
            bbmax = ssmax * ion->a[imaterial][ielem]; // max impact parameter in Angs

            // then, using smax(E), estimate sqrt of flight length according to Poisson's probability
            // (fl0 = 1/(density * pi * smax^2), fl = ln(r)*fl0
            sqrtfl  = sqrtloglist[*iloglist]/(sqrtLayerDensityPi[imaterial]*bbmax); // or bb?
        } else {
            // in this version, estimate sqrt of flight length according to Poisson's probability
            // given a mean free path provided by the user
            sqrtfl  = sqrtloglist[*iloglist]*sqrtMeanFreePath[imaterial];

            // (below, note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist] used to set sqrtfl)
            // here, smax was previouly set based on density and mean free path provided by the user
            ss = sqrand*sqrtloglist1[*iloglist]*ion->smax[imaterial][ielem];
            //ss = sqrtrandomlist[(*irandomlist)++]*sqrtloglist1[*iloglist]*ion->smax[ilayer][ielem]; // original version
            //ss = sqrtf(randomx())*smax/sqrtf(-log(randomx())); // using this makes program 3 times slower!
        }
        fl = sqrtfl*sqrtfl;

        if(++*iloglist>=MAXLOGLIST) *iloglist=0;

        // compute collision
        if(ss * ion->a[imaterial][ielem]>=interatomicDist[imaterial]) {
            // impact parameter larger than interatomic distance: assume collision has missed
            sin2thetaby2 = 0.0f;
            Er = 0;
        } else { // compute collision

            // get matrix index
            i = iE*DIMS+Sindex(ss);
            // energy transfer to recoil
            sin2thetaby2 = Matrix(i);
            Er = *E*ion->K[imaterial][ielem]*sin2thetaby2;

            // rotate ion trajectory
            if(doRotate) {
                rotate(l, m, n, iazimangle, ion->CosScat[imaterial][ielem][i], ion->SinScat[imaterial][ielem][i]);
                if(minCosScat) {
                    if(writeDetails) // find if in NUMMINCOSSCAT smallest values
                        storeMinCosScat(ion->CosScat[imaterial][ielem][i], *x, minCosScat, minCosScatDepth);
                    else // only keep smallest value
                        if(ion->CosScat[imaterial][ielem][i]<minCosScat[0]) {
                            minCosScat[0] = ion->CosScat[imaterial][ielem][i];
                            minCosScatDepth[0] = *x;
                        }
                }

                *E -= Er;
            }
        }

        // readjust flight length so we don't go beyond depth
        //fl_2ndPart = fl;
        temp = *l*depth;
        if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
                && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
            fl = (depth-*x)/ *l*1.001f;     // readjust flight length so projectile flights slightly over depth
        }

        // electronic stopping
        i = Dindex(*E*amuRatio); // energy corrected for velocity in dE/dx calculation
        S = ion->sp[imaterial][i]*fl;
        // *** IMPORTANT: DON'T CHANGE i, MAY BE NEEDED BELOW

        // find if after flying the distance fl we are in a new material
        // (*** It must be understood that if fl is large in comparison to target features
        //  we may wrongly suppose that we travelled in a certain material while we actually went
        //  through 2 or more. The user must select a fl that is smaller than the features he wants
        //  to simulate)
        retval = getIndices(*x+*l*fl, *y+*m*fl, *x+*n*fl, *l, &ix, &iy, &iz);
        if(retval==4)
            im = substrateLayer; // we are in substrate (ix, iy, iz not set, don't use them)
        else if(retval)
            im = UCHAR_MAX; // all other non-0 values = vacuum (ix, iy, iz not set, don't use them)
        else {
            iii = (iz*yDim+iy)*xDim+ix;
            if(iii>=bitmapDim) {
                // something went wrong...
                fprintf(stdout, "?");
                *numcol+=ncol;
                return -3;
            }
            im = bitmap[iii];
        }

        // reajust stopping and flight length if we end up in another materal
        if(im != imaterial) {
            // We are not in the same material. We will get the value of the pixels along the trajectory
            // in order to find the fraction of the trajectory we spent in each material. If we meet a third
            // material (different from imaterial and im) along the way, we will not consider it, and increment
            // a warning counter.

            fraction = 1.; // fraction will be the fraction of fl that is in 'imaterial'
            // at least the first pixel is in 'imaterial'...

            fl_pix = (unsigned long)(fl*pixelsPerAngs+0.5);  // fl in pixels
            dx = *l*AngsPerPixel; // step size will be one pixel, so we will be looking each pixel along the way
            dy = *m*AngsPerPixel;
            dz = *n*AngsPerPixel;

            for(ii=1; ii<fl_pix; ii++) { // ... and we scan from the second pixel
                retval = getIndices(*x+ii*dx, *y+ii*dy, *x+ii*dz, *l, &ix, &iy, &iz);
                if(retval == 4)
                    imm = substrateLayer; // we are in substrate (ix, iy, iz not set, don't use them)
                else if(retval)
                    imm = UCHAR_MAX; // all other non-0 values = vacuum (ix, iy, iz not set, don't use them)
                else
                    imm = bitmap[(iz*yDim+iy)*xDim+ix];

                if(imm==imaterial)
                    fraction++;
                else if(imm!=im) {
                    moreThanTwoMaterials++;
                    // we are in another material than im or imaterial
                    // probably a not to bad approximation to consider we are half into 'imaterial'.
                    // Otherwise, if we do not increment 'fraction', it is as if we would be considering that this
                    // other material 'imm' is 100% 'im'. (Comuputing time cost of a more detailed computation of
                    // stopping vs effect on a spectrum remains to be evaluated)
                    fraction+=0.5f;
                }
                // else, imm==im and we don't want to increment fraction

            }
             // normalise fraction to the number of steps
            if(fl_pix)
                fraction /= fl_pix;
            else
                fraction = 1;

            // finally, adjust stopping for fraction of trajectory spent in each material
            S *= fraction; // adjust the stopping that occured in 'imaterial'
            if(fraction< 1.0 && im!=UCHAR_MAX) {
                // compute fl in the part of the trajectory that is in material 'im'
                // and ajust it to account for changes in density (so it may end up shorter or
                // longer, but we will not carry out 2nd order corrections)
                fl_2ndPart = fl*(1.0f-fraction)*layerDensity[imaterial]/layerDensity[im];
                S += ion->sp[im][i]*fl_2ndPart;
                fl = fl*fraction+fl_2ndPart;

                // readjust flight length so we don't go beyond depth
                temp = *l*depth;
                if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
                        && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
                    fl = (depth-*x)/ *l*1.001f;     // readjust flight length so projectile flights slightly over depth
                }
            }
            // else, we are in vacuum and there is no stopping!

        }

        // advance to collision site
        *x+=fl* *l;
        *y+=fl* *m;
        *z+=fl* *n;

        // compute energy straggling and subtract electronic energy loss
        sstr = ion->sqrtStrag[imaterial][i]*sqrtfl*erflist[(*ierflist)++];  // energy straggling; not reajusted, already a 2nd order correction
        if(*ierflist>=MAXERFLIST) *ierflist=0;
        *E -= S + sstr;

        imaterial = im;

        //fprintf(stderr, "\n%g %g %g %f %f %f %f %f %f %f %f",*E,sin2thetaby2,S,fl, sstr,*x,*y,*z,*l,*m,*n);
        if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
            *numcol+=ncol;
            if(*E<minE) return -2; // last collision made energy lower than minimum energy
            return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
        }
        if( ncol++  > 1000000000) {
            fprintf(stdout, "?");
            *numcol+=ncol;
            return -3;
        }

        if(isEnerDepth) {
            if(simulationType == TECH_ENERDEPTHMAP) {
                if(*x>=0 && *x<targetBottom && *E<E0 && *E>=0) {
                    bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
                    binE = (unsigned int) (*E/E0*NUMBINPROF);
                    if(bin<NUMBINPROF && binE<NUMBINPROF)
                        energy_depth_map[bin*NUMBINPROF+binE]++;
                }
            } else if(simulationType == TECH_ENERDEPTHLIST) {
#ifdef MAX_THREAD
                pthread_mutex_lock(&mut);
#endif
                fprintf(ofdetect, "%e %e", *E, *x);
                if(writeDetails) {
                    fprintf(ofdetect, " %e %e  %e %e %e", *y, *z, *l, *m, *n);
                }
                fprintf(ofdetect, "\n");
                fflush(ofdetect);

#ifdef MAX_THREAD
                pthread_mutex_unlock(&mut);
#endif
            }
        }

    }
    *numcol+=ncol;
    if(simulationType == TECH_IMPLANT) {
        bin = (unsigned int) (*x/targetBottom*NUMBINPROF);
        if(bin<NUMBINPROF)
            beamImplHist[bin]++;
    }
    return -2; // we have reach minimum energy (normal function end for implantation)
}




// Find a new traveling direction within a cone of angle angSpread 
// around the line relating sample origin and detector "det" origin
void mainCollision(int det, float *cosx, float *cosy, float *cosz) {
  // Pic a random tilting angle uniformly distributed in terms of its cos component
  // (cosAngSpread randomly but uniformly distributed between cosAngSpread0 and 1)

  float cosAngSpread = (1.0f-cosAngSpread0)*(float)randomx()+cosAngSpread0; 
  float sinAngSpread = -sqrtf(1-cosAngSpread*cosAngSpread);
  unsigned int iazim = (unsigned int)(MAXAZILIST*randomx());

  // new flying direction: direction of detector central point
  *cosx = pcnormx[det];
  *cosy = pcnormy[det];
  *cosz = pcnormz[det];

  // tilt this vector off its axis by an angle between angleSpread0 and angleSpread1
  // and rotate it by a random azimuthal angle
  rotate(cosx, cosy, cosz, &iazim, cosAngSpread, sinAngSpread);
  //printf("99    %f %f\n", *cosx*pcnormx[det]+ *cosy*pcnormy[det]+ *cosz*pcnormz[det], atan((*cosz-pcnormz[det])/(*cosx-pcnormx[det])) );
  //printf("99   %f %f %f %f\n", *cosx*pcnormx[det]+ *cosy*pcnormy[det]+ *cosz*pcnormz[det],*cosx-pcnormx[det], *cosy-pcnormy[det], *cosz-pcnormz[det]   );
}


// compute kinematic factor for scattering collision
// return 0 if collision forbiden by kinematics, 1 otherwise
// value of kinematic factor stored in *kinfact
int kinematicFactorRBS(float massRatio,
  float cosxin,  float cosyin,  float coszin,
  float cosxout, float cosyout, float coszout, 
    float *kinfact) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float sin2thetaLab, mr2s2tl, sqrtfact, sqrtK;

  if(costhetaLab<=0.0f && massRatio>=1.0f) return 0;

  sin2thetaLab = 1.0f-costhetaLab*costhetaLab; // sin(theta_lab)^2
  mr2s2tl = massRatio*massRatio*sin2thetaLab;
  if(mr2s2tl>1.) {
    return 0; // scattering above critical angle
  } else {
    sqrtfact = sqrtf(1.0f-mr2s2tl);
    //sqrtfact = mySqrt(1.0f-mr2s2tl);
    sqrtK = (sqrtfact+massRatio*costhetaLab)/(1.0f+massRatio);
    *kinfact = sqrtK*sqrtK;
    return 1;
  }
}


// compute kinematic factor for a recoil
// return 0 if collision forbiden by kinematics, 1 otherwise
// value of kinematic factor stored in *kinfact
int kinematicFactorERD(float massRatio,
  float cosxin,  float cosyin,  float coszin,
  float cosxout, float cosyout, float coszout, 
	float *kinfact ) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float temp;

	if(costhetaLab<0.) {
    return 0; // recoils only possible in forward direction
  } else {
    temp = 1.0f+massRatio;
    *kinfact = 4.0f*massRatio/(temp*temp)*costhetaLab*costhetaLab;
    return 1;
  }
}

// compute Andersen screening factor
double AndersenScreening (float E, double sinThetaCMby2, ion_params *_recoil) {
  double V1 = 48.73*beam.Z*_recoil->Z*sqrt(pow(beam.Z,2./3.)+pow(_recoil->Z,2./3.));
  double massRatio = beam.mass/_recoil->mass;
  double v1byEcm = V1/E*(massRatio+1.0f)*(massRatio+1.0f);
  double temp = v1byEcm/2.0f/sinThetaCMby2;
  temp =  (1.0f+v1byEcm/2.0f)/(1.0f+v1byEcm+temp*temp);
  return temp*temp;
}


// compute the cross-section in Angs^2 given the incomming and outgoing directional cosines
float crossSection(float E,
               float cosxin, float cosyin, float coszin,
               float cosxout, float cosyout, float coszout, double *minimalApproach, int isRBS, ion_params *_recoil ) {

  double costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  double sin2thetaLab = 1.0-costhetaLab*costhetaLab; // sin(theta_lab)^2
  double thetaCM, sinThetaCMby2;
  double crossSectionCM, screeningFact = 1.0, temp;
  double massRatio = beam.mass/_recoil->mass;

  if(isRBS && massRatio*massRatio*sin2thetaLab>1.0) return 0.0f; // above critical angle
  if(!isRBS && costhetaLab<0.0) return 0.0f; // no backward recoils!

  // center-of-mass angle
  thetaCM = isRBS?acos(costhetaLab)-asin(-sqrt(sin2thetaLab)*massRatio):PI-2.0*acos(costhetaLab);
  *minimalApproach = thetaCM*180./PI;
  sinThetaCMby2 = sin(thetaCM*0.5);

  if(costhetaLab>coscutoffAngle) {
    belowCutoff++;
    return 0.0f;
  }

  switch (screen) {
  case 1: // Andersen
    screeningFact = AndersenScreening(E, sinThetaCMby2, _recoil);
  case 0: // Rutherford
    temp = beam.Z*_recoil->Z*E2/(2.0f*E/(1.0+massRatio))/(sinThetaCMby2*sinThetaCMby2);
    crossSectionCM = 0.25*temp*temp*screeningFact;  // Goldstein Eq. 3-102 p. 109 x screening
    break;
  case 2: // Universal potential screening function
  case 3: // Lenz-Jensen screening function
    crossSectionCM = crossSectionScreenPot(E/(1.0+massRatio), beam.Z, _recoil->Z, thetaCM, screen);
    break;
  default:
    fprintf(stderr,"\nScreening type %u unknown.\n",screen);
    exit(1);
  }

  if(isRBS) {
  //  sqrtfact = sqrtf(1.0f-mr2s2tl);
  //  sqrtCrosssect = beam.Z*_recoil.Z*E2/(2.0f*E)*(sqrtfact+costhetaLab)/sin2thetaLab;
  //  return sqrtCrosssect*sqrtCrosssect/sqrtfact; // in A^2/str

    // convert cross-section from CM to lab frame (Goldstein Eq. 3-116, p. 117)
    temp = massRatio*cos(thetaCM);
    return d2f(crossSectionCM*pow(1.0+2.0*temp+massRatio*massRatio,1.5)/(1.0+temp))/l0;
  } 
  
  // else (ERD)
  return d2f(crossSectionCM*costhetaLab*4.0)/l0;
  //  temp = 1.0f+massRatio;
  //  temp =  beam.Z*_recoil.Z*E2*temp/(2.0f*E*costhetaLab);
  //  return  temp*temp/costhetaLab; // in A^2/str

}


// return squared distance to center if the trajectory defined by [x,y,z]+t[l,m,n] intersects detector, -1 otherwise
// i is the detector number to consider
// set [xx yy zz] as the intersection point in detector reference frame
// set [ll mm nn] as the directional vector in detector reference frame
float intersection(int i, float x, float y, float z, float l, float m, float n, 
  float *xx, float *yy, float *zz, float *ll, float *mm, float *nn, float *t0, float winv, float hinv ) {
  float temp, temp1, temp2;
  float vx, vy, vz;

  *ll = l*xpx[i]+m*xpy[i]+n*xpz[i]; 
  *mm = l*ypx[i]+m*ypy[i]+n*ypz[i]; 
  *nn = l*ypx[i]+m*ypy[i]+n*ypz[i];

  *t0 = -(d[i]+xpx[i]*x+xpy[i]*y+xpz[i]*z)/(*ll); // t0=-(d-xp.r0)/(xp.u)
  if(*t0<0.0f)
    return -1; // particle directed away from detector

  // impact point r0+t0*u in sample reference frame
  x += l**t0;
  y += m**t0;
  z += n**t0;

  // vector from detector center to impact point in sample reference frame
  vx = x-pcx[i];
  vy = y-pcy[i];
  vz = z-pcz[i];
  temp = vx*vx+vy*vy+vz*vz; // v^2

  // impact point in detector reference frame
  *xx = 0.0f;                          // v.xp but v in plane and xp normal to plane
  *yy = vx*ypx[i]+vy*ypy[i]+vz*ypz[i]; // v.yp
  *zz = vx*zpx[i]+vy*zpy[i]+vz*zpz[i]; // v.zp

  temp1 = *yy*ainv[i]*winv;
  temp2 = *zz*binv[i]*hinv;
  if(isEllipse[i]) {
    // ellipse equation:          (v.yp/a)^2    + (v.zp/b)^2    = 1
    // virtual ellipse equation:  (v.yp/a Wf)^2 + (v.zp/b Hf)^2 = 1
    if( temp1*temp1+temp2*temp2 > 1.0f )  
      return -1; // particle outside ellipse
  } else {
    // rectangle
    // left of left side or right of right side or below bottom or above top
    if(   temp1<-1.0f    ||      temp1>1.0f     || temp2<-1.0f  || temp2>1.0f)
      return -1;
  }
  return temp;
}


int correctTrajectory(int det,  double *E, double Ebc, double Eac, float k,
                      float *l, float *m, float *n, float lbc, float mbc, float nbc, float lac, float mac, float nac, 
                      float *x, float *y, float *z, float xc, float yc, float zc, ion_params * _recoil) {
  double temp, temp2;
  double cx, cy, cz;
  double costh, sinth;
  double lp, mp, np;
  float K;

  // positions in double precision because calculations involve the subtraction of nearly equal quantities
  double rvx, rvy, rvz;      // vector relating collision site and spot on virtual detector
  double rrx, rry, rrz;      // vector relating collision site and randomly selected spot on detector
  double xx = *x, yy = *y, zz = *z, kk;  // position in double
  double tempd, tempd2;
  double massRatio = beam.mass/_recoil->mass;


  int validKinematic;

  // let rv be a vector relating the collision site and virtual detector intersection point
  rvx = k* *l+xx-xc; 
  rvy = k* *m+yy-yc;
  rvz = k* *n+zz-zc;

  // let rr be the vector between collision site and the rescaled intersection point (in sample reference frame)
  // temporarly, rr = intersection point - pc
  rrx = k* *l+xx-pcx[det];
  rry = k* *m+yy-pcy[det];
  rrz = k* *n+zz-pcz[det];
  
  // tempd, tempd2 are the rescaled y,z components 
  // of the impact point in the detector reference frame
  tempd  = (rrx*ypx[det]+rry*ypy[det]+rrz*ypz[det])/Wf;
  tempd2 = (rrx*zpx[det]+rry*zpy[det]+rrz*zpz[det])/Hf;

  // finally, rr = temp*y' + temp2*z' + pc - colllision site
  rrx = tempd*ypx[det]+tempd2*zpx[det]+pcx[det]-xc;
  rry = tempd*ypy[det]+tempd2*zpy[det]+pcy[det]-yc;
  rrz = tempd*ypz[det]+tempd2*zpz[det]+pcz[det]-zc;

  // c = (rr x rv)/|rr x rv|: normal vector perpendicular to plane defined by rr and rv
  // first, compute rr x rv
  cx = rry*rvz-rrz*rvy;
  cy = rrz*rvx-rrx*rvz;
  cz = rrx*rvy-rry*rvx;
  // normalisation
  temp = sqrt((cx*cx+cy*cy+cz*cz)); // |rr x rv| also required for sinth computation below
  temp2 = 1.0/temp;
  cx*=temp2;
  cy*=temp2;
  cz*=temp2;

  // angle components for rotation from rv to rr
  temp2 = 1.0/sqrt((rvx*rvx+rvy*rvy+rvz*rvz)*(rrx*rrx+rry*rry+rrz*rrz)); // 1/|rr||rv|
  costh = (rvx*rrx+rvy*rry+rvz*rrz)*temp2;  // costh = rr.rv/|rr||rv|
  //sinth = sqrtf(1.-costh*costh);
  sinth = temp*temp2; // sinth = |rr x rv|/|rr||rv|

  // correction to the directionnal cosines at the exit of main collision
  // considering that the particle now flies to the real detector
  lp = lac*costh + (mac*cz-nac*cy)*sinth;
  mp = mac*costh + (nac*cx-lac*cz)*sinth;
  np = nac*costh + (lac*cy-mac*cx)*sinth;

  // store corrected flying direction at the exit of main collision for later cross section computation
  lac = d2f(lp); mac = d2f(mp); nac = d2f(np);

  // recompute collision kinematic factor and cross section considering corrected directional cosines lp, mp, np
  switch(simulationType) {
  case TECH_RBS:
  case TECH_RBS_TOF:
    validKinematic = kinematicFactorRBS(massRatio, lbc, mbc, nbc, d2f(lp), d2f(mp), d2f(np), &K);
    break;
  case TECH_ERD:
  case TECH_ERD_TOF:
    validKinematic = kinematicFactorERD(massRatio, lbc, mbc, nbc, d2f(lp), d2f(mp), d2f(np), &K);
    break;
  default:
    fprintf(stderr, "Virtual detection not implemented for simulation type %i.\n", simulationType);
    exit(1);
    break;
  }
  if(!validKinematic) return 0;

  // correction to the flying direction at target exit
  lp = *l*costh + (*m*cz-*n*cy)*sinth;
  mp = *m*costh + (*n*cx-*l*cz)*sinth;
  np = *n*costh + (*l*cy-*m*cx)*sinth;

  // correction of the energy loss considering the new flying direction
  // (since more or less distance is traveled through the sample)

  // rv now defines the vector between main collision site and target exit point in original direction
  rvx = xx-xc; // x ~ 0
  rvy = yy-yc;
  rvz = zz-zc;

  // rr = rotation of rv 
  rrx = rvx*costh + (rvy*cz-rvz*cy)*sinth;
  rry = rvy*costh + (rvz*cx-rvx*cz)*sinth;
  rrz = rvz*costh + (rvx*cy-rvy*cx)*sinth;

  // find new exit point
  kk = -xc/rrx; // trajectory is k*rr+[xc,yc,zc] and intersects surface at x=0 => k*rrx+xc = 0
  *x = d2f(kk*rrx+xc); 
  *y = d2f(kk*rry+yc);
  *z = d2f(kk*rrz+zc);
  
  // corrected flying direction
  *l = d2f(lp);
  *m = d2f(mp);
  *n = d2f(np);

  // rr now defines the vector between main collision site and new target exit point in corrected direction
  rrx = *x-xc;
  rry = *y-yc;
  rrz = *z-zc;

  // correction to energy transfer during main collision   and   energy loss
  //*E = Ebc*K-(Eac-*E)*mySqrt(d2f((rrx*rrx+rry*rry+rrz*rrz)/(rvx*rvx+rvy*rvy+rvz*rvz)));
  tempd = (rrx*rrx+rry*rry+rrz*rrz)/(rvx*rvx+rvy*rvy+rvz*rvz);
  *E = Ebc*K-(Eac-*E)*sqrt((rrx*rrx+rry*rry+rrz*rrz)/(rvx*rvx+rvy*rvy+rvz*rvz));
  return 1;
}

/*
void myswap(float *a, float *b) {
    float temp;
    temp = *a;
    *a   = *b;
    *b   =  temp;
}

void myswapd(double *a, double *b) {
    double temp;
    temp = *a;
    *a   = *b;
    *b   =  temp;
}
*/

// main loop (called using a structure so we can also call it using POSIX threads)
// For each ion, it calls functions to compute:
//  - slowdown of incident ion
//  - new direction after main collision, kinematic factor
//  - slowdown of scattered ion or recoil
// The function then :
//  - determines intersection with real or virtual detector
//  - for virtual detector, selects a spot on real detector and make trajectory corrections
//  - calls function to compute slowdown through absorber/foil, if any
//  - determines intersection second detector, if any
//  - if particle is detected, calls function to compute screened cross-section and increment histogram by this amount
void mainloop(void *arg) {

  thread_params *param =(thread_params *)arg;
  unsigned long numProj    = param->numProjectile;
  unsigned long totalnumbercoll  = 0;
  unsigned int  i, ilayer, collisionLayer = param->collisionLayer;

  // local variables
  unsigned long iProjectile, numbercoll = 0, numbercoll2 = 0;
  int exitStatus, validKinematic;
  double E, EE;            // energy, energy of recoil in coincidence
  float x, y, z, y0, z0;
  float l, m, n;       // position, directional cosines
  float x2, y2, z2;
  float l2, m2, n2; // position, directional cosines of recoil in coincidence detection
  float xp, yp, zp;
  float lp, mp, np;       // ...same, detector coordinates
  float x2p, y2p, z2p;
  float l2p, m2p, n2p; 

  float xa, ya, za, la, ma, na, xb, yb, zb, lb, mb, nb, Ea; // absober and TOF

  float crosssect;
  //float V1 = d2f(48.73*beam.Z*recoil->Z*sqrt(pow(beam.Z,2./3.)+pow(recoil->Z,2./3.))); // parameter for cross-section sreening
  float temp = d2f(randomx()), temp2;
  float *minCosScat = NULL, *minCosScatDepth = NULL; // location to store cos(scat_angle) and corresp. depth
  double minimalApproach;

  float WfInv = 1.0f/Wf, HfInv = 1.0f/Hf;

  // create and initialize counters so each thread has its own
  unsigned int ir = (unsigned int)(temp*MAXRANLIST); // random
  unsigned int il = (unsigned int)(temp*MAXLOGLIST); // log
  unsigned int ie = (unsigned int)(temp*MAXERFLIST); // erf 
  unsigned int ia = (unsigned int)(temp*MAXAZILIST); // azimulal angle

  // variable for picking an element in the material of which a bitmap pixel consists
  unsigned long ix, iy, iz;
  long ielem, j;
  int retval;
  float ran, sumAbundance;
  unsigned char imaterial;

  float collDepth;         // collision depth
  float t0;                // parameter defining collision position
  double TOF;

  // variables related to virtual detector calculations
  float xc, yc, zc;         // position of main collision
  float lbc, mbc, nbc, Ebc; // directional cosines and energy just before main collision
  float lac, mac, nac, Eac; // directional cosines and energy just after main collision
  float lac2, mac2, nac2, Eac2;
  float cx, cy, cz;         
  float K;                  // kinematic factor 

  int isQuick = numberLayers==1 && numberElements[0]==1 && simulationType!=-2;

  float layerThickness[MAXLAYERS+1];    // thickness of each layer [A]
  float layerTop[MAXLAYERS+1];          // depth of the top of each layer [A]
  float layerBottom[MAXLAYERS+1];       // depth of the bottom of each layer [A]  

  double massRatio, isotopMass;
  float local_factorNtOmegaByNumdetpart = factorNtOmegaByNumdetpart;
  ion_params *_recoil = recoil; // we work with a copy of the global
  // variable because in the case of bitmaps, the value has to be changed

  int _currentCScorrType; // value of the cross-section correction type
  char CSstamp[100]; // name of the cross-section correction file to load (if _currentCScorrType > 0)

  int _currentIsNatural; // stores the value of isNatural, which may change in the case of bitmap simulations

  if(simulationType>0 && (!isBitmap || collisionLayer==substrateLayer)) {
      massRatio = beam.mass/_recoil->mass;
      _currentCScorrType = currentCScorrType;
      _currentIsNatural = isNatural;
  }

  if(simulationType==TECH_FULLRBS) {
      if(writeDetails) {
          // if writeDetails !=0, we will save NUMMINCOSSCAT values
          minCosScatDepth  = (float *) malloc(NUMMINCOSSCAT*sizeof(float));
          minCosScat  = (float *) malloc(NUMMINCOSSCAT*sizeof(float));
      } else {
          // we save only the maximum scattering (one value)
          minCosScatDepth  = (float *) malloc(sizeof(float));
          minCosScat  = (float *) malloc(sizeof(float));
      }
  }

#ifdef MAX_THREAD
  pthread_mutex_lock(&mut);
#endif
  fflush(stderr);
#ifdef MAX_THREAD
  pthread_mutex_unlock(&mut);
#endif

  for(iProjectile=0; iProjectile<numProj; iProjectile++) { // ions loop

    if(numProj<10 || (iProjectile+1)%(numProj/10)==0) {
      fprintf(stdout, "_");
      fflush(stdout);
    }

    if(ie>MAXERFLIST-6) ie=0; 

    // starting projectile energy
    E = E0;

    // starting at the origin to avoid single precision problems
    // beam size effects added afterwards
    x = 1e-20f;
    y = 0;
    z = 0;

    // In the case of a bitmap, we want to impact the target at a random location in (y,z)
    // We will add to y,z an offset y0,z0 and we will subtract it back at the end
    // as to ensure that we are not broadening the beam by yDim or zDim
    if(isBitmap) {
        y0 = (float) (yDim/pixelsPerAngs*randomx());
        y += y0;
        if(zDim>0) {
            z0 = (float) (zDim/pixelsPerAngs*randomx());
            z += z0;
        }
    }

    // starting directional cosines
    l = l0;
    m = m0;
    n = n0;

    lp = 99;
    l2p = 99;
    lac = 99;
    lac2 = 99;

    if(!isBitmap) {
        // find layer thicknesses for the current ion simulation, based on thickness distribution
        for(ilayer=0; ilayer<numberLayers; ilayer++) {
          temp = erflist[ie++]*layerThickDist[ilayer];
          layerThickness[ilayer] = layerThickness0[ilayer]+temp; // add thickness variation of current layer
          if(ilayer>0 && layerThickDist[ilayer-1]<0.0f) layerThickness[ilayer]-=temp2; // if negative, subtract thickness variation of previous layer
          // (this is for the special case where an interface between layer can be rough while external surfaces are flat)

          if(layerThickness[ilayer]<0) layerThickness[ilayer] = 0;
          temp2 = temp; // record current layer thickness in case it is needed for next layer thickness variation

          // depth of layer top and bottom
          if(ilayer==0 || ilayer>=numberLayers) {  // top layer or foils
            layerTop[ilayer] = 0.;
            layerBottom[ilayer] = layerThickness[ilayer];
          } else {
            layerTop[ilayer] = layerBottom[ilayer-1];
            layerBottom[ilayer] = layerTop[ilayer] + layerThickness[ilayer];
          }
        }
    }

    if(simulationType>0) {
        // follow projectile to predefined depth of main collision
        if(isBitmap) {
            // in the case of a bitmap, defined collision depth somewhere in the substrate if we are in the substrate...
            if(collisionLayer==substrateLayer) {
                collDepth = bitmapThichness+layerThickness0[substrateLayer]*(float)iProjectile/(float)numProj+1e-20f;
            } else {
                // ... or in the bitmap if we are in the bitmap (i.e. any other value of collisionLayer
                // since in this case layers are actually material description for a pixels value)
                collDepth = bitmapThichness*(float)iProjectile/(float)numProj+1e-20f;
            }
        } else {
            // in the case of the usual stack of layers, find a location in collisionLayer
            collDepth = layerTop[collisionLayer]+layerThickness[collisionLayer]*(float)iProjectile/(float)numProj+1e-20f;
        }
    } else {
        // we want to follow implanted ions wherever they go, down to minimum energy or until they exit target
        collDepth = 1e32f; // ... so put collison depth to infinity and beyond!
    }
    // compute projectile slowdown
    numbercoll = 0;

    if(simulationType==TECH_FULLRBS) {
        if(writeDetails) {
            for(i=0; i<NUMMINCOSSCAT; i++) {
                minCosScat[i] = 1.0f;
                minCosScatDepth [i] = 0.0f;
            }
        } else {
            minCosScat[0] = 1.0f;
            minCosScatDepth [0] = 0.0f;
        }
    }


    if(isQuick) {
      exitStatus = quickslowdown(&x, &y, &z, &l, &m, &n, &E, collDepth, 1, &numbercoll, &ir, &il, &ie, &ia,
              MostAbundantIsotope[beam.Z]/beam.mass, interatomicDist[0], layerTop[0], layerBottom[0], 
              beam.a[0][0], beam.f[0][0], beam.K[0][0], beam.smax[0][0], beam.CosScat[0][0], beam.SinScat[0][0],
              beam.sp[0], beam.sqrtStrag[0], sqrtMeanFreePath[0], layerTop[0], layerBottom[0],
              minCosScat, minCosScatDepth);
    } else if(isBitmap) {
      exitStatus = pixmap_slowdown( &x, &y, &z, &l, &m, &n, &E, collDepth,
                                    &numbercoll, &ir, &il, &ie, &ia, &beam, minCosScat, minCosScatDepth);
      if(x>collDepth) {
          // it is possible that x exceeds collDepth because we have made some distance through vacuum
          // in that case, subtract the excess length to x, y, z
          temp = (x-collDepth)/l; // excess traveled length
          x = collDepth;
          y -= temp*m;
          z -= temp*n;
      }
    } else {
      exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, collDepth, 1,
                            &numbercoll, &ir, &il, &ie, &ia, &beam,0,layerTop, layerBottom, minCosScat, minCosScatDepth);
    }
    totalnumbercoll+=numbercoll;

    // when the main collision occurs in a pixel, we have to find which
    // element (and isotope, if more than one) will be our main collision partner
    if(simulationType>0 && exitStatus==0 && isBitmap && collisionLayer!=substrateLayer) {
        // select a recoil randomly among the composition of the pixel we are in
        retval = getIndices(x, y, z, l, &ix, &iy, &iz);

        if(retval>0) {
            fprintf(stderr, "\nError: main collision locus supposed to be inside bitmap at this "
                   "step, but located outside of it. Giving up.\n");
            exit(1);
        } else {
            imaterial = bitmap[(iz*yDim+iy)*xDim+ix];
        }

        if(imaterial==UCHAR_MAX) {
            // we have collided with vacuum! No scattering, no event
            exitStatus = -7;
        } else {
            // randomly pick the partner of the main collision
            if(ir>=MAXRANLIST-2) ir = 0;
            ran = randomlist[ir++];
            for(ielem=0;ielem<numberElements[imaterial]; ielem++)
                if(elementFractionSum[imaterial][ielem]>=ran)
                    break;

            _recoil = element[imaterial][ielem]; // recoil changes
            _currentIsNatural = natural[imaterial][ielem];
            if(nsim[imaterial][ielem]==1) {
                _recoil->mass = massIsotope[imaterial][ielem][0]; // recoil mass
                _currentCScorrType = CScorrType[imaterial][ielem][0]; // CScorrType changes
            } else if(nsim[imaterial][ielem]>1) {
                // we have to select (randomly) the mass o the isotope we are making a main collision with
                ran = randomlist[ir++];
                sumAbundance = 0;
                for(j=0;j<nsim[imaterial][ielem]; j++) {
                    sumAbundance+=abundance[imaterial][ielem][j];
                    if(sumAbundance>=ran)
                        break;
                }
                _recoil->mass = massIsotope[imaterial][ielem][j];
                _currentCScorrType = CScorrType[imaterial][ielem][j];
            }

            // Here, we do not include a factor for elemental concentration
            // because we selected it randomly, so its weight should already
            // be represented in its share of the number of detected particles.
            // (See detailed explanation for factorNtOmegaByNumdetpart elsewhere in the code)
            local_factorNtOmegaByNumdetpart = d2f(bitmapThichness
                                            *layerDensity[imaterial]*2.0*PI*(1.0-cosAngSpread0)/numProjectile/(Wf*Hf));
            massRatio = beam.mass/_recoil->mass;
        }
    }

    // for IBA techiques, compute main collision and slowdown of emerging ion(s)
    if(simulationType>0 && exitStatus==0) {

      // store direction and energy before main collision (bc)
      lbc = l; mbc = m; nbc = n; Ebc = E;
      // ... and locus of main collision
      xc = x; yc = y; zc = z; 
      x2 = x; y2 = y; z2 = z; // used for coincidence 

      // find new flying direction towards detector 0
      mainCollision(0, &l, &m, &n); 

      // store direction after main collision (ac)
      lac = l; mac = m; nac = n;

      // find new flying direction and compute kinematic factor
      switch(simulationType) {
      case TECH_RBS: //RBS
      case TECH_RBS_TOF: //RBS-TOF
        validKinematic = kinematicFactorRBS(massRatio, lbc, mbc, nbc, l, m, n, &K);
        E*=K;  // apply kinematic factor
        Eac=E; // store energy after main collision
        break;
      case TECH_ERD: //ERDA
      case TECH_ERD_TOF: //ERD-TOF
        validKinematic = kinematicFactorERD(massRatio, lbc, mbc, nbc, l, m, n, &K);
        E*=K; // apply kinematic factor
        Eac=E; // store energy after main collision
        break;
      case TECH_COINC: //coincidence (two detectors)
        // scattering
        validKinematic = kinematicFactorRBS(massRatio, lbc, mbc, nbc, l, m, n, &K);
        E*=K;       // apply kinematic factor
        EE = Ebc-E; // recoil energy considering an elastic collision

        // store energy of the two particles after collision
        Eac = E;
        Eac2 = EE;

        // we assume identical masses, i.e. recoiling at 90 deg. from scattering
        if(massRatio!=1.0) {
          fprintf(stderr, "\nCoincidence for different masses not implemented.\n");
          exit(1);
        }
        // determine recoil direction: [l2 m2 n2] = [l m n]_ac x c
        // with c a unit vector parallel to [l m n]_bc x [l m n]_ac
        cx = mbc*nac-nbc*mac;
        cy = nbc*lac-lbc*nac;
        cz = lbc*mac-mbc*lac;
        temp = 1.0f/sqrtf(cx*cx+cy*cy+cz*cz);
        cx*=temp;
        cy*=temp;
        cz*=temp;

        // [l2 m2 n2] = [l m n]_ac x c/|[l m n]_ac x c| 
        // but |[l m n]_ac x c|=1 because they are unit vectors perp. to each other
        l2 = mac*cz-nac*cy;
        m2 = nac*cx-lac*cz;
        n2 = lac*cy-mac*cx;
        lac2 = l2; mac2 = m2; nac2 = n2;
        break;

      default:
        fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
        exit(1);
        break;
      }
   
      if(!validKinematic) {
        // we are not continuing, so add beam size to positions

        if(isBitmap) {
              // in the case of a bitmap, we have offset the impact point to hit the target at a random location
              y -= y0; // subtract back the y0 value added
              if(zDim>0) {
                  z -= z0; // subtract back the z0 value added
              }
        }

        temp = ySpotSize*erflist[ie++];
        temp2 = zSpotSize*erflist[ie++]; 
        y+=temp;
        z+=temp2;
        y2+=temp;
        z2+=temp2;
        yc+=temp;
        zc+=temp2;

      } else {
            
        // slowdown of outgoing ion(s)
        switch(simulationType) {
        case TECH_RBS: //RBS
        case TECH_RBS_TOF: //RBS-TOF
          // slowdown of scattered ion up to just above surface (-1e-10) or until something else happen
          if(isBitmap)
              exitStatus = pixmap_slowdown( &x, &y, &z, &l, &m, &n, &E, -1e-10f,
                                            &numbercoll, &ir, &il, &ie, &ia, &beam, NULL, NULL);
          else
              exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, -1e-10f, 1,
                                &numbercoll, &ir, &il, &ie, &ia, &beam,0,layerTop, layerBottom, NULL, NULL);
          totalnumbercoll+=numbercoll;
          break;
        case TECH_ERD: //ERDA
        case TECH_ERD_TOF: //ERD-TOF
            if(isBitmap)
                exitStatus = pixmap_slowdown( &x, &y, &z, &l, &m, &n, &E, -1e-10f,
                                              &numbercoll, &ir, &il, &ie, &ia, _recoil, NULL, NULL);
            else
                exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, -1e-10f, 1,
                                &numbercoll, &ir, &il, &ie, &ia, _recoil,0,layerTop, layerBottom, NULL, NULL);
          totalnumbercoll+=numbercoll;
          break;
        case TECH_COINC: // coincidence: follow both scattered ion and recoil
            if(isBitmap)
                exitStatus = pixmap_slowdown( &x, &y, &z, &l, &m, &n, &E, xDim*AngsPerPixel,
                                              &numbercoll, &ir, &il, &ie, &ia, &beam, NULL, NULL);
            else
                exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, layerBottom[numberLayers-1], 1,
                                &numbercoll, &ir, &il, &ie, &ia, &beam,0,layerTop, layerBottom, NULL, NULL);
          totalnumbercoll+=numbercoll;

          if(exitStatus==0) {// only follow recoil if scattered ion made it to sample back
              if(isBitmap)
                  exitStatus = pixmap_slowdown( &x2, &y2, &z2, &l2, &m2, &n2, &EE, xDim*AngsPerPixel,
                                                &numbercoll, &ir, &il, &ie, &ia, _recoil, NULL, NULL);
              else
                  exitStatus = slowdown(&x2, &y2, &z2, &l2, &m2, &n2, &EE, layerBottom[numberLayers-1], 1,
                                  &numbercoll2, &ir, &il, &ie, &ia, _recoil,0,layerTop, layerBottom, NULL, NULL);
            totalnumbercoll+=numbercoll2;
          }
          break;
        default:
          fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
          exit(1);
          break;
        }

        // we have finished crossing the target, add beam size to positions
        if(isBitmap) {
              // in the case of a bitmap, we have offset the impact point to hit the target at a random location
              y -= y0; // subtract back the y0 value added
              if(zDim>0) {
                  z -= z0; // subtract back the z0 value added
              }
        }
        temp = ySpotSize*erflist[ie++];
        temp2 = zSpotSize*erflist[ie++]; 
        y+=temp;
        z+=temp2;
        y2+=temp;
        z2+=temp2;
        yc+=temp;
        zc+=temp2;

        // detection
        if(exitStatus==0) {
          switch(simulationType ) {
          case TECH_COINC: // annular detector coincidence
            // intersection of scattered ion with detector 0 or 1
            temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0,1,1); 
            temp2 = intersection(1, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0,1,1);
            if(temp<r2annular[0] && temp2<r2annular[1]) {
              exitStatus = -3;
              break; // scattered ion not detected: no coincidence, don't bother to check recoil
            }
            // intersection of recoil with detector 1 or 0
            temp = intersection(1, x2, y2, z2, l2, m2, n2, &x2p, &y2p, &z2p, &l2p, &m2p, &n2p, &t0,1,1);
            temp2 = intersection(0, x2, y2, z2, l2, m2, n2, &x2p, &y2p, &z2p, &l2p, &m2p, &n2p, &t0,1,1);
            if(temp<r2annular[1] && temp2<r2annular[0]) {
              exitStatus = -3;
              break; // recoil not detected: no coincidence
            }
            exitStatus = 0;
            break;
          case TECH_RBS: // RBS
          case TECH_RBS_TOF: //RBS-TOF
          case TECH_ERD: // ERDA
          case TECH_ERD_TOF: // ERD-TOF
            // intersection with detector 0
            temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0, WfInv, HfInv); // detector 0
            if(temp<0) {
              exitStatus = -3;
              break; // ion not detected
            }

            if(Hf!=1.0f || Wf!=1.0f) { 
              // correct for virtual detection
              validKinematic = correctTrajectory(0, &E, Ebc, Eac, t0, 
                &l, &m, &n, lbc, mbc, nbc, lac, mac, nac, 
                &x, &y, &z, xc, yc, zc, _recoil);

              // recompute corrected impact point to see if it is within annulus
              temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0,1,1);
              if(!validKinematic || temp<r2annular[0]) {
                exitStatus = -5;
                break; // ion not detected
              }

            }
            break;
          default:
            fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
            exit(1);
            break;
          }
        }

        // compute cross section for detected ions
        if(exitStatus ==0) {
          switch(simulationType) {
          case TECH_RBS:
          case TECH_RBS_TOF:
          case TECH_COINC:
            crosssect = crossSection(Ebc, lbc, mbc, nbc, lac, mac, nac, &minimalApproach, 1, _recoil);
            break;
          case TECH_ERD: // ERD
          case TECH_ERD_TOF: // ERD-TOF
            crosssect = crossSection(Ebc, lbc, mbc, nbc, lac, mac, nac, &minimalApproach, 0, _recoil);
            break;
          default:
            fprintf(stderr, "\nNo cross-section computation possible for simulation type %i\n", simulationType);
            exit(1);
            break;
          }

        } else {
          crosssect = 0; // just to make sure no one uses an old value
        }

        if(exitStatus==0) {

            // if absorber present, compute slowdown through it
            if(absorberThick) {

              // because absorber is thin compared to detector distances
              // we just move current depth from 0 to a depth corresponfing
              // to absorber thickness and compute slowdown through it
              xa = absorberThick;

              ya = yp;
              za = zp;
              la = lp;
              ma = mp;
              na = np;
              Ea = E;

              // slowdown through absorber (outcome could be other than exitStatus==0)
              switch(simulationType) {
              case TECH_RBS: //RBS
              case TECH_RBS_TOF: //RBS-TOF
                exitStatus = slowdown(&xa, &ya, &za, &la, &ma, &na, &E,
                  -1e-10f, 0, &numbercoll, &ir, &il, &ie, &ia, &beam,0,layerTop, layerBottom, NULL, NULL);
                totalnumbercoll+=numbercoll;
                break;
              case TECH_ERD: //ERD
              case TECH_ERD_TOF: //ERD-TOF
                exitStatus = slowdown(&xa, &ya, &za, &la, &ma, &na, &E,
                  -1e-10f, 0, &numbercoll, &ir, &il, &ie, &ia, _recoil,0,layerTop, layerBottom, NULL, NULL);
                totalnumbercoll+=numbercoll;
                break;
              default: //including TECH_COINC
                fprintf(stderr,"\nNo absorber/foil allowed for simulation type %i",simulationType);
                exit(1);
                break;
              }

              // time of flight techniques: detector 1 is stop detector, positioned in reference frame of detector 0
              if( exitStatus==0 && (simulationType==TECH_RBS_TOF || simulationType==TECH_ERD_TOF) ) {
                // did the particle reach the final detector? If yes, t0 gives the distance traveled between the start and stop detector
                if( intersection(1, xa, ya, za, la, ma, na, &xb, &yb, &zb, &lb, &mb, &nb, &t0,1,1)<0 ) {
                  exitStatus=-4;
                  TOF = -1;
                } else {
                  // compute time of flight [ns]
                  if(E>0) {
                      if(simulationType==TECH_RBS_TOF) TOF = t0*0.1*sqrt(beam.mass/E*AMUbyE); // factor 0.1 converts Angs to meters and seconds to ns
                      else TOF = t0*0.1*sqrt(_recoil->mass/E*AMUbyE*0.5); // factor 0.1 converts Angs to meter and seconds to ns
                  } else {
                    TOF = -2;
                    exitStatus=-4;
                  }
                }
              }
            }
        }

        if(exitStatus==0) {


#ifdef MAX_THREAD
          pthread_mutex_lock(&mut);
#endif

          // write detected particle data

          if(simulationType == TECH_COINC) {
              fprintf(ofdetect, "%e ", EE); // in case of coincidence, write second particle energy first
          }

          // in case of TOF technique, write TOF in detector first
          if(simulationType == TECH_ERD_TOF || simulationType == TECH_RBS_TOF) {
              fprintf(ofdetect, "%e ", TOF);
          }

          if(_currentCScorrType) {
              isotopMass = _recoil->mass;
              if(_currentIsNatural) // we are using the natural composition...
                  isotopMass = 0; // indicate that by setting M2 to 0 in CSstamp
               sprintf(CSstamp, "%1i%1i.%03.0f%02i%03.0f%02i", simulationType, _currentCScorrType, beam.mass, beam.Z, isotopMass, _recoil->Z);
          } else
              sprintf(CSstamp, "0");

          fprintf(ofdetect, "%e  %e %e  %e %e  %e %e  %u %u %s",
                  E, crosssect, local_factorNtOmegaByNumdetpart, Ebc,
                  lbc*lac+mbc*mac+nbc*nac,  -lac*xpx[0]-mac*xpy[0]-nac*xpz[0], xc,
                  _recoil->Z, collisionLayer, CSstamp);

          if(writeDetails) {
              fprintf(ofdetect,
                      // the following lines are a single line format
                      "  %e %e  "
                      "%e %e %e  %e %e %e  %e %e %e  "
                      "%e %e %e  %e %e %e  "
                      "%e %e %e  %e %e %e  ",
                      minimalApproach, Eac,
                      xc, yc, zc, lbc, mbc, nbc, lac, mac, nac,
                      x, y, z, l, m, n,
                      xp, yp, zp, lp, mp, np);

              if(simulationType == TECH_COINC) {
                  fprintf(ofdetect,
                          // the following lines are a single line format
                          "%e %e %e %e   "
                          "%e %e %e  %e %e %e  "
                          "%e %e %e  %e %e %e  ",
                          Eac2, lac2, mac2, nac2,
                          x2, y2, z2, l2, m2, n2,
                          x2p, y2p, z2p, l2p, m2p, n2p);
//                  fprintf(ofdetect, "    %e %e", crossSection(Ebc, lbc, mbc, nbc, lac2, mac2, nac2, &minimalApproach, 0),
//                           crossSection(Ebc, lbc, mbc, nbc, lac2, mac2, nac2, &minimalApproach, 1) );
              }

              if(absorberThick) {
                  fprintf(ofdetect,
                          // the following lines are a single line format
                          "%e %e %e  %e %e %e  "
                          "%e %e %e  %e %e %e  %e ",
                          xa, ya, za, la, ma, na, xb, yb, zb, lb, mb, nb, Ea);
              }

          }
          fprintf(ofdetect, "\n");
          fflush(ofdetect);

#ifdef MAX_THREAD
          pthread_mutex_unlock(&mut);
#endif
        } 

      }
    }

    // case of full RBS: check intersection of ions that emerged surface with detector 0
    if(exitStatus==-1 && simulationType==TECH_FULLRBS) {
        // add beam size to positions
        if(isBitmap) {
              // in the case of a bitmap, we have offset the impact point to hit the target at a random location
              y -= y0; // subtract back the y0 value added
              if(zDim>0) {
                  z -= z0; // subtract back the z0 value added
              }
        }
        y += ySpotSize*erflist[ie++];
        z += zSpotSize*erflist[ie++]; 

        // intersection with detector 0
        temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0, WfInv, HfInv); // detector 0
        if(temp<0) {
          exitStatus = -3;
        } else {
            exitStatus = 0;
            fprintf(ofdetect, "%e %e %e", E, minCosScat[0], minCosScatDepth[0]);

            if(writeDetails) {
                fprintf(ofdetect, "  %e %e %e  %e %e %e  %e %e %e  %e %e %e",
                        x, y, z,  l, m, n,  xp,  yp,  zp,  lp,  mp,  np);
                fprintf(ofdetect, "  %u", NUMMINCOSSCAT);
                for(i=0; i<NUMMINCOSSCAT; i++)
                    fprintf(ofdetect, "  %e %e", minCosScat[i], minCosScatDepth[i]);
            }
            fprintf(ofdetect, "\n");
            fflush(ofdetect);
        }
    }

    // compute sums
    switch (exitStatus) {
    case 0:
        if(simulationType<0) {
            // Unexpected but extreemly rare outcome due to a rotate() call with [l m n] = [0 0 1] that return infinite value
            // ... and hoped for but rare outcome in case of full RBS: ion reached detector
            fprintf(stdout, "*");
            if(simulationType == TECH_FULLRBS)
                detectedParticles++;
        } else {
            if(validKinematic)
                detectedParticles++;
            else {
                exitStatus = -6;
                badKinematics++;
            }
        }
        break;
    case -3:
      exitUndetected++;
      break;
    case -4:
      exitUndetectedFinal++;
      break;
    case -5:
      badRescale++;
      break;
    case -6:
      badKinematics++;
      break;
    case -7:
      collVacuum++;
      break;
    case 1:
      exitBottom++;
      break;
    case -1:
      exitTop++;
      break;
    case -2:
      exitMinEnergy++;
      break;
    default:
      fprintf(stderr, "\nExit status %i unknown.\n",exitStatus);
      exit(1);
      break;
    }

  } // end of ion loop

  param->totalnumbercoll = totalnumbercoll;

  if(simulationType==TECH_FULLRBS) {
      free(minCosScatDepth);
      free(minCosScat);
  }


#ifdef MAX_THREAD
  pthread_exit(0);
#endif

}


// skip comment lines, i.e. lines beggining by anything else than a digit, +, -, '.' or space
void skipComments(FILE * ifp) {
  unsigned int ch;

  if(ftell(ifp)>0) ignoreline(ifp); // if something was already read from file, ignore the rest of current line (e.g. return)

  ch = fgetc(ifp); // probe frist character of new line
  while(!isdigit(ch) && ch!=' ' && ch!='+' && ch!='-' && ch!='.' && ch != EOF) {
    // if its not a digit, +, -, '.' or a space, its a comment, ignore line ..
    ignoreline(ifp);
    //... and probe first character of next line
    ch = fgetc(ifp);
  }
  ungetc(ch, ifp); // put back last character into stream so it is ready to read
}


// check if elem already exists in element[][] table
// it it does, assign element[ilayer][ielement] to that element and return 1
// if it doesn't, allocate a new element structure and return 0
int assignElement(unsigned int Z, float mass, unsigned int ilayer, unsigned int ielement) {
  unsigned int i1, i2;
  ion_params * ion;
  
  for(i1=0; i1<=MAXLAYERS; i1++) {
    for(i2=0; i2<MAXELEMENT; i2++) {
      if(element[i1][i2] != NULL && element[i1][i2]->Z==Z && fabs(element[i1][i2]->mass - mass) < 0.001) {
        // identical element found, point to the same structure
        element[ilayer][ielement] = element[i1][i2];
        return 1;
      }
    }
  }


  // not found, allocate a new structure
  element[ilayer][ielement] = malloc(sizeof(ion_params));

  // initialize newly created structure
  ion = element[ilayer][ielement];
  ion->Z    = Z;
  ion->mass = mass;
  ion->roundMass = (int)round(mass);
  ion->tablesComputed = 0;

  for(i1=0; i1<=MAXLAYERS; i1++) {
    for(i2=0; i2<MAXELEMENT; i2++) {
      ion->CosScat[i1][i2] = NULL;
      ion->SinScat[i1][i2] = NULL;
    }
    ion->sp[i1] = NULL;
    ion->sqrtStrag[MAXLAYERS] = NULL; // sqrt of energy straggling * sqrt(2) (computed according to Bragg's rule)
  }
  return 0;
}


// Prepare for calculations: read the input file, compute several parameters, call functions to compute tables
// convention: energies in eV, lengths in A, densities in at/A3 (=10^24 at/cm4)
int input( ) {

    float         beamRadius;    // sigma of a gaussian, radially simetric beam (0 to ignore)
    float         tempcc[MAXLAYERS];
    float         meanFreePath;    // mean free path, 0 to use interatomic distance

    unsigned int ilayer, ielem, nth, numTravel;

    // other working variables
    unsigned long i;
    unsigned int j, Z;
    float sum, temp, mass;
    int randomseed, itemp;

    // file handle
    FILE *ifp;
    char *mode = "r";
    char filename[FILENAMESIZE];

    ofdetect   = NULL;

    // fill randomlist, sqrtloglist, sinAzimAngle and cosAzimAngle
    computelists();
    //mySqrtTableFill();

    // load scatteing matrix
    if(!loadMatrix()) {
        fprintf(stderr, "\nError loading scattering matrix.\n");
        return 1;
    }

    // load the list of Chu straggling correction parameters
    if(!loadChu()) {
        fprintf(stderr, "\nError loading Chu parameters.\n");
        return 1;
    }

    // load a list of inverse error function values
    if(!loadErfInv()) {
        fprintf(stderr, "\nError loading inverse error function file.\n");
        return 1;
    }
    randomizelist(erflist, MAXERFLIST); // put the list in random order

    // load parameters from file filenameWithoutExt +".in"
    // (floating point values are read as strings then converted to avoid formatting problems)
    sprintf(filename,"%s.in", filenameWithoutExt);
    ifp = fopen(filename, mode);
    if (ifp == NULL) {
        fprintf(stderr, "\nCan't open input file %s\n", filename);
        return 1;
    }

    // simulation parameters
    skipComments(ifp);
    fscanf(ifp, "%d %g %g %g %u %u %u %g %d %u", &simulationType, &meanFreePath, &cosAngSpread0, &minE0,
           &enerStragType,  &writeDetails, &screen, &coscutoffAngle, &randomseed, &nth);

    if(randomseed) randomxinit(randomseed,0);

    if(meanFreePath<0) {
        // if meanFreePath<0, deflections "turned off": no rotations during trajectory computation
        doRotate = 0;
        meanFreePath = -meanFreePath;
        // if meanFreePath<0, than the user actually entered the minimum deflection angle in degrees
        //sinMinDeflAngle = -meanFreePath; // sinf(-meanFreePath*PI/180.0); // currently actually minimum energie transfer
        sinMinDeflAngle = 0;
        //meanFreePath = 0;
    } else {
        sinMinDeflAngle = 0;
        doRotate = 1;
    }

    cosAngSpread0 = d2f(cos(cosAngSpread0*PI/180.));
    coscutoffAngle = d2f(cos(coscutoffAngle*PI/180.));

    if(enerStragType>3) {
        fprintf(stderr, "\nEnergy straggling calculation type %u unknown.\n",enerStragType);
        return 1;
    }
    if(screen>3) {
        fprintf(stderr, "\nScreening type  %u unknown.\n",screen);
        return 1;
    }

#ifdef MAX_THREAD
   nthread = nth;
   if(nthread<1) {
       fprintf(stderr, "\nWarning: number of threads set to 1.\n");
       nthread = 1;
   }
   if(nthread>MAX_THREAD) {
       fprintf(stderr, "\nError: number of threads %u larger than maximum value %u.\n", nthread, MAX_THREAD);
       return 1;
   }
#else
    nthread =1;
#endif

    // beam parameters
    skipComments(ifp);
    fscanf(ifp, "%u %f %lu %g %g", &beam.Z, &beam.mass, &numProjectile, &E0, &beamRadius);
    if( beam.mass == 0.0f) beam.mass = MostAbundantIsotope[beam.Z]; // if mass==0 use default
    beam.roundMass = (int)round(beam.mass);

    // directional cosines of incident beam
    skipComments(ifp);
    fscanf(ifp, "%g %g %g", &l0, &m0, &n0);

    // normalize beam directinal cosines
    sum = sqrtf(l0*l0+m0*m0+n0*n0);
    l0/=sum;
    m0/=sum;
    n0/=sum;

    // projection of beam size on surface
    ySpotSize = beamRadius/sqrtf(1.0-m0*m0);
    zSpotSize = beamRadius/sqrtf(1.0-n0*n0);

    // total number of layers, total number of different ions for which the trajectory is followed
    skipComments(ifp);
    fscanf(ifp, "%i %u", &itemp, &numTravel);
    // absolute value of itemp is the number of layers (or materials in case of a bitmap)
    // if itemp>0, target is a series of layers; if itemp<0, target is described by a bitmap
    if(itemp>=0) {
        // usual layer structure
        numberLayers = (unsigned int) itemp;
        isBitmap = 0;
        targetBottom = 0;
    } else {
        // target described by bitmap (and in this case, layers are actually the different compounds in the bitmap)
        numberLayers = (unsigned int) (-itemp);
        isBitmap = 1;
        // read bitmap dimensions, which layer is the substrate, and pixel-to-Angstroms conversion factor
        fscanf(ifp, "%lu %lu %lu %i %f", &xDim, &yDim, &zDim, &substrateLayer, &pixelsPerAngs);
        //printf("pixelsPerAngs=%f\n",pixelsPerAngs);
        if(pixelsPerAngs==0) {
            fprintf(stderr, "nError: pixels per Angstroms set to 0.\n");
            return 1;
        }
        AngsPerPixel = 1.0f/pixelsPerAngs;
        bitmapThichness  =  xDim*AngsPerPixel;
        targetBottom = bitmapThichness;
    }

    // bound checking
    if(numberLayers>MAXLAYERS || numberLayers<1) {
        fprintf(stderr, "\nError: number of layers is %u, but must be between 1 and %u.\n", numberLayers, MAXLAYERS);
        return 1;
    }

    // layer and foil description
    for(ilayer=0; ilayer<=MAXLAYERS; ilayer++)
        for(ielem=0; ielem<MAXELEMENT; ielem++)
            element[ilayer][ielem] = NULL;

    for(ilayer=0; ilayer<=numberLayers; ilayer++) {
        skipComments(ifp);
        fscanf(ifp, "%g %g %g %u %lu",
               &layerThickness0[ilayer], &layerThickDist[ilayer], &layerDensity[ilayer],
               &numberElements[ilayer], &numProjectileLayer[ilayer] );
        if(!isBitmap || ilayer==substrateLayer) targetBottom += layerThickness0[ilayer];

        // bound checking
        if(numberElements[ilayer]>MAXELEMENT || numberElements[ilayer]<1) {
            if(ilayer<numberLayers)
                fprintf(stderr, "\nError: number of elements in layer %u must be between 1 and %u.\n", ilayer, MAXELEMENT);
            else
                fprintf(stderr, "\nError: number of elements in foil must be between 1 and %i.\n", MAXELEMENT);
            return 1;
        }

        // compute interatomic distance, and use it as mean free path unless it was set by user
        interatomicDist[ilayer] = d2f(pow(3./(4*PI*layerDensity[ilayer]),1./3.));
        sqrtLayerDensityPi[ilayer] = sqrtf(layerDensity[ilayer]*PI);

        if(meanFreePath==0) sqrtMeanFreePath[ilayer] = sqrtf(interatomicDist[ilayer]);
        else sqrtMeanFreePath[ilayer] = sqrtf(meanFreePath);

        sum = 0.;
        // for each element in the layer
        for(ielem=0; ielem<numberElements[ilayer]; ielem++) {
            // Z, mass, fraction, displacement energy, binding energy
            skipComments(ifp);
            fscanf(ifp, "%u %g %g %g %g %u %u", &Z, &mass, &elementFraction[ilayer][ielem],
                   &Ed[ilayer][ielem], &Eb[ilayer][ielem], &natural[ilayer][ielem], &nsim[ilayer][ielem]);
            if(nsim[ilayer][ielem]>MAXELEMENT) {
                fprintf(stderr, "\nError: number of isotopes to simulate for element %u in layer %u must be between 0 and %u.\n",
                        ielem, ilayer, MAXELEMENT);
                return 1;
            }
            temp = 0;
            for(j=0; j<nsim[ilayer][ielem]; j++) {
                fscanf(ifp, "%g %g %i", &abundance[ilayer][ielem][j], &massIsotope[ilayer][ielem][j], &CScorrType[ilayer][ielem][j]);
                temp += abundance[ilayer][ielem][j];
            }
            if(layerThickness0[ilayer]>0 && (temp>1.001 || temp < 0.99))
                fprintf(stderr, "\nWarning: total isotopic abunance of element %u in layer %u is %f, which is different from 1.\n",
                        ielem, ilayer, temp);

            if(mass==0.) mass = AtomicMass[Z];  // default is average atomic mass
            // NOTE: this mass is set here to pre-compute the stopping and scattering matrices (cos & sin) before the simulation
            //       but is then modified when simulating the different isotopes
            sum += elementFraction[ilayer][ielem];
            assignElement(Z, mass, ilayer, ielem); // set pointer to already existing element structure or allocate a new one
        }

        // concentration normalization and fraction sum calculation
        for(ielem=0; ielem<numberElements[ilayer]; ielem++) {
            elementFraction[ilayer][ielem]/=sum;
            if(ielem==0) elementFractionSum[ilayer][ielem]=elementFraction[ilayer][ielem];
            else elementFractionSum[ilayer][ielem]=elementFractionSum[ilayer][ielem-1]+elementFraction[ilayer][ielem];
        }

        // H stopping needed to find charge state in straggling calculation
        hsp[ilayer] = fillDedxTable(ilayer, 1, 1.007f, layerDensity[ilayer], 1.0f );
        if(!hsp[ilayer]) {
            fprintf(stderr, "\nError: can't load hydrogen stopping power table\n");
            return 1;
        }

    }

    // detector parameters (dummy values in case of implantation)
    skipComments(ifp);
    for(i=0;i<NUMDET;i++) {
        // is the dectector an ellipse (!=0) or a rectangle (==0)
        fscanf(ifp, "%i", &isEllipse[i]);

        // coordinates of 3 points in space defining a rectangle containing the detector surface [A]
        fscanf(ifp, "%g %g %g", &p1x[i], &p1y[i], &p1z[i]);
        fscanf(ifp, "%g %g %g", &p2x[i], &p2y[i], &p2z[i]);
        fscanf(ifp, "%g %g %g", &p3x[i], &p3y[i], &p3z[i]);

        //fprintf(stderr, "d=[ %g, %g, %g,   %g, %g, %g,   %g, %g, %g, 0,0,0]\n", p1x[i], p1y[i], p1z[i], p2x[i], p2y[i], p2z[i], p3x[i], p3y[i], p3z[i]);

        // width and height multiplying factor for virtual detector, annulus radius [A]
        fscanf(ifp, "%g", &r2annular[i]);
        r2annular[i]*=r2annular[i];

        // compute normal vectors defining reference frame axes of this detector
        // zp = (p2-p1)/|p2-p1|
        zpx[i] = p2x[i]-p1x[i];
        zpy[i] = p2y[i]-p1y[i];
        zpz[i] = p2z[i]-p1z[i];
        temp = sqrtf(1.0/(zpx[i]*zpx[i]+zpy[i]*zpy[i]+zpz[i]*zpz[i]));
        binv[i] = 2.0f*temp; // 1/half-size of detector along detector z axis
        zpx[i] *= temp;
        zpy[i] *= temp;
        zpz[i] *= temp;

        // yp = (p3-p2)/|p3-p2|
        ypx[i] = p3x[i]-p2x[i];
        ypy[i] = p3y[i]-p2y[i];
        ypz[i] = p3z[i]-p2z[i];
        temp = sqrtf(1.0/(ypx[i]*ypx[i]+ypy[i]*ypy[i]+ypz[i]*ypz[i]));
        ainv[i] = 2.0f*temp; // 1/half-size of detector along detector y axis
        ypx[i] *= temp;
        ypy[i] *= temp;
        ypz[i] *= temp;

        // check that zp and yp are perpendicular
        temp = zpx[i]*ypx[i]+zpy[i]*ypy[i]+zpz[i]*ypz[i];
        if(temp>1e-4f) {
            fprintf(stderr,"\nError: coordinates of detector %lu not properly defined. Sides not perpendicular. yp.zp = %g > 1E-4\n",i, temp);
            return 1;
        }

        // xp = yp x zp
        xpx[i] = ypy[i]*zpz[i]-ypz[i]*zpy[i];
        xpy[i] = ypz[i]*zpx[i]-zpz[i]*ypx[i];
        xpz[i] = ypx[i]*zpy[i]-ypy[i]*zpx[i];

        //fprintf(stderr, ">> %g %g %g\n>> %g %g %g\n>> %g %g %g\n", xpx[i], xpy[i], xpz[i], ypx[i], ypy[i], ypz[i], zpx[i], zpy[i], zpz[i]); //xyz

        // detector center
        pcx[i] = (p3x[i]+p1x[i])/2.0f;
        pcy[i] = (p3y[i]+p1y[i])/2.0f;
        pcz[i] = (p3z[i]+p1z[i])/2.0f;

        // compute d in detector plane equation xpx*x+xpy*y+xpz*z+d=0
        d[i] = -(xpx[i]*p1x[i]+xpy[i]*p1y[i]+xpz[i]*p1z[i]); // d=-xp.p1

        //  check if detector is pointing towards target
        if(pcx[i]/xpx[i]>0) {
            fprintf(stderr,"\nError: detector %lu pointing away from target.\n",i);
            return 1;
        }
        temp = 1.0f/sqrtf(pcx[i]*pcx[i]+pcy[i]*pcy[i]+pcz[i]*pcz[i]);
        pcnormx[i] = pcx[i]*temp;
        pcnormy[i] = pcy[i]*temp;
        pcnormz[i] = pcz[i]*temp;

        // coordinates of points 1 and 3 in detector reference frame
        p1yp[i] = (p1x[i]-pcx[i])*ypx[i]+(p1y[i]-pcy[i])*ypy[i]+(p1z[i]-pcz[i])*ypz[i];
        p1zp[i] = (p1x[i]-pcx[i])*zpx[i]+(p1y[i]-pcy[i])*zpy[i]+(p1z[i]-pcz[i])*zpz[i];

        p3yp[i] = (p3x[i]-pcx[i])*ypx[i]+(p3y[i]-pcy[i])*ypy[i]+(p3z[i]-pcz[i])*ypz[i];
        p3zp[i] = (p3x[i]-pcx[i])*zpx[i]+(p3y[i]-pcy[i])*zpy[i]+(p3z[i]-pcz[i])*zpz[i];

    }

    // virtual detector enlargement factors
    skipComments(ifp);
    fscanf(ifp, "%g %g", &Wf, &Hf);
    if(Wf!=1.0f || Hf!=1.0f) {
      if( simulationType==TECH_COINC ) {
        fprintf(stderr, "\nWarning: virtual detection not implemented for coincidence. Factors set to 1.\n");
        Wf = 1.0f;
        Hf = 1.0f;
      } else if( simulationType==TECH_FULLRBS ) {
          fprintf(stderr, "\nWarning: using virtual detector in Full RBS actually only increases detector size. No trajectory correction.\n");
      }
    }

    // load compound correction coefficients
    skipComments(ifp);
    switch (simulationType) {
    case TECH_IMPREC:
    case TECH_ERD:
    case TECH_ERD_TOF:
    case TECH_COINC:
        for(i=0;i<numTravel;i++) {
            // read Z, type of cross-section (0: Rutherford, 1: Tabulated), and compound corr. values
            fscanf(ifp, "%u", &Z);
            for(ilayer=0; ilayer<numberLayers; ilayer++)
                fscanf(ifp, "%g", &tempcc[ilayer]);

            // apply it to the beam if same Z
            if(Z==beam.Z) {
                for(ilayer=0; ilayer<numberLayers; ilayer++)
                     beam.compoundCorr[ilayer] = tempcc[ilayer];
                beam.compoundCorr[numberLayers] = 1.0f; // compound correction of absorber always 1
            }
            // then check for all other elements
            // (a set of coefficients may apply to more than one element, including the beam!!)
            for(ilayer=0; ilayer<numberLayers; ilayer++)
                for(ielem=0; ielem<numberElements[ilayer]; ielem++)
                    if(element[ilayer][ielem] != NULL && element[ilayer][ielem]->Z == Z) {
                        for(j=0; j<numberLayers; j++)
                             element[ilayer][ielem]->compoundCorr[j] = tempcc[j];
                        // compound correction of absorber always 1
                        element[ilayer][ielem]->compoundCorr[numberLayers] = 1.0f;
                    }
        }
        break;
    case TECH_IMPLANT:
    case TECH_ENERDEPTHMAP:
    case TECH_ENERDEPTHLIST:
    case TECH_RBS:
    case TECH_RBS_TOF:
    case TECH_FULLRBS:
        // read only one value, for the beam
        fscanf(ifp, "%u", &Z);
        if(Z!=beam.Z) {
            fprintf(stderr,"\nError reeding compound corr.: Z=%u not equal to beam Z=%u\n",Z, beam.Z);
            return 1;
        }
        for(ilayer=0; ilayer<numberLayers; ilayer++)
            fscanf(ifp, "%g", &(beam.compoundCorr[ilayer]) );
        beam.compoundCorr[numberLayers] = 1.0f; // compound correction of absorber always 1
        break;
    default:
        fprintf(stderr, "\nSimulation type %d unknown.\n", simulationType);
        return 1;
    }

    // end of file input
    fclose(ifp);

    EminErr=0; EmaxErr=0;
    SminErr=0; SmaxErr=0;
    DminErr=0; DmaxErr=0;
    //printf("numTravel=%i\n",numTravel);

    return 0;

}


// write simulation parameters, each line preceeded with '#' to filter them out easily in matlab, python, etc.
// (both to be able to check if parameters were read correctly and so we keep parameters with the output data;
// would be better to use XML but difficult to implement in plain C and in a platform-independent way)
int copyinput(FILE *ofp ) {

    // other working variables
    unsigned int i, ilayer, ielem, Z, numTravel = 0, roundMass;
    unsigned int j, Zalready[NUMBERELEMENTS];

    for(i=0; i<NUMBERELEMENTS; i++)
        Zalready[i]=0;

    for(ilayer=0; ilayer<numberLayers; ilayer++)
        for(ielem=0; ielem<numberElements[ilayer]; ielem++)
            numTravel += nsim[ilayer][ielem];

    if(ofdetect == NULL)
        return 1;

    // simulation parameters
    fprintf(ofp,"#%% technique, mean free path, cone angle [deg], minimum energy [eV], energy straggling, writeDetails, screening, cutoff angle, random seed, nthreads\n");
    fprintf(ofp, "#%d %g %g %g %u %u %u %g %ld %u\n", simulationType, (doRotate?1:-1)*powf(sqrtMeanFreePath[0],2), acos(cosAngSpread0)*180./PI,
            minE0, enerStragType,  writeDetails, screen, acos(coscutoffAngle)*180./PI, getseed1(), nthread);

    // beam parameters
    fprintf(ofp,"#%% beam Z, atomic mass, number of ions, beam energy [eV], beam radius [A]\n");
    fprintf(ofp, "#%u %f %lu %g %g\n", beam.Z, beam.mass, numProjectile, E0, ySpotSize*sqrtf(1.0-m0*m0));

    // directional cosines of incident beam
    fprintf(ofp, "#%% incident beam directional cosines\n");
    fprintf(ofp, "#%g %g %g\n", l0, m0, n0);

    // total number of layers, total number of different ions for which the trajectory is followed
    fprintf(ofp, "#%% number of layers, number of traveling atoms\n");
    if(isBitmap)
        fprintf(ofp, "#-%u %u   %lu %lu %lu %i %f\n", numberLayers, numTravel, xDim, yDim, zDim, substrateLayer, pixelsPerAngs);
        // (note the minus sign in front of unsigned format string for unsigned variable 'numberLayers')
    else
        fprintf(ofp, "#%u %u\n", numberLayers, numTravel);

    for(ilayer=0; ilayer<=numberLayers; ilayer++) {
        fprintf(ofp, "#%% layer thickness, roughness, atomic density, number of elements, number of projectiles\n");
        fprintf(ofp, "#%g %g %g %u %lu\n",
               layerThickness0[ilayer], layerThickDist[ilayer], layerDensity[ilayer],
               numberElements[ilayer], numProjectileLayer[ilayer] );

        // for each element in the layer
        for(ielem=0; ielem<numberElements[ilayer]; ielem++) {
            // Z, mass, fraction, displacement energy, binding energy
            Z = element[ilayer][ielem]->Z;
            fprintf(ofp,"#%% Z, mass, proportion, Edisp, Ebind, simNaturalAbundance, N isotopes to simulate, N*[fraction, mass, use CS table]\n");
            fprintf(ofp, "#%u %g %g %g %g %u %u", Z, element[ilayer][ielem]->mass, elementFraction[ilayer][ielem],
                   Ed[ilayer][ielem], Eb[ilayer][ielem], natural[ilayer][ielem], nsim[ilayer][ielem]);
            for(j=0; j<nsim[ilayer][ielem]; j++) {
                roundMass = (int)round(massIsotope[ilayer][ielem][j]);
                fprintf(ofp, " %g %g", abundance[ilayer][ielem][j], massIsotope[ilayer][ielem][j]);
            }
            fprintf(ofp,"\n");
        }
    }

    // detector parameters
    if(simulationType>0 || simulationType==TECH_FULLRBS) {
        fprintf(ofp,"#%% detectors: isEllipse p1x p1y p1z p2x p2y p2z p3x p3y p3z r_annulus\n");
        for(i=0;i<NUMDET;i++) {
            // is the dectector an ellipse (!=0) or a rectangle (==0)
            fprintf(ofp, "#%i", isEllipse[i]);

            // coordinates of 3 points in space defining a rectangle containing the detector surface [A]
            fprintf(ofp, " %g %g %g", p1x[i], p1y[i], p1z[i]);
            fprintf(ofp, " %g %g %g", p2x[i], p2y[i], p2z[i]);
            fprintf(ofp, " %g %g %g", p3x[i], p3y[i], p3z[i]);

            // width and height multiplying factor for virtual detector, annulus radius [A]
            fprintf(ofp, " %g\n", sqrtf(r2annular[i]));

        }

        // virtual detector enlargement factors
        fprintf(ofp, "#%% virtual detector enlargement factors: width, height\n");
        fprintf(ofp, "#%g %g\n", Wf, Hf);

        // write compound correction coefficients
        fprintf(ofp,"#%% ion Z, and its compound correction in each layer\n");
        switch (simulationType) {
        case TECH_IMPREC:
        case TECH_ERD:
        case TECH_ERD_TOF:
        case TECH_COINC:
            // write Z and compound corr. values for the beam
            fprintf(ofp, "#%u", beam.Z);
            for(ilayer=0; ilayer<numberLayers; ilayer++)
                fprintf(ofp, " %g", beam.compoundCorr[ilayer]);
            Zalready[beam.Z]=1;
            fprintf(ofp, "\n");

            // then check if we need to write them for the other elements
            for(ilayer=0; ilayer<numberLayers; ilayer++)
                for(ielem=0; ielem<numberElements[ilayer]; ielem++) {
                    Z = element[ilayer][ielem]->Z;
                    if(nsim[ilayer][ielem] > 0 && Zalready[Z]==0) {
                        fprintf(ofp, "#%u", Z);
                        for(i=0; i<numberLayers; i++)
                            fprintf(ofp, " %g", element[ilayer][ielem]->compoundCorr[i]);
                        fprintf(ofp, "\n");
                        Zalready[Z] = 1;
                    }
                }
            break;
        case TECH_IMPLANT:
        case TECH_ENERDEPTHMAP:
        case TECH_ENERDEPTHLIST:
        case TECH_RBS:
        case TECH_RBS_TOF:
        case TECH_FULLRBS:
            // write Z and compound corr. values for the beam only
            fprintf(ofp, "#%u", beam.Z);
            for(ilayer=0; ilayer<numberLayers; ilayer++)
                fprintf(ofp, " %g", beam.compoundCorr[ilayer]);
            fprintf(ofp, "\n");
            break;
        default:
            fprintf(stderr, "\nSimulation type %d unknown.\n", simulationType);
            return 1;
        }

    }

    return 0;

}


int fillAllTables() {

    unsigned int ilayer, ielem, j;

    // alllocate tables depending on technique
    followRecoils = 0; // default value
    switch (simulationType) {
    case TECH_IMPREC:

        followRecoils = 1;

        for(ilayer=0; ilayer<NUMBERELEMENTS; ilayer++) {
            dispHist[ilayer] = NULL;
            implHist[ilayer] = NULL;
        }

        for(ilayer=0; ilayer<=MAXLAYERS; ilayer++) {
            for(ielem=0; ielem<MAXELEMENT; ielem++) {
                if(fillTables(element[ilayer][ielem])==0) { // assume no compound correction for recoils during implantation
                    if(dispHist[element[ilayer][ielem]->Z]==NULL) {
                        dispHist[element[ilayer][ielem]->Z] = calloc(NUMBINPROF,sizeof(double)); // allocate and initialize to 0
                        implHist[element[ilayer][ielem]->Z] = calloc(NUMBINPROF,sizeof(double)); // allocate and initialize to 0
                    }
                }
            }
        }

    case TECH_ERD:
    case TECH_ERD_TOF:
    case TECH_COINC:
        for(ilayer=0; ilayer<MAXLAYERS; ilayer++) {
            for(ielem=0; ielem<MAXELEMENT; ielem++) {
                if(nsim[ilayer][ielem]) {
                    recoil = element[ilayer][ielem];
                    fillTables(recoil);
                }
            }
        }
        // NO BREAK
    case TECH_IMPLANT:
    case TECH_RBS:
    case TECH_RBS_TOF:
    case TECH_FULLRBS:
    case TECH_ENERDEPTHMAP:
    case TECH_ENERDEPTHLIST:
        fillTables(&beam);

        if(simulationType == TECH_IMPLANT && beamImplHist==NULL)
            beamImplHist = calloc(NUMBINPROF,sizeof(double)); // allocate and initialize to 0
        else if(simulationType == TECH_ENERDEPTHMAP && energy_depth_map==NULL)
            energy_depth_map = calloc(NUMBINPROF*NUMBINPROF,sizeof(unsigned int)); // allocate and initialize to 0
        break;
    default:
        fprintf(stderr, "\nSimulation type %d unknown.\n", simulationType);
        return 1;
    }
    return 0;

}


int runSimul(unsigned long numProj, unsigned int collLayer) {

    float temp, temp2, sum;
    clock_t t1;
    unsigned long totalnumbercoll=0, ichan, ielem; // number of multiple scattering collisions

#ifdef MAX_THREAD
    unsigned int ithread;
#endif

    // structures to pass parameters to mainloop
#ifdef MAX_THREAD
    unsigned int nth;
    thread_params param[MAX_THREAD];
    pthread_t threads[MAX_THREAD];
    pthread_mutex_init (&mut, NULL);
#else
    thread_params param[1];
#endif

    // set counters to 0
    detectedParticles = 0;
    exitBottom = 0;
    belowCutoff = 0;
    exitUndetected = 0;
    badRescale=0;
    collVacuum=0;
    exitUndetectedFinal = 0;
    exitTop = 0;
    exitMinEnergy = 0;
    badKinematics=0;
    moreThanTwoMaterials = 0;

    t1=clock();

    // pass threadf specific arguments and start mainloop
#ifdef MAX_THREAD
    nth = nthread;
    if(numProj<100) nth=1;

    // start nthread threads
    for (ithread=0; ithread<nth; ithread++) {
        param[ithread].id=ithread;
        param[ithread].numProjectile   = numProj/nth;
        param[ithread].totalnumbercoll = 0;
        param[ithread].collisionLayer = collLayer;

        pthread_create(&threads[ithread], NULL, (void *)&mainloop, (void *)(param+ithread));
    }

    // wait for threads completionF
    for (ithread=0; ithread<nth; ithread++)	{
        pthread_join(threads[ithread],NULL);
    }

    // number of collisions
    totalnumbercoll = 0;
    for (ithread=0; ithread<nth; ithread++)	{
        totalnumbercoll+=param[ithread].totalnumbercoll;
    }
#else
    // no threads (or only one!)
    param[0].id=0;
    param[0].numProjectile   = numProj;
    param[0].totalnumbercoll = 0;
    param[0].collisionLayer = collLayer;

    mainloop(&param);

    totalnumbercoll=param[0].totalnumbercoll;

#endif

    // simulation finished
    temp = (float)(clock()-t1)/(float)CLOCKS_PER_SEC;
    fprintf(stdout, "\nDone in %.0f s, %.1f coll/ion, %.1f us/ion, %.1f ns/coll\n", temp,
            (float)totalnumbercoll/(float)numProj, temp*1.e6/(float)numProj,temp*1.e9/(float)totalnumbercoll );

    // Multiply histogram by density*thickness*solidAngle/number_of_detected_particle, with
    //   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle,
    //   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)
    // The spectrum then needs to be multipllied by the number of incident ions q and the concentration
    // of the detected element in the layer. See documentation for a more detailed explanation.
    if(simulationType>0 || simulationType == TECH_FULLRBS) {

        // result of summations
        fprintf(stdout, "%13lu total incident ions\n", numProj);
        fprintf(stdout, "%13lu detected atoms\n", detectedParticles);
        fprintf(stdout, "%13lu particles emerged at surface but were undetected\n", exitUndetected);
        fprintf(stdout, "%13lu main collisions below cutoff angle\n", belowCutoff);
        fprintf(stdout, "%13lu particles entered detection system but were finally not detected\n", exitUndetectedFinal);
        fprintf(stdout, "%13lu incomming particles abnormally exited at top\n", exitTop);
        fprintf(stdout, "%13lu particles exited at target bottom\n", exitBottom);
        fprintf(stdout, "%13lu particles implanted or below channel 0\n", exitMinEnergy);
        fprintf(stdout, "%13lu particles generated in a direction not allowed by kinetics\n", badKinematics);
        fprintf(stdout, "%13lu particles not detected after correction for virtual detector\n", badRescale);
        fprintf(stdout, "%13lu main collisions with vacuum in pixel target\n\n", collVacuum);
        fprintf(stdout, "%13lu pixels in a 3rd material between 2 collisions\n", moreThanTwoMaterials);

    } else {

        if(simulationType == TECH_IMPLANT || simulationType == TECH_IMPREC) {
            // divide the distribution by the number of projectiles
            // and bin size, so we get a distribution in at/A/ion
            temp = 1.0f/(numProj*(targetBottom/NUMBINPROF));
            for(ichan=0;ichan<NUMBINPROF;ichan++) beamImplHist[ichan] *= temp;

            if(followRecoils) {
                for(ielem=1;ielem<NUMBERELEMENTS;ielem++) {
                    if(implHist[ielem]) {
                        for(ichan=0;ichan<NUMBINPROF;ichan++) implHist[ielem][ichan] *= temp;
                    }
                    if(dispHist[ielem]) {
                        for(ichan=0;ichan<NUMBINPROF;ichan++) dispHist[ielem][ichan] *= temp;
                    }
                }
            }

            // compute moments for implanted beam
            sum = 0;
            for(ichan=0;ichan<NUMBINPROF;ichan++) sum+=beamImplHist[ichan];

            temp = 0; // mean range
            for(ichan=0;ichan<NUMBINPROF;ichan++) temp+=ichan*targetBottom/NUMBINPROF*beamImplHist[ichan];
            temp /= sum;

            temp2 = 0; // range straggling
            for(ichan=0;ichan<NUMBINPROF;ichan++) temp2+=powf(ichan*targetBottom/NUMBINPROF-temp,2)*beamImplHist[ichan];
            temp2 /= sum;
            temp2 = sqrtf(temp2);
        }

        // output results
        fprintf(stdout, "%13lu total incident ions\n", numProj);
        fprintf(stdout, "%13lu implanted ions\n", exitMinEnergy);
        fprintf(stdout, "%13lu backscattered ions\n", exitTop);
        fprintf(stdout, "%13lu transmitted ions\n", exitBottom);
        fprintf(stdout, "beam mean range = %g A\n", temp);
        fprintf(stdout, "beam straggling = %g A\n", temp2);

        if(detectedParticles)
            fprintf(stderr, "\n%13lu detected atoms (no ion should be detected during implantation).\n", detectedParticles);
    }
    return 0;
}


// Allocate *bitmap and store 'corteo.bitmap' into it.
// (The file is actually filenameWithoutExt + ".bitmap")
// Return:
// 0 on success
// 1 if cannot open 'corteo.bitmap'
// 2 if cannot allocate *bitmap
// 3 if wrong number of bytes read from 'corteo.bitmap'
// 4 if a byte value in 'corteo.bitmap' is larger than the number of materials (i.e. layers)
int loadBitmap() {

    char bitmapfile[FILENAMESIZE];
    char *inp;
    FILE *fp;
    unsigned long i, n0, n, ix, iy, iz;

    sprintf(bitmapfile,"%s.bitmap", filenameWithoutExt);
    fp = fopen(bitmapfile, "r");

    if(fp == NULL) {
        fprintf(stderr, "\nCannot open '%s'.\n", bitmapfile);
        return 1;
    }

    bitmap = (unsigned char *) malloc(sizeof(char)*xDim*yDim*zDim+10);

    if(bitmap == NULL) {
        fprintf(stderr, "\nCannot allocate memory for bitmap.\n");
        return 2;
    }

    // In order to ease debugging et al., the bitmap file is actually a text file where material numbers
    // (in this mode, a layer is actually a material) appear as their ASCII character (i.e. material 1
    // appears as '1', ASCII code 49. An exception is vacuum, which shows up as a space (' ', ASCII 32)
    // and must be converted to code 255 (UCHAR_MAX).
    // It is notable that in this system, if there are more than 10 layers, we continue with the
    // ASCII code subsequent to '9'=57. When reaching code 127, this is the 'DEL' character, so
    // we are probably bound for troubles if there are more than 78 different materials in the bitmap.
    //
    i = 0; // line counter
    for(iz=0; iz<zDim; iz++) {
        for(iy=0; iy<yDim; iy++) {
            i++; // line counter
            n0 = (iz*yDim+iy)*xDim;
            inp = fgets((char *)bitmap+n0 , xDim+10, fp);
            if (inp[xDim] == '\n')
                inp[xDim] = '\0';
            if( strlen(inp) != xDim ) {
                    fprintf(stderr, "\nAt line %lu of file '%s', expected to read %lu bytes but got %lu'.\n",
                            i, bitmapfile, xDim, strlen(inp) );
                    return 3;
            }
            for(ix=0; ix<xDim; ix++) {
                n = n0+ix;
                bitmap[n] = (bitmap[n] == ' ') ? UCHAR_MAX : bitmap[n]-48; // convert ' ' to 255, else convert '0' to 0, etc.
                if(bitmap[n] != 255 && bitmap[n]>=numberLayers) {
                    fprintf(stderr, "\nIn file '%s', the byte #%lu of line %lu as an index value %i >= number of layer %u.\n",
                            bitmapfile, ix, i, bitmap[n], numberLayers);
                    return 4;
                }
            }
        }
    }

    return 0;

}


// call input(), then run mainloop, and writes spectrum
// convention: energies in eV, lengths in A, densities in at/A3 (=10^24 at/cm4)
int main(int argc, char *argv[] ) {

    // working variables
    int ret;
    unsigned int j, ilayer, ielem, iSimul, numSimul, ichan, Echan;
    char filename[FILENAMESIZE];
    char *modew = "w";
    FILE *ofp;

    if(checkType()!=0) return checkType();

    if(argc>3 || (argc==2 && atoi(argv[1])==0) ) {
        fprintf(stderr, "\nUsage: %s [random_integer [input/output_file_without_ext]]\n", argv[0]);
        return 1;
    }

    filenameWithoutExt = NULL;
    if(argc==3) {
        // set name for input and output file to the 2nd argument
        j = strlen(argv[2]);
        if(j>FILENAMESIZE-50) {
            fprintf(stderr, "\nName (including path) provided for input/output files is more than %i characters. Name is:\n%s\n", FILENAMESIZE-50, argv[2]);
            return 1;
        }
        filenameWithoutExt = calloc(j+2, sizeof(char));
        if(filenameWithoutExt) strcpy(filenameWithoutExt, argv[2]);
    } else {
        // set name for input and output file to the default value 'corteo'
        filenameWithoutExt = calloc(10, sizeof(char));
        if(filenameWithoutExt) strcpy(filenameWithoutExt, "corteo");
    }
    if(filenameWithoutExt==NULL) {
        fprintf(stderr, "\nAllocation error for input/output file name without extention.\n");
        return 1;
    }

    // read filenameWithoutExt + ".in", set parameters
    if( (ret=input())!=0 ) return ret;

#ifdef MAX_THREAD
    fprintf(stdout, "%i thread(s).\n", nthread);
#else
    fprintf(stdout, "1 thread.\n");
#endif

    if(argc>=2) randomxinit(0, atoi(argv[1]));
    fprintf(stdout, "Random seeds = %li %li\n", getseed1(), getseed2());

    // open required detail files
    if(simulationType>0 || simulationType==TECH_FULLRBS || simulationType == TECH_ENERDEPTHLIST) {
        if(argc>1) {
            sprintf(filename,"%s.detect.%s", filenameWithoutExt, argv[1]);
            ofdetect = fopen(filename, modew);
        } else {
            sprintf(filename,"%s.detect", filenameWithoutExt);
            ofdetect = fopen(filename, modew);
        }
        copyinput(ofdetect);
        if(simulationType>0) {

            // write detected particle data
            fprintf(ofdetect, "#");
            if(simulationType == TECH_COINC) {
                fprintf(ofdetect, "E_2nd_part_coin ");
            }
            if(simulationType == TECH_RBS_TOF || simulationType == TECH_ERD_TOF) {
                fprintf(ofdetect, "TOF[ns] ");
            }

            fprintf(ofdetect, "E crosssect Nt_Omega/Num_det_part E_before_coll cos_coll_angle cos_angle_det0 colldepth Z2 collision_layer CS_stamp ");

            if(writeDetails) {
                fprintf(ofdetect,
                        "minimal_approach_dist_coll E_after_col  "
                        "x_coll y_coll z_coll  l_before_coll m_before_coll n_before_coll  l_after_coll m_after_coll n_after_coll  "
                        "x_exit y_exit z_exit  l_exit m_exit n_exit  "
                        "x_det y_det z_det  l_det m_det n_det  ");

                if(simulationType == TECH_COINC) {
                    fprintf(ofdetect,
                            "E_after_col_coin l_before_coll_coin m_before_coll_coin n_before_coll_coin"
                            "x_exit_coin y_exit_coin z_exit_coin  l_exit_coin m_exit_coin n_exit_coin  "
                            "x_det_coin y_det_coin z_det_coin  l_det_coin m_det_coin n_det_coin ");
                }

                if(absorberThick) {
                    fprintf(ofdetect,
                            "x_abs y_abs z_abs  l_abs m_abs n_abs  "
                            "x_tof y_tof z_tof  l_tof m_tof n_tof E_before_abs  ");
                }

            }
        } else if(simulationType == TECH_FULLRBS) {
            fprintf(ofdetect, "#Energy min_cos_coll_angle");
            if(writeDetails) {
                fprintf(ofdetect,
                        " x_exit y_exit z_exit  l_exit m_exit n_exit  "
                        "x_det y_det z_det  l_det m_det n_det ");
                fprintf(ofdetect, "  NUMMINCOSSCAT");
                for(j=0; j<NUMMINCOSSCAT; j++)
                    fprintf(ofdetect, "  minCosScat_%u minCosScatDepth_%u", j, j);
            }
        } else if (simulationType == TECH_ENERDEPTHLIST) {
            fprintf(ofdetect, "#Energy x");
            if(writeDetails)
                fprintf(ofdetect, " y z l m n");

        }
        fprintf(ofdetect, "\n");
        fflush(ofdetect);

    }

    switch (simulationType) {
    case TECH_IMPLANT:
        fprintf(stdout, "Simulating implantation of ");
        break;
    case TECH_IMPREC:
        fprintf(stdout, "Simulating full cascade of ");
        break;
    case TECH_ENERDEPTHMAP:
        fprintf(stdout, "Simulating energy-depth map of ");
        break;
    case TECH_ENERDEPTHLIST:
        fprintf(stdout, "Simulating energy-depth list of ");
        break;
    case TECH_RBS:
        fprintf(stdout, "Simulating RBS spectrum of ");
        break;
    case TECH_RBS_TOF:
        fprintf(stdout, "Simulating RBS-TOF spectrum of ");
        break;
    case TECH_FULLRBS:
        fprintf(stdout, "Simulating full RBS spectrum of ");
        break;
    case TECH_ERD:
        fprintf(stdout, "Simulating ERD spectrum of ");
        break;
    case TECH_ERD_TOF:
        fprintf(stdout, "Simulating ERD-TOF spectrum of ");
        break;
    case TECH_COINC:
        fprintf(stdout, "Simulating 2-detectors coincidence of ");
        break;
    default:
        fprintf(stderr, "\nSimulation type %d unknown.\n", simulationType);
        return 1;
    }

    if(E0<1e3f)
        fprintf(stdout, "%.2f eV ",E0);
    else if(E0<1e6f)
        fprintf(stdout, "%.2f keV ",E0/1e3f);
    else
        fprintf(stdout, "%.2f MeV ",E0/1e6f);

    fprintf(stdout, "%u%s\n",beam.roundMass, AtomicNames[beam.Z]);

    if(simulationType>0 || simulationType==TECH_FULLRBS) {
        fprintf(stdout, "beam incident at %.1f deg. and detector at %.1f deg. from surface normal.\n",
                acos(l0)*180./PI,acos(xpx[0])*180./PI);
        if(simulationType==TECH_COINC)
            fprintf(stdout, "  Second detector at %.1f deg. from surface normal.\n",acos(xpx[1])*180./PI);
        fprintf(stdout, "beam and detector %.1f deg. apart.\n",acos(l0*xpx[0]+m0*xpy[0]+n0*xpz[0])*180./PI);
        //fprintf(stderr, "%f %f %f %f %f %f\n", l0, m0, n0, xpx[0], xpy[0], xpz[0]); //xyz
        if(simulationType==TECH_COINC)
            fprintf(stdout, "  Beam and second detector %.1f deg. apart.\n",acos(l0*xpx[1]+m0*xpy[1]+n0*xpz[1])*180./PI);
    } else {
        fprintf(stdout, "beam incident at %.1f deg. from surface normal.\n",acos(l0)*180./PI);
    }

    // print layer composition
    for(ilayer=0; ilayer<numberLayers; ilayer++) {
        if(isBitmap) {
            if(ilayer == substrateLayer)
                fprintf(stdout, "\nSubstrate (layer %u): %6.1f+/-%6.1f A of", ilayer, layerThickness0[ilayer], layerThickDist[ilayer]);
            else
                fprintf(stdout, "\nPixmap material %u:x", ilayer);
        } else {
            fprintf(stdout, "\nLayer %u: %6.1f+/-%6.1f A of", ilayer, layerThickness0[ilayer], layerThickDist[ilayer]);
        }
        for(ielem=0; ielem<numberElements[ilayer]; ielem++) {
            if(elementFraction[ilayer][ielem]>=0.01f)
                fprintf(stdout, " %3.0f%2s (%5.1f%%)", element[ilayer][ielem]->mass, AtomicNames[element[ilayer][ielem]->Z], elementFraction[ilayer][ielem]*100);
            else
                fprintf(stdout, " %3.0f%2s (%5g%%)", element[ilayer][ielem]->mass, AtomicNames[element[ilayer][ielem]->Z], elementFraction[ilayer][ielem]*100);
        }
        fprintf(stdout, ", N=%5.3fe22/cm3, Nt=%0.fe15/cm2", layerDensity[ilayer]*100, layerThickness0[ilayer]*layerDensity[ilayer]*10.);
    }

    if(isBitmap) {
        if(zDim<2)
            fprintf(stdout, "\n2D pixmap %lu pixels high and %lu pixels wide, or %g Angs high by %g Angs wide.", xDim, yDim, xDim*AngsPerPixel, yDim*AngsPerPixel);
        else
            fprintf(stdout, "\n3D pixmap %lu pixels high by %lu pixels wide by %lu long, or %g Angs high by %g Angs wide by %g Angs long.", xDim, yDim, zDim, xDim*AngsPerPixel, yDim*AngsPerPixel, zDim*AngsPerPixel);
    }


    // last layer is absorber; thickness = 0 for no absorber
    absorberThick = layerThickness0[numberLayers];
    if(simulationType>0) {
        fprintf(stdout, "\nAbsorber: ");
        if(absorberThick) {
            fprintf(stdout, "%.0f A of ", absorberThick);
            for(ielem=0; ielem<numberElements[numberLayers]; ielem++)
                fprintf(stdout, "%.0f%s (%g%%) ", element[numberLayers][ielem]->mass, AtomicNames[element[numberLayers][ielem]->Z], elementFraction[numberLayers][ielem]*100);
        } else
            fprintf(stdout, "not present.");
    }

    if( (ret=fillAllTables())!=0)
        return ret;

    if( isBitmap && (ret=loadBitmap()) != 0)
        return ret;


    // run the simulation(s)
    if (simulationType < 0) {
        // in the case of implantation and full RBS, it's only one simulation
        fprintf(stdout, "\n  1 simulation running"); // TWO SPACES NEEDED BEFORE 1 because CorteoUi expects "%3u" format
        fflush(stdout);
        runSimul(numProjectile, 0);

    } else {
        if(isBitmap) {
            numSimul = 1;
            if(substrateLayer>=0)
                for(ielem=0; ielem<MAXELEMENT; ielem++)
                    numSimul+=nsim[substrateLayer][ielem];
        } else {
            // for IBA without pixmap, it's one for each element in each layer the user wants to simulate
            // first, compute total number of simulations
            numSimul = 0;
            for(ilayer=0; ilayer<MAXLAYERS; ilayer++)
                 for(ielem=0; ielem<MAXELEMENT; ielem++)
                     numSimul+=nsim[ilayer][ielem];
        }

        fprintf(stdout, "\n%3u simulation", numSimul); // Warning: CorteoUi expects "%3u" format
        if(numSimul>1) fprintf(stdout, "s"); // add an "s" if more than one!

        // then run the simulations
        if(isBitmap) {
            fprintf(stdout, "\nRunning simulation for all elements selected for simulation in pixmap\n");
            fflush(stdout);
            // in the case or a bitmap, we don't define the recoil yet, it will be selected during the main
            // collision processing, depending where we are in the bitmap
            runSimul(numProjectile, numberLayers+10); // dummy value for numberLayers

            // substrate simulation is similar to a normal layer simulation
            // (see description of factorNtOmegaByNumdetpart below)
            if(substrateLayer>=0) {
                iSimul = 1; // already one simulation completed
                for(ielem=0; ielem<MAXELEMENT; ielem++) {
                    recoil = element[substrateLayer][ielem];
                    for(j=0;j<nsim[substrateLayer][ielem]; j++) {
                        iSimul++;
                        fprintf(stdout, "\nRunning simulation %u of %u: %.0f%s in sustrate\n",
                                iSimul, numSimul, massIsotope[substrateLayer][ielem][j], AtomicNames[recoil->Z]);
                        fflush(stdout);
                        factorNtOmegaByNumdetpart = d2f(elementFraction[substrateLayer][ielem]*layerThickness0[substrateLayer]
                                                        *layerDensity[substrateLayer]*2.0*PI*(1.0-cosAngSpread0)/numProjectile/(Wf*Hf));
                        // (note that we use numProjectile rather than numProjectileLayer[] below)

                        recoil->mass = massIsotope[substrateLayer][ielem][j]; // set recoil mass to simulate
                        currentCScorrType = CScorrType[substrateLayer][ielem][j];
                        runSimul(numProjectileLayer[substrateLayer]*abundance[substrateLayer][ielem][j], substrateLayer);

                    }
                }

            }
        } else {
            iSimul = 0;
            for(ilayer=0; ilayer<MAXLAYERS; ilayer++) {
                for(ielem=0; ielem<MAXELEMENT; ielem++) {
                    recoil = element[ilayer][ielem];
                    for(j=0;j<nsim[ilayer][ielem]; j++) {
                        iSimul++;
                        fprintf(stdout, "\nRunning simulation %u of %u: %.0f%s in layer %u\n",
                                iSimul, numSimul, massIsotope[ilayer][ielem][j],
                                AtomicNames[recoil->Z], ilayer);
                        fflush(stdout);

                        // Compute factor by which the histogram is multiplied: density*thickness*solidAngle/number_of_detected_particle, with
                        //   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle,
                        //   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)
                        // The spectrum then needs to be multipllied by the number of incident ions q and the concentration
                        // of the detected element in the layer. See documentation for a more detailed explanation.
                        factorNtOmegaByNumdetpart = d2f(elementFraction[ilayer][ielem]*layerThickness0[ilayer]
                                                        *layerDensity[ilayer]*2.0*PI*(1.0-cosAngSpread0)/numProjectileLayer[ilayer]/(Wf*Hf));

                        recoil->mass = massIsotope[ilayer][ielem][j]; // set recoil mass to simulate
                        currentCScorrType = CScorrType[ilayer][ielem][j];
                        isNatural = natural[ilayer][ielem];
                        runSimul(numProjectileLayer[ilayer]*abundance[ilayer][ielem][j], ilayer);
                    }
                }
            }
        }


    }

    fprintf(stderr, "\nSimulation(s) finished.\n");

    // write histograms
    if(simulationType==TECH_IMPREC || simulationType==TECH_IMPLANT) {
        ofp = fopen("impl_beam.out", modew);
        if(ofp) {
            copyinput(ofp);
            for(ichan=0;ichan<NUMBINPROF;ichan++) {
                fprintf(ofp, "%g %g\n", ichan*targetBottom/NUMBINPROF, beamImplHist[ichan]);  // depth is that of the beginning of a channel
            }
            fclose(ofp);
        } else {
            fprintf(stderr, "\nError: unable to open impl_beam.out\n\n");
        }
    }

    if(simulationType==TECH_IMPREC) {
        for(ielem=1;ielem<NUMBERELEMENTS;ielem++) {
            if(dispHist[ielem]) {
                sprintf(filename, "disp%03i.out", ielem);
                ofp = fopen(filename, modew);
                if(ofp) {
                    for(ichan=0;ichan<NUMBINPROF;ichan++) {
                        fprintf(ofp, "%g %g\n", ichan*targetBottom/NUMBINPROF, dispHist[ielem][ichan]);
                    }
                    fclose(ofp);
                }
                sprintf(filename, "impl%03i.out", ielem);
                ofp = fopen(filename, modew);
                if(ofp) {
                    for(ichan=0;ichan<NUMBINPROF;ichan++) {
                        fprintf(ofp, "%g %g\n", ichan*targetBottom/NUMBINPROF, implHist[ielem][ichan]);
                    }
                    fclose(ofp);
                }
            }
        }
    }
    
    if(simulationType==TECH_ENERDEPTHMAP) {
        sprintf(filename,"%s.enermap", filenameWithoutExt);
        ofp = fopen(filename, modew);
        if(ofp) {
            copyinput(ofp);
            fprintf(ofp, "%d %g %g\n", NUMBINPROF, E0, targetBottom);
            for(ichan=0;ichan<NUMBINPROF;ichan++) {
                for(Echan=0;Echan<NUMBINPROF;Echan++) {
                    fprintf(ofp, "%u ", energy_depth_map[ichan*NUMBINPROF+Echan]);
                }
                fprintf(ofp, "\n");
            }
            fclose(ofp);
        } else {
            fprintf(stderr, "\nError: unable to open %s\n", filename);
        }
    }


    // list errors in index functions
    if(EminErr) fprintf(stderr, "%13u reduced energies below minimum value\n", EminErr);
    if(EmaxErr) fprintf(stderr, "%13u reduced energies above maximum value\n", EmaxErr);
    if(SminErr) fprintf(stderr, "%13u reduced impact parameters below minimum value\n", SminErr);
    if(SmaxErr) fprintf(stderr, "%13u reduced impact parameters above maximum value\n", SmaxErr);
    if(DminErr) fprintf(stderr, "%13u stopping power energies below minimum value\n", DminErr);
    if(DmaxErr) fprintf(stderr, "%13u stopping power energies above maximum value\n", DmaxErr);

    if(simulationType>0) {
        fflush(ofdetect);
        fclose(ofdetect);
    }

    return 0;
}



// used by compCS() only
double crossSectionRBSThetaCM(double E, double thetaCM, int screen2) {

  double sinThetaCMby2 = sin(thetaCM*0.5);
  double crossSectionCM, screeningFact = 1.0, temp;
  double massRatio = beam.mass/recoil->mass;

  switch (screen2) {
  case 1: // Andersen
    screeningFact = AndersenScreening(E, sinThetaCMby2, recoil);
  case 0: // Rutherford
    temp = beam.Z*recoil->Z*E2/(2.0f*E/(1.0+massRatio))/(sinThetaCMby2*sinThetaCMby2);
    crossSectionCM = 0.25*temp*temp*screeningFact;  // Goldstein Eq. 3-102 p. 109 x screening
    break;
  case -1: // Rutherford computed with Gauss-Mehler quadrature (for comparison)
    crossSectionCM = crossSectionScreenPot(E/(1.0+massRatio), beam.Z, recoil->Z, thetaCM, 0);
    break;
  case 2: // Universal potential screening function
  case 3: // Lenz-Jensen screening function
    crossSectionCM = crossSectionScreenPot(E/(1.0+massRatio), beam.Z, recoil->Z, thetaCM, screen2);
    break;
  default:
    fprintf(stderr,"\n\nScreening type %u unknown.\n",screen2);
    exit(1);
  }

    // convert cross-section from CM to lab frame (Goldstein Eq. 3-116, p. 117)
    temp = massRatio*cos(thetaCM);
    return d2f(crossSectionCM*pow(1.0+2.0*temp+massRatio*massRatio,1.5)/(1.0+temp));

}

// patch to compute cross-sections
void compCS() {

  double CSRuth, CSRuthQ, CSAnd, CSUniv,CSLJ, massratio, thetaCM, thetaLab, E;
  unsigned int Z1, Z2, j=0, i;
  FILE * off = fopen("compCS.dat","w");

  Z1 = 2;
  Z2 = 79;
  massratio = 4./196.970;

  for(E=10; E<1e7; E*=1.1) {
    if((j++)%(DIME/10)==0) {
      fprintf(stdout, "x");
      fflush(stdout);
    }

    for(i=1; i<180; i+=2) {
      thetaCM = (double)i*PI/180.0;
      thetaLab =  atan(sin(thetaCM)/(cos(thetaCM)+massratio))*180.0/PI;
      if(thetaLab<0) thetaLab+=180.0;
      CSRuth = crossSectionRBSThetaCM(E, thetaCM, 0); // Rutherford (Rutherford CS)
      CSRuthQ = crossSectionRBSThetaCM(E, thetaCM, -1); // Rutherford (gaussian quadrature)
      CSAnd = crossSectionRBSThetaCM(E, thetaCM, 1); // Andersen
      CSUniv = crossSectionRBSThetaCM(E, thetaCM, 2); // Univ
      CSLJ = crossSectionRBSThetaCM(E, thetaCM, 3); //LJ

      fprintf(off,"%g %g %g %g %g %g %g %g\n", E, thetaCM, thetaLab, CSRuth, CSRuth, CSAnd, CSUniv, CSLJ);
    }
  }
  fclose(off);
}

