/*
    Copyright 2007, Francois Schiettekatte

    This file is part of Corteo.

    Corteo is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Corteo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Corteo.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "corteomatrix.h"
#include "constants.h"
#include "corteoindex.h"

// matrix elements calulation parameters
#define NSUM  1000   // number of terms in Gauss-Mehler quadrature sum when computing the matrix
#define NSUM2 100     // number of terms in Gauss-Mehler quadrature sum when evaluation the screened cross-section
#define THETAERR -1000.0

// parameters used to compute matrix, stored in matrix file
#define HEADERSIZE 7
float headerRef[HEADERSIZE] = {MINE, DIME, MAXE, MINS, DIMS, MAXS, NSUM}; 

// scattering matrix: sin^2(theta_CM/2)
float matrix[DIME*DIMS];

// Universal screening function 
double PHIUniv(double x) {
    return 0.1818*exp(-3.2*x)+0.5099*exp(-0.9423*x)+0.2802*exp(-0.4028*x)+0.02817*exp(-0.2016*x);
}

// Lenz-Jensen screening function
double PHILJ(double x) {
  double y = 3.108*sqrt(x);
  return exp(-y)*(1+y+0.3344*y*y+0.0485*y*y*y+2.647e-3*y*y*y*y);
}

// no screening (screening = 1)
double PHInone(double x) {
    return 1.0;
}

// Following functions used to compute scattering angle following Gauss-Mehler quadrature
double H(double u, double x0, double epsilon, double s, double (*PHI)(double x)) {
  double x0u = x0/u;
  double eps1 = 1./epsilon;
  double ss = s*s;
  return sqrt((1-u*u)/(1-PHI(x0u)/x0u*eps1-ss/x0u/x0u));
}

// the root of this function (vs x0) is the minimal 
// approach distance at energy epsilon and impact parameter s
double funcX0(double x0, double epsilon, double s, double (*PHI)(double x)) {
    double sx0 = s/x0;
    return 1.-PHI(x0)/(x0*epsilon)-sx0*sx0;
}

// use bisection (Newton) method to find the minimal approach distance x0
// (return THETAERR on error)
double findX0(double epsilon, double s, double (*PHI)(double x)) {
   double funcX0val1, funcX0val2, funcX0valm;

   double temp = 1.0/(2.0*epsilon);
   double x2 = temp+sqrt(temp+s*s); // inital guesses: Mendenhall & Weller NIMB 58(1991)11, eq. 15 

   double x1 = x2/10.;
   funcX0val1 = funcX0(x1, epsilon, s, PHI);  // should be always negative
   funcX0val2 = funcX0(x2, epsilon, s, PHI);  // should be always positive (starting ~1.0)
   funcX0valm = funcX0((x1+x2)/2., epsilon, s, PHI);
   if(funcX0val1>=0.) {
     // initial guess for x1 too optimistic, start with a safe value (much longer)
     x1 = MINS;
     funcX0val1 = funcX0(x1, epsilon, s, PHI);  // should be always negative
   }

   if(funcX0val1>=0. || funcX0val2<=0.)
     return THETAERR;    // error, values should be on each side of 0

   do {
     if(funcX0valm<0.) {
       funcX0val1 = funcX0valm;
       x1 = (x1+x2)/2.;
     } else {
       funcX0val2 = funcX0valm;
       x2 = (x1+x2)/2.;
     }
     funcX0valm = funcX0((x1+x2)/2., epsilon, s, PHI);
   } while (fabs(funcX0valm)>1e-14);
    // 1.-XXX wont be more precise than 2^(-52)=2e-16
    // but using 1e-10 saves time and is enough precise for float conversion later on

   return (x1+x2)/2.;
}

// use bisection method to find the reduced 
// impact parameter s given epsilon and thetaCM 
// (return THETAERR on error)
double finds(double epsilon, double thetaCM, double (*PHI)(double x)) {
   double funcX0val1, funcX0val2, funcX0valm;

   // inital guesses: Mendenhall & Weller NIMB 58 (1991) 11, eqs. 23-25
   double gamma = (PI-thetaCM)/PI;
   double x0 = findX0((1.0-gamma*gamma)*epsilon, MINS, PHI);
   double x1 = 0.7*gamma*x0;
   double x2 = 1.0/(2.0*epsilon*tan(thetaCM/2.0)); 

   funcX0val1 = thetaCM-THETA(epsilon, x1, NSUM2, PHI);  // should be always negative
   funcX0val2 = thetaCM-THETA(epsilon, x2, NSUM2, PHI);  // should be always positive (starting ~1.0)
   funcX0valm = thetaCM-THETA(epsilon, (x1+x2)/2., NSUM2, PHI);

   if(funcX0val1>=0. || funcX0val2<=0.)
     return THETAERR;    // error, values should be on each side of 0

   do {
     if(funcX0valm<0.) {
       funcX0val1 = funcX0valm;
       x1 = (x1+x2)/2.;
     } else {
       funcX0val2 = funcX0valm;
       x2 = (x1+x2)/2.;
     }
     funcX0valm = thetaCM-THETA(epsilon, (x1+x2)/2., NSUM2, PHI);
   } while (fabs(funcX0valm)>1e-5);

   return (x1+x2)/2.;
}

// scattering angle calculated using Gauss-Mehler quadrature
// (return THETAERR on error)
double THETA(double epsilon, double s, unsigned int nsum, double (*PHI)(double x)) {
  double temp, sum = 0., sum1 = 0;
  double x0 = findX0(epsilon, s, PHI);
  unsigned int j;
  double theta;

  if(x0==THETAERR) {
    return THETAERR;
  }

  for(j=0; j<nsum/2; j++) {
    temp = H(cos((2.*j-1.)*PI/(2.*nsum)),x0, epsilon, s, PHI);
    sum += temp;
  }
  theta = 2.*PI*s/nsum/x0*sum;

  return PI-theta;

}

// compute all the elements of matrix and write matrix to file 'corteo.mat'
// user sets showProgress!=0 to display the progress of this (long) calculation to the console
// return 1 if successful, 0 if not able to write file
int calcMatrix(unsigned int screeningType, int showProgress) {
  unsigned long i, j, nThetaErr = 0;
  double theta, sinThetaBy2;
  char filename[1000];

  // pointer to screening function (I like C!)
  double (*PHI)(double x);

  FILE *ofp;
  char *mode = "wb";

  switch (screeningType) {
  case 0:
    PHI = PHInone; // unscreened potential (screening = 1)
    sprintf(filename, "data/corteo.none.mat");
    break;
  case 2:
    PHI = PHIUniv; // Universal screening function
    sprintf(filename, "data/corteo.univ.mat");
    break;
  case 3:
    PHI = PHILJ; // Lenz-Jensen screening function
    sprintf(filename, "data/corteo.lj.mat");
    break;
  default:
    fprintf(stderr, "Error, screening function of type %u unknown. Stopping here.", screeningType);
    exit(1);
  }

  ofp = fopen(filename, mode);
  if (ofp == NULL) {
      fprintf(stderr, "Error opening matrix file %s. Stopping here.", filename);
      exit(1);
  }

  if(showProgress) fprintf(stdout, "Computing scattering matrix to be stored in %s\n", filename);
  // compute matrix for each reduced energy, reduced impact parameter pair

  for(j=0; j<DIME; j++) {
    if(showProgress && j%10==0) {
      fprintf(stdout, "\b\b\b\b\b\b%.1f%%", j*100./DIME);
      fflush(stdout);
    }
    for(i=0; i<DIMS; i++) {
       // calculations (summation) made using double to decrease numerical noise
       theta = THETA(Eval(j), Sval(i), NSUM, PHI);
       if(theta == THETAERR) nThetaErr++;
       sinThetaBy2 = sin(theta/2.);

       // store in matrix sin^2(theta_CM/2) as float 
       setMatrix(i+j*DIMS,(float)(sinThetaBy2*sinThetaBy2));
    }
  }

  // write matrix, including a header indicating the parameters used to compute matrix
  fwrite((void*)headerRef, HEADERSIZE, sizeof(float), ofp);
  fwrite((void*)matrix,    DIME*DIMS,  sizeof(float), ofp);
  fclose(ofp);

  if(showProgress) { 
    fprintf(stdout, "done,\n");
    fflush(stdout);
  }
  if(nThetaErr) fprintf(stderr, "%lu error(s) evaluating theta\n", nThetaErr);
  
  return 1;
}


// read scattering matrix from file corteo.mat
// return 1 if file read correctly, 0 otherwise
int loadMatrix(void) {
  unsigned int i;
  FILE *ifp;
  char *mode = "rb";

  float header[HEADERSIZE];
  unsigned int tail;

  ifp = fopen("data/corteo.mat", mode);
  if (ifp == NULL) {
    fprintf(stderr,"Can't open data/corteo.mat ");
    return 0;
  }

  // file exists, get header
  fread((void*)header, HEADERSIZE, sizeof(float), ifp);

  for(i=0; i<HEADERSIZE; i++)
    if(header[i] != headerRef[i]) {
	    // "corteo.mat" header does not correspond to currently defined parameters
	    fclose(ifp);
      fprintf(stderr,"Header element %i does not correspond to current parameters in corteo.mat",i);
      return 0;
    }

  // read file into matrix
  fread((void*)matrix, DIME*DIMS, sizeof(float), ifp);

  // read control data
  fread((void*)(&tail), 1, sizeof(int), ifp);
  fclose(ifp);

  return 1;
}


// return element i of matrix
float Matrix(unsigned long i) { 
	return matrix[i]; 
}

// set element i of matrix
void setMatrix(unsigned long i, float val) {
	matrix[i] = val; 
}


// returns the cross section in the center of mass frame considering a screened potential
double crossSectionScreenPot(double E, unsigned int Z1, unsigned int Z2,
                             double thetaCM, unsigned int screeningType) {
  double ds, dsdTheta, s, screenLength, epsilon;

  // pointer to screening function (I like C!)
  double (*PHI)(double x);

  // set screening lenth and screening function
  switch (screeningType) {
  case NOSCREEN:
    screenLength = 0.5291772108; // Bohr radius
    PHI = PHInone; // unscreened potential (screening = 1)
    break;
  case ANDERSEN:
      fprintf(stderr, "Andersen screening not implemented in this function. Stopping here.");
      exit(1);
  case UNIVERSAL:
    screenLength = SCREENCONST/(pow(Z1,0.23)+pow(Z2,  0.23));   // Universal screening length
    PHI = PHIUniv; // Universal screening function
    break;
  case LENZJENSEN:
    screenLength = SCREENCONST/sqrt(pow(Z1,2/3)+pow(Z2, 2/3));  // Lindhard screening length
    PHI = PHILJ; // Lenz-Jensen screening function
    break;
  default:
    fprintf(stderr, "Error, screening function of type %u unknown. Stopping here.", screeningType);
    exit(1);
  }

  // reduced energy according to screening length
  epsilon = E*screenLength/(Z1*Z2*E2);

  // find corresponfding reduced impact parameter 
  s = finds(epsilon, thetaCM, PHI);
  if(s==THETAERR) return 0;

  // ds/dTheta using five-point stencil 
  ds = s*0.01;
  dsdTheta = (12.0*ds)/(-THETA(epsilon,s+2.0*ds,NSUM2, PHI)+8.0*THETA(epsilon,s+ds,NSUM2, PHI)
                        -8.0*THETA(epsilon,s-ds,NSUM2, PHI)+THETA(epsilon,s-2.0*ds,NSUM2, PHI));

  // return "non-reduced" center-of-mass cross section 
  return screenLength*screenLength*s/sin(thetaCM)*fabs(dsdTheta);

}

